@@ -5,7 +5,7 @@
  * hash.h	Structures and prototypes
  *		for fast hashing.
  *
- * Version:	$Id: hash.h,v 1.3 2007/11/23 13:46:55 aland Exp $
+ * Version:	$Id: 332413875ea5ca7090bcdb74cc05d862c2706f6f $
  *
  *   This program is free software; you can redistribute it and/or modify
  *   it under the terms of the GNU General Public License as published by
@@ -25,7 +25,11 @@
  */
 
 #include <freeradius-devel/ident.h>
-RCSIDH(hash_h, "$Id: hash.h,v 1.3 2007/11/23 13:46:55 aland Exp $")
+RCSIDH(hash_h, "$Id: 332413875ea5ca7090bcdb74cc05d862c2706f6f $")
+
+#ifdef __cplusplus
+extern "C" {
+#endif
 
 /*
  *	Fast hash, which isn't too bad.  Don't use for cryptography,
@@ -61,4 +65,9 @@ int		fr_hash_table_num_elements(fr_hash_table_t *ht);
 int		fr_hash_table_walk(fr_hash_table_t *ht,
 				     fr_hash_table_walk_t callback,
 				     void *ctx);
+
+#ifdef __cplusplus
+}
+#endif
+
 #endif /* FR_HASH_H */
