@@ -1,7 +1,7 @@
 /*
  * packet.c	Generic packet manipulation functions.
  *
- * Version:	$Id: packet.c,v 1.21 2008/05/29 15:00:54 aland Exp $
+ * Version:	$Id: 93c2d7ecef1907721ce82987e702c3424171e6e6 $
  *
  *   This library is free software; you can redistribute it and/or
  *   modify it under the terms of the GNU Lesser General Public
@@ -21,7 +21,7 @@
  */
 
 #include	<freeradius-devel/ident.h>
-RCSID("$Id: packet.c,v 1.21 2008/05/29 15:00:54 aland Exp $")
+RCSID("$Id: 93c2d7ecef1907721ce82987e702c3424171e6e6 $")
 
 #include	<freeradius-devel/libradius.h>
 
@@ -154,6 +154,30 @@ int fr_packet_cmp(const RADIUS_PACKET *a, const RADIUS_PACKET *b)
 }
 
 
+int fr_inaddr_any(fr_ipaddr_t *ipaddr)
+{
+
+	if (ipaddr->af == AF_INET) {
+		if (ipaddr->ipaddr.ip4addr.s_addr == INADDR_ANY) {
+			return 1;
+		}
+
+#ifdef HAVE_STRUCT_SOCKADDR_IN6
+	} else if (ipaddr->af == AF_INET6) {
+		if (IN6_IS_ADDR_UNSPECIFIED(&(ipaddr->ipaddr.ip6addr))) {
+			return 1;
+		}
+#endif
+
+	} else {
+		fr_strerror_printf("Unknown address family");
+		return -1;
+	}
+
+	return 0;
+}
+
+
 /*
  *	Create a fake "request" from a reply, for later lookup.
  */
@@ -175,18 +199,17 @@ void fr_request_from_reply(RADIUS_PACKET *request,
 int fr_socket(fr_ipaddr_t *ipaddr, int port)
 {
 	int sockfd;
-	uint16_t sport;
 	struct sockaddr_storage salocal;
 	socklen_t	salen;
 
 	if ((port < 0) || (port > 65535)) {
-		librad_log("Port %d is out of allowed bounds", port);
+		fr_strerror_printf("Port %d is out of allowed bounds", port);
 		return -1;
 	}
 
 	sockfd = socket(ipaddr->af, SOCK_DGRAM, 0);
 	if (sockfd < 0) {
-		librad_log("cannot open socket: %s", strerror(errno));
+		fr_strerror_printf("cannot open socket: %s", strerror(errno));
 		return sockfd;
 	}
 
@@ -196,36 +219,18 @@ int fr_socket(fr_ipaddr_t *ipaddr, int port)
 	 */
 	if (udpfromto_init(sockfd) != 0) {
 		close(sockfd);
-		librad_log("cannot initialize udpfromto: %s", strerror(errno));
+		fr_strerror_printf("cannot initialize udpfromto: %s", strerror(errno));
 		return -1;
 	}
 #endif
 
-	sport = port;
-	sport = htons(sport);
-	memset(&salocal, 0, sizeof(salocal));
-	if (ipaddr->af == AF_INET) {
-		struct sockaddr_in s4;
 
-		s4.sin_family = AF_INET;
-		s4.sin_addr = ipaddr->ipaddr.ip4addr;
-		s4.sin_port = sport;
-		salen = sizeof(s4);
-		memset(&salocal, 0, sizeof(salocal));
-		memcpy(&salocal, &s4, salen);
+	if (!fr_ipaddr2sockaddr(ipaddr, port, &salocal, &salen)) {
+		return sockfd;
+	}
 
 #ifdef HAVE_STRUCT_SOCKADDR_IN6
-	} else if (ipaddr->af == AF_INET6) {
-		struct sockaddr_in6 s6;
-
-		s6.sin6_family = AF_INET6;
-		s6.sin6_addr = ipaddr->ipaddr.ip6addr;
-		s6.sin6_port = sport;
-		salen = sizeof(s6);
-		memset(&salocal, 0, sizeof(salocal));
-		memcpy(&salocal, &s6, salen);
-
-#if 1
+	if (ipaddr->af == AF_INET6) {
 		/*
 		 *	Listening on '::' does NOT get you IPv4 to
 		 *	IPv6 mapping.  You've got to listen on an IPv4
@@ -241,15 +246,36 @@ int fr_socket(fr_ipaddr_t *ipaddr, int port)
 				   (char *)&on, sizeof(on));
 		}
 #endif /* IPV6_V6ONLY */
-#endif
+	}
 #endif /* HAVE_STRUCT_SOCKADDR_IN6 */
-	} else {
-		return sockfd;	/* don't bind it */
+
+	if (ipaddr->af == AF_INET) {
+		UNUSED int flag;
+		
+#if defined(IP_MTU_DISCOVER) && defined(IP_PMTUDISC_DONT)
+		/*
+		 *	Disable PMTU discovery.  On Linux, this
+		 *	also makes sure that the "don't fragment"
+		 *	flag is zero.
+		 */
+		flag = IP_PMTUDISC_DONT;
+		setsockopt(sockfd, IPPROTO_IP, IP_MTU_DISCOVER,
+			   &flag, sizeof(flag));
+#endif
+
+#if defined(IP_DONTFRAG)
+		/*
+		 *	Ensure that the "don't fragment" flag is zero.
+		 */
+		flag = 0;
+		setsockopt(sockfd, IPPROTO_IP, IP_DONTFRAG,
+			   &flag, sizeof(flag));
+#endif
 	}
 
 	if (bind(sockfd, (struct sockaddr *) &salocal, salen) < 0) {
 		close(sockfd);
-		librad_log("cannot bind socket: %s", strerror(errno));
+		fr_strerror_printf("cannot bind socket: %s", strerror(errno));
 		return -1;
 	}
 
@@ -284,9 +310,9 @@ typedef struct fr_packet_socket_t {
  *	that should be managed.
  */
 struct fr_packet_list_t {
-	fr_hash_table_t *ht;
+	rbtree_t	*tree;
 
-	fr_hash_table_t *dst2id_ht;
+	fr_hash_table_t	*dst2id_ht;
 
 	int		alloc_id;
 	int		num_outgoing;
@@ -380,29 +406,20 @@ int fr_packet_list_socket_add(fr_packet_list_t *pl, int sockfd)
 		return 0;
 	}
 
+	if (!fr_sockaddr2ipaddr(&src, sizeof_src, &ps->ipaddr, &ps->port)) {
+		return 0;
+	}
+
 	/*
 	 *	Grab IP addresses & ports from the sockaddr.
 	 */
-	ps->ipaddr.af = src.ss_family;
 	if (src.ss_family == AF_INET) {
-		struct sockaddr_in	s4;
-
-		memcpy(&s4, &src, sizeof(s4));
-		ps->ipaddr.ipaddr.ip4addr = s4.sin_addr;
-		ps->port = ntohs(s4.sin_port);
-
 		if (ps->ipaddr.ipaddr.ip4addr.s_addr == INADDR_ANY) {
 			ps->inaddr_any = 1;
 		}
 
 #ifdef HAVE_STRUCT_SOCKADDR_IN6
 	} else if (src.ss_family == AF_INET6) {
-		struct sockaddr_in6	s6;
-
-		memcpy(&s6, &src, sizeof(s6));
-		ps->ipaddr.ipaddr.ip6addr = s6.sin6_addr;
-		ps->port = ntohs(s6.sin6_port);
-
 		if (IN6_IS_ADDR_UNSPECIFIED(&ps->ipaddr.ipaddr.ip6addr)) {
 			ps->inaddr_any = 1;
 		}
@@ -415,16 +432,13 @@ int fr_packet_list_socket_add(fr_packet_list_t *pl, int sockfd)
 	return 1;
 }
 
-static uint32_t packet_entry_hash(const void *data)
-{
-	return fr_request_packet_hash(*(const RADIUS_PACKET * const *) data);
-}
-
 static int packet_entry_cmp(const void *one, const void *two)
 {
 	const RADIUS_PACKET * const *a = one;
 	const RADIUS_PACKET * const *b = two;
 
+	if (!a || !*a || !b || !*b) return -1; /* work-around for bug #35 */
+
 	return fr_packet_cmp(*a, *b);
 }
 
@@ -496,7 +510,7 @@ void fr_packet_list_free(fr_packet_list_t *pl)
 {
 	if (!pl) return;
 
-	fr_hash_table_free(pl->ht);
+	rbtree_free(pl->tree);
 	fr_hash_table_free(pl->dst2id_ht);
 	free(pl);
 }
@@ -514,10 +528,8 @@ fr_packet_list_t *fr_packet_list_create(int alloc_id)
 	if (!pl) return NULL;
 	memset(pl, 0, sizeof(*pl));
 
-	pl->ht = fr_hash_table_create(packet_entry_hash,
-					packet_entry_cmp,
-					NULL);
-	if (!pl->ht) {
+	pl->tree = rbtree_create(packet_entry_cmp, NULL, 0);
+	if (!pl->tree) {
 		fr_packet_list_free(pl);
 		return NULL;
 	}
@@ -551,9 +563,7 @@ int fr_packet_list_insert(fr_packet_list_t *pl,
 {
 	if (!pl || !request_p || !*request_p) return 0;
 
-	(*request_p)->hash = fr_request_packet_hash(*request_p);
-
-	return fr_hash_table_insert(pl->ht, request_p);
+	return rbtree_insert(pl->tree, request_p);
 }
 
 RADIUS_PACKET **fr_packet_list_find(fr_packet_list_t *pl,
@@ -561,7 +571,7 @@ RADIUS_PACKET **fr_packet_list_find(fr_packet_list_t *pl,
 {
 	if (!pl || !request) return 0;
 
-	return fr_hash_table_finddata(pl->ht, &request);
+	return rbtree_finddata(pl->tree, &request);
 }
 
 
@@ -598,27 +608,33 @@ RADIUS_PACKET **fr_packet_list_find_byreply(fr_packet_list_t *pl,
 
 	my_request.dst_ipaddr = reply->src_ipaddr;
 	my_request.dst_port = reply->src_port;
-	my_request.hash = 0;
 
 	request = &my_request;
 
-	return fr_hash_table_finddata(pl->ht, &request);
+	return rbtree_finddata(pl->tree, &request);
 }
 
 
 RADIUS_PACKET **fr_packet_list_yank(fr_packet_list_t *pl,
 				      RADIUS_PACKET *request)
 {
+	RADIUS_PACKET **packet_p;
+
 	if (!pl || !request) return NULL;
 
-	return fr_hash_table_yank(pl->ht, &request);
+	packet_p = rbtree_finddata(pl->tree, &request);
+	if (!packet_p) return NULL;
+
+	rbtree_deletebydata(pl->tree, packet_p);
+	return packet_p;
+
 }
 
 int fr_packet_list_num_elements(fr_packet_list_t *pl)
 {
 	if (!pl) return 0;
 
-	return fr_hash_table_num_elements(pl->ht);
+	return rbtree_num_elements(pl->tree);
 }
 
 
@@ -639,11 +655,46 @@ int fr_packet_list_id_alloc(fr_packet_list_t *pl,
 			      RADIUS_PACKET *request)
 {
 	int i, id, start;
+	int src_any = 0;
 	uint32_t free_mask;
 	fr_packet_dst2id_t my_pd, *pd;
 	fr_packet_socket_t *ps;
 
-	if (!pl || !pl->alloc_id || !request) return 0;
+	if (!pl || !pl->alloc_id || !request) {
+		fr_strerror_printf("Invalid arguments");
+		return 0;
+	}
+
+	/*
+	 *	Error out if no destination is specified.
+	 */
+	if ((request->dst_ipaddr.af == AF_UNSPEC) ||
+	    (request->dst_port == 0)) {
+		fr_strerror_printf("No destination address/port specified");
+		return 0;
+	}
+
+	/*
+	 *	Special case: unspec == "don't care"
+	 */
+	if (request->src_ipaddr.af == AF_UNSPEC) {
+		memset(&request->src_ipaddr, 0, sizeof(request->src_ipaddr));
+		request->src_ipaddr.af = request->dst_ipaddr.af;
+	}
+
+	src_any = fr_inaddr_any(&request->src_ipaddr);
+	if (src_any < 0) {
+		fr_strerror_printf("Error checking src IP address");
+		return 0;
+	}
+
+	/*
+	 *	MUST specify a destination address.
+	 */
+	if (fr_inaddr_any(&request->dst_ipaddr) != 0) {
+		fr_strerror_printf("Error checking dst IP address");
+		return 0;
+	}
 
 	my_pd.dst_ipaddr = request->dst_ipaddr;
 	my_pd.dst_port = request->dst_port;
@@ -660,6 +711,7 @@ int fr_packet_list_id_alloc(fr_packet_list_t *pl,
 
 		if (!fr_hash_table_insert(pl->dst2id_ht, pd)) {
 			free(pd);
+			fr_strerror_printf("Failed inserting into hash");
 			return 0;
 		}
 	}
@@ -680,7 +732,10 @@ int fr_packet_list_id_alloc(fr_packet_list_t *pl,
 	while (pd->id[id] == pl->mask) { /* all sockets are using this ID */
 		id++;
 		id &= 0xff;
-		if (id == start) return 0;
+		if (id == start) {
+			fr_strerror_printf("All IDs are being used");
+			return 0;
+		}
 	}
 
 	free_mask = ~((~pd->id[id]) & pl->mask);
@@ -689,13 +744,38 @@ int fr_packet_list_id_alloc(fr_packet_list_t *pl,
 	for (i = 0; i < MAX_SOCKETS; i++) {
 		if (pl->sockets[i].sockfd == -1) continue; /* paranoia */
 
+		ps = &(pl->sockets[i]);
+
+		/*
+		 *	Address families don't match, skip it.
+		 */
+		if (ps->ipaddr.af != request->dst_ipaddr.af) continue;
+
+		/*
+		 *	We're sourcing from *, and they asked for a
+		 *	specific source address: ignore it.
+		 */
+		if (ps->inaddr_any && !src_any) continue;
+
+		/*
+		 *	We're sourcing from a specific IP, and they
+		 *	asked for a source IP that isn't us: ignore
+		 *	it.
+		 */
+		if (!ps->inaddr_any && !src_any &&
+		    (fr_ipaddr_cmp(&request->src_ipaddr,
+				   &ps->ipaddr) != 0)) continue;
+
 		if ((free_mask & (1 << i)) == 0) {
 			start = i;
 			break;
 		}
 	}
 
-	if (start < 0) return 0; /* bad error */
+	if (start < 0) {
+		fr_strerror_printf("Internal sanity check failed");
+		return 0; /* bad error */
+	}
 
 	pd->id[id] |= (1 << start);
 	ps = &pl->sockets[start];
@@ -737,7 +817,6 @@ int fr_packet_list_id_free(fr_packet_list_t *pl,
 	if (!pd) return 0;
 
 	pd->id[request->id] &= ~(1 << ps->offset);
-	request->hash = 0;	/* invalidate the cached hash */
 
 	ps->num_outgoing--;
 	pl->num_outgoing--;
@@ -750,7 +829,7 @@ int fr_packet_list_walk(fr_packet_list_t *pl, void *ctx,
 {
 	if (!pl || !callback) return 0;
 
-	return fr_hash_table_walk(pl->ht, callback, ctx);
+	return rbtree_walk(pl->tree, InOrder, callback, ctx);
 }
 
 int fr_packet_list_fd_set(fr_packet_list_t *pl, fd_set *set)
@@ -817,7 +896,7 @@ int fr_packet_list_num_incoming(fr_packet_list_t *pl)
 
 	if (!pl) return 0;
 
-	num_elements = fr_hash_table_num_elements(pl->ht);
+	num_elements = rbtree_num_elements(pl->tree);
 	if (num_elements < pl->num_outgoing) return 0; /* panic! */
 
 	return num_elements - pl->num_outgoing;
