@@ -1,7 +1,7 @@
 /*
  * valuepair.c	Functions to handle VALUE_PAIRs
  *
- * Version:	$Id$
+ * Version:	$Id: fcfdb4bbc3c6e852f5c636e0a2dfca57fcd586ef $
  *
  *   This library is free software; you can redistribute it and/or
  *   modify it under the terms of the GNU Lesser General Public
@@ -21,7 +21,7 @@
  */
 
 #include <freeradius-devel/ident.h>
-RCSID("$Id$")
+RCSID("$Id: fcfdb4bbc3c6e852f5c636e0a2dfca57fcd586ef $")
 
 #include	<freeradius-devel/libradius.h>
 
@@ -135,6 +135,33 @@ VALUE_PAIR *pairalloc(DICT_ATTR *da)
 }
 
 
+VALUE_PAIR *paircreate_raw(int attr, int type, VALUE_PAIR *vp)
+{
+	char *p = (char *) (vp + 1);
+
+	if (!vp->flags.unknown_attr) {
+		pairfree(&vp);
+		return NULL;
+	}
+
+	vp->vendor = VENDOR(attr);
+	vp->attribute = attr;
+	vp->operator = T_OP_EQ;
+	vp->name = p;
+	vp->type = type;
+	vp->length = 0;
+	memset(&vp->flags, 0, sizeof(vp->flags));
+	vp->flags.unknown_attr = 1;
+	
+	if (!vp_print_name(p, FR_VP_NAME_LEN, vp->attribute)) {
+		free(vp);
+		return NULL;
+	}
+
+	return vp;
+}
+
+
 /*
  *	Create a new valuepair.
  */
@@ -153,19 +180,7 @@ VALUE_PAIR *paircreate(int attr, int type)
 	/*
 	 *	It isn't in the dictionary: update the name.
 	 */
-	if (!da) {
-		char *p = (char *) (vp + 1);
-		
-		vp->vendor = VENDOR(attr);
-		vp->attribute = attr;
-		vp->name = p;
-		vp->type = type; /* be forgiving */
-
-		if (!vp_print_name(p, FR_VP_NAME_LEN, vp->attribute)) {
-			free(vp);
-			return NULL;
-		}
-	}
+	if (!da) return paircreate_raw(attr, type, vp);
 
 	return vp;
 }
@@ -308,6 +323,8 @@ VALUE_PAIR *paircopyvp(const VALUE_PAIR *vp)
 {
 	size_t name_len;
 	VALUE_PAIR *n;
+
+	if (!vp) return NULL;
 	
 	if (!vp->flags.unknown_attr) {
 		name_len = 0;
@@ -320,6 +337,12 @@ VALUE_PAIR *paircopyvp(const VALUE_PAIR *vp)
 		return NULL;
 	}
 	memcpy(n, vp, sizeof(*n) + name_len);
+
+	/*
+	 *	Reset the name field to point to the NEW attribute,
+	 *	rather than to the OLD one.
+	 */
+	if (vp->flags.unknown_attr) n->name = (char *) (n + 1);
 	n->next = NULL;
 
 	if ((n->type == PW_TYPE_TLV) &&
@@ -422,6 +445,8 @@ void pairmove(VALUE_PAIR **to, VALUE_PAIR **from)
 			case T_OP_CMP_TRUE:
 			case T_OP_CMP_FALSE:
 			case T_OP_CMP_EQ:
+			case T_OP_REG_EQ:
+			case T_OP_REG_NE:
 				tailfrom = i;
 				continue;
 
@@ -791,13 +816,11 @@ static int gettime(const char *valstr, time_t *date)
 		f[2] = strchr(f[1], ':'); /* find : separator */
 		if (f[2]) {
 		  *(f[2]++) = '\0';	/* nuke it, and point to SS */
-		} else {
-		  strcpy(f[2], "0");	/* assignment would discard const */
-		}
+		  tm->tm_sec = atoi(f[2]);
+		}			/* else leave it as zero */
 
 		tm->tm_hour = atoi(f[0]);
 		tm->tm_min = atoi(f[1]);
-		tm->tm_sec = atoi(f[2]);
 	}
 
 	/*
@@ -821,6 +844,27 @@ static const char *hextab = "0123456789abcdef";
  *  double-check the parsed value, to be sure it's legal for that
  *  type (length, etc.)
  */
+static uint32_t getint(const char *value, char **end)
+{
+	if ((value[0] == '0') && (value[1] == 'x')) {
+		return strtoul(value, end, 16);
+	}
+
+	return strtoul(value, end, 10);
+}
+
+static int check_for_whitespace(const char *value)
+{
+	while (*value) {
+		if (!isspace((int) *value)) return 0;
+
+		value++;
+	}
+
+	return 1;
+}
+
+
 VALUE_PAIR *pairparsevalue(VALUE_PAIR *vp, const char *value)
 {
 	char		*p, *s=0;
@@ -829,6 +873,8 @@ VALUE_PAIR *pairparsevalue(VALUE_PAIR *vp, const char *value)
 	size_t		length;
 	DICT_VALUE	*dval;
 
+	if (!value) return NULL;
+
 	/*
 	 *	Even for integers, dates and ip addresses we
 	 *	keep the original string in vp->vp_strvalue.
@@ -872,6 +918,10 @@ VALUE_PAIR *pairparsevalue(VALUE_PAIR *vp, const char *value)
 						c = '\'';
 						cp++;
 						break;
+					case '\\':
+						c = '\\';
+						cp++;
+						break;
 					case '`':
 						c = '`';
 						cp++;
@@ -895,6 +945,7 @@ VALUE_PAIR *pairparsevalue(VALUE_PAIR *vp, const char *value)
 				*p++ = c;
 				length++;
 			}
+			vp->vp_strvalue[length] = '\0';
 			vp->length = length;
 			break;
 
@@ -916,7 +967,7 @@ VALUE_PAIR *pairparsevalue(VALUE_PAIR *vp, const char *value)
 				cs = s = strdup(value);
 				if (!s) return NULL;
 				p = strrchr(s, '+');
-				*p = 0;
+				if (p) *p = 0;
 				vp->flags.addport = 1;
 			} else {
 				p = NULL;
@@ -927,8 +978,8 @@ VALUE_PAIR *pairparsevalue(VALUE_PAIR *vp, const char *value)
 				fr_ipaddr_t ipaddr;
 
 				if (ip_hton(cs, AF_INET, &ipaddr) < 0) {
-					free(s);
 					fr_strerror_printf("Failed to find IP address for %s", cs);
+					free(s);
 					return NULL;
 				}
 
@@ -939,73 +990,48 @@ VALUE_PAIR *pairparsevalue(VALUE_PAIR *vp, const char *value)
 			break;
 
 		case PW_TYPE_BYTE:
-			if (value && (value[0] == '0') && (value[1] == 'x')) {
-				goto do_octets;
-			}
+			vp->length = 1;
 
 			/*
 			 *	Note that ALL integers are unsigned!
 			 */
-			vp->vp_integer = (uint32_t) strtoul(value, &p, 10);
+			vp->vp_integer = getint(value, &p);
 			if (!*p) {
 				if (vp->vp_integer > 255) {
 					fr_strerror_printf("Byte value \"%s\" is larger than 255", value);
 					return NULL;
 				}
-				vp->length = 1;
 				break;
 			}
-
-			/*
-			 *	Look for the named value for the given
-			 *	attribute.
-			 */
-			if ((dval = dict_valbyname(vp->attribute, value)) == NULL) {
-				fr_strerror_printf("Unknown value %s for attribute %s",
-					   value, vp->name);
-				return NULL;
-			}
-			vp->vp_integer = dval->value;
-			vp->length = 1;
-			break;
+			if (check_for_whitespace(p)) break;
+			goto check_for_value;
 
 		case PW_TYPE_SHORT:
 			/*
 			 *	Note that ALL integers are unsigned!
 			 */
-			vp->vp_integer = (uint32_t) strtoul(value, &p, 10);
+			vp->vp_integer = getint(value, &p);
+			vp->length = 2;
 			if (!*p) {
 				if (vp->vp_integer > 65535) {
 					fr_strerror_printf("Byte value \"%s\" is larger than 65535", value);
 					return NULL;
 				}
-				vp->length = 2;
 				break;
 			}
-
-			/*
-			 *	Look for the named value for the given
-			 *	attribute.
-			 */
-			if ((dval = dict_valbyname(vp->attribute, value)) == NULL) {
-				fr_strerror_printf("Unknown value %s for attribute %s",
-					   value, vp->name);
-				return NULL;
-			}
-			vp->vp_integer = dval->value;
-			vp->length = 2;
-			break;
+			if (check_for_whitespace(p)) break;
+			goto check_for_value;
 
 		case PW_TYPE_INTEGER:
 			/*
 			 *	Note that ALL integers are unsigned!
 			 */
-			vp->vp_integer = (uint32_t) strtoul(value, &p, 10);
-			if (!*p) {
-				vp->length = 4;
-				break;
-			}
+			vp->vp_integer = getint(value, &p);
+			vp->length = 4;
+			if (!*p) break;
+			if (check_for_whitespace(p)) break;
 
+	check_for_value:
 			/*
 			 *	Look for the named value for the given
 			 *	attribute.
@@ -1016,7 +1042,6 @@ VALUE_PAIR *pairparsevalue(VALUE_PAIR *vp, const char *value)
 				return NULL;
 			}
 			vp->vp_integer = dval->value;
-			vp->length = 4;
 			break;
 
 		case PW_TYPE_DATE:
@@ -1048,8 +1073,10 @@ VALUE_PAIR *pairparsevalue(VALUE_PAIR *vp, const char *value)
 			}
 
 		  	if (ascend_parse_filter(vp) < 0 ) {
-				fr_strerror_printf("failed to parse Ascend binary attribute: %s",
-					   fr_strerror());
+				char buffer[256];
+
+				snprintf(buffer, sizeof(buffer), "failed to parse Ascend binary attribute: %s", fr_strerror());
+				fr_strerror_printf("%s", buffer);
 				return NULL;
 			}
 			break;
@@ -1059,8 +1086,8 @@ VALUE_PAIR *pairparsevalue(VALUE_PAIR *vp, const char *value)
 			 *	then fall through to raw octets, so that
 			 *	the user can at least make them by hand...
 			 */
-#endif
 	do_octets:
+#endif
 			/* raw octets: 0x01020304... */
 		case PW_TYPE_OCTETS:
 			if (strncasecmp(value, "0x", 2) == 0) {
@@ -1106,12 +1133,21 @@ VALUE_PAIR *pairparsevalue(VALUE_PAIR *vp, const char *value)
 			break;
 
 		case PW_TYPE_IPV6ADDR:
-			if (inet_pton(AF_INET6, value, &vp->vp_ipv6addr) <= 0) {
-				fr_strerror_printf("failed to parse IPv6 address "
-					   "string \"%s\"", value);
-				return NULL;
+			{
+				fr_ipaddr_t ipaddr;
+
+				if (ip_hton(value, AF_INET6, &ipaddr) < 0) {
+					char buffer[1024];
+
+					strlcpy(buffer, fr_strerror(), sizeof(buffer));
+
+					fr_strerror_printf("failed to parse IPv6 address "
+                                                           "string \"%s\": %s", value, buffer);
+					return NULL;
+				}
+				vp->vp_ipv6addr = ipaddr.ipaddr.ip6addr;
+				vp->length = 16; /* length of IPv6 address */
 			}
-			vp->length = 16; /* length of IPv6 address */
 			break;
 
 		case PW_TYPE_IPV6PREFIX:
@@ -1257,11 +1293,11 @@ static VALUE_PAIR *pairmake_any(const char *attribute, const char *value,
 	 *	Unknown attributes MUST be of type 'octets'
 	 */
 	if (value && (strncasecmp(value, "0x", 2) != 0)) {
-		fr_strerror_printf("Invalid octet string \"%s\" for attribute name \"%s\"", value, attribute);
+		fr_strerror_printf("Unknown attribute \"%s\" requires a hex string, not \"%s\"", attribute, value);
 		return NULL;
 	}
 
-	attr = vendor = 0;
+	vendor = 0;
 
 	/*
 	 *	Pull off vendor prefix first.
@@ -1372,6 +1408,9 @@ static VALUE_PAIR *pairmake_any(const char *attribute, const char *value,
 		return NULL;
 	}
 
+	vp->operator = (operator == 0) ? T_OP_EQ : operator;
+	if (!value) return vp;
+
 	size = strlen(value + 2);
 
 	/*
@@ -1385,10 +1424,20 @@ static VALUE_PAIR *pairmake_any(const char *attribute, const char *value,
 		vp->type = PW_TYPE_OCTETS;
 		/* FALL-THROUGH */
 		
-	case PW_TYPE_STRING:
 	case PW_TYPE_OCTETS:
 	case PW_TYPE_ABINARY:
 		vp->length = size >> 1;
+		if (vp->length > sizeof(vp->vp_octets)) {
+			vp->length = sizeof(vp->vp_octets);
+		}
+		break;
+
+	case PW_TYPE_STRING:
+		vp->length = size >> 1;
+		memset(&vp->vp_strvalue, 0, sizeof(vp->vp_strvalue));
+		if (vp->length >= sizeof(vp->vp_strvalue)) {
+			vp->length = sizeof(vp->vp_strvalue) - 1;
+		}
 		break;
 	}
 
@@ -1414,8 +1463,6 @@ static VALUE_PAIR *pairmake_any(const char *attribute, const char *value,
 		break;
 	}
        
-	vp->operator = (operator == 0) ? T_OP_EQ : operator;
-
 	return vp;
 }
 
@@ -1490,13 +1537,12 @@ VALUE_PAIR *pairmake(const char *attribute, const char *value, int operator)
 	if (value && (*value == ':' && da->flags.has_tag)) {
 	        /* If we already found a tag, this is invalid */
 	        if(found_tag) {
-		        pairbasicfree(vp);
 			fr_strerror_printf("Duplicate tag %s for attribute %s",
 				   value, vp->name);
 			DEBUG("Duplicate tag %s for attribute %s\n",
 				   value, vp->name);
+		        pairbasicfree(vp);
 			return NULL;
-
 		}
 	        /* Colon found and attribute allows a tag */
 	        if (value[1] == '*' && value[2] == ':') {
@@ -1539,6 +1585,15 @@ VALUE_PAIR *pairmake(const char *attribute, const char *value, int operator)
 		 */
 	case T_OP_REG_EQ:	/* =~ */
 	case T_OP_REG_NE:	/* !~ */
+		if (!value) {
+			fr_strerror_printf("No regular expression found in %s",
+					   vp->name);
+		        pairbasicfree(vp);
+			return NULL;
+		}
+	  
+		strlcpy(vp->vp_strvalue, value, sizeof(vp->vp_strvalue));
+		vp->length = strlen(vp->vp_strvalue);
 		/*
 		 *	If anything goes wrong, this is a run-time error,
 		 *	not a compile-time error.
@@ -1619,7 +1674,7 @@ VALUE_PAIR *pairread(const char **ptr, FR_TOKEN *eol)
 
 	q = attr;
 	for (len = 0; len < sizeof(attr); len++) {
-		if (valid_attr_name[(int)*p]) {
+	  if (valid_attr_name[(int)*((const uint8_t *)p)]) {
 			*q++ = *p++;
 			continue;
 		}
@@ -1635,7 +1690,7 @@ VALUE_PAIR *pairread(const char **ptr, FR_TOKEN *eol)
 	/*
 	 *	We may have Foo-Bar:= stuff, so back up.
 	 */
-	if (attr[len - 1] == ':') {
+	if ((len > 0) && (attr[len - 1] == ':')) {
 		p--;
 		len--;
 	}
@@ -1643,6 +1698,12 @@ VALUE_PAIR *pairread(const char **ptr, FR_TOKEN *eol)
 	attr[len] = '\0';
 	*ptr = p;
 
+	if (!*attr) {
+		*eol = T_OP_INVALID;
+		fr_strerror_printf("Invalid attribute name");
+		return NULL;
+	}
+
 	/* Now we should have an operator here. */
 	token = gettoken(ptr, buf, sizeof(buf));
 	if (token < T_EQSTART || token > T_EQEND) {
@@ -1778,7 +1839,7 @@ VALUE_PAIR *pairread(const char **ptr, FR_TOKEN *eol)
  */
 FR_TOKEN userparse(const char *buffer, VALUE_PAIR **first_pair)
 {
-	VALUE_PAIR	*vp;
+	VALUE_PAIR	*vp, *head, **tail;
 	const char	*p;
 	FR_TOKEN	last_token = T_OP_INVALID;
 	FR_TOKEN	previous_token;
@@ -1789,20 +1850,30 @@ FR_TOKEN userparse(const char *buffer, VALUE_PAIR **first_pair)
 	if (buffer[0] == 0)
 		return T_EOL;
 
+	head = NULL;
+	tail = &head;
+
 	p = buffer;
 	do {
 		previous_token = last_token;
 		if ((vp = pairread(&p, &last_token)) == NULL) {
-			return last_token;
+			break;
 		}
-		pairadd(first_pair, vp);
+		*tail = vp;
+		tail = &((*tail)->next);
 	} while (*p && (last_token == T_COMMA));
 
 	/*
 	 *	Don't tell the caller that there was a comment.
 	 */
 	if (last_token == T_HASH) {
-		return previous_token;
+		last_token = previous_token;
+	}
+
+	if (last_token == T_OP_INVALID) {
+		pairfree(&head);
+	} else {
+		pairadd(first_pair, head);
 	}
 
 	/*
@@ -1880,6 +1951,8 @@ VALUE_PAIR *readvp2(FILE *fp, int *pfiledone, const char *errprefix)
  *	e.g. "foo" != "bar"
  *
  *	Returns true (comparison is true), or false (comparison is not true);
+ *
+ *	FIXME: Ignores tags!
  */
 int paircmp(VALUE_PAIR *one, VALUE_PAIR *two)
 {
@@ -1972,11 +2045,23 @@ int paircmp(VALUE_PAIR *one, VALUE_PAIR *two)
 	case PW_TYPE_SHORT:
 	case PW_TYPE_INTEGER:
 	case PW_TYPE_DATE:
-		compare = two->vp_integer - one->vp_integer;
+		if (two->vp_integer < one->vp_integer) {
+			compare = -1;
+		} else if (two->vp_integer == one ->vp_integer) {
+			compare = 0;
+		} else {
+			compare = +1;
+		}
 		break;
 
 	case PW_TYPE_IPADDR:
-		compare = ntohl(two->vp_ipaddr) - ntohl(one->vp_ipaddr);
+		if (ntohl(two->vp_ipaddr)  < ntohl(one->vp_ipaddr) ) {
+			compare = -1;
+		} else if (ntohl(two->vp_ipaddr)  == ntohl(one->vp_ipaddr) ) {
+			compare = 0;
+		} else {
+			compare = +1;
+		}
 		break;
 
 	case PW_TYPE_IPV6ADDR:
