@@ -1,7 +1,7 @@
 /*
  * auth.c	User authentication.
  *
- * Version:	$Id$
+ * Version:	$Id: fd36a262d76c3c0b515e719f38be293d0876673f $
  *
  *   This program is free software; you can redistribute it and/or modify
  *   it under the terms of the GNU General Public License as published by
@@ -23,7 +23,7 @@
  */
 
 #include <freeradius-devel/ident.h>
-RCSID("$Id$")
+RCSID("$Id: fd36a262d76c3c0b515e719f38be293d0876673f $")
 
 #include <freeradius-devel/radiusd.h>
 #include <freeradius-devel/modules.h>
@@ -40,16 +40,25 @@ char *auth_name(char *buf, size_t buflen, REQUEST *request, int do_cli)
 	VALUE_PAIR	*cli;
 	VALUE_PAIR	*pair;
 	int		port = 0;
+	const char	*tls = "";
 
 	if ((cli = pairfind(request->packet->vps, PW_CALLING_STATION_ID)) == NULL)
 		do_cli = 0;
 	if ((pair = pairfind(request->packet->vps, PW_NAS_PORT)) != NULL)
 		port = pair->vp_integer;
 
+	if (request->packet->dst_port == 0) {
+		if (pairfind(request->packet->vps, PW_FREERADIUS_PROXIED_TO)) {
+			tls = " via TLS tunnel";
+		} else {
+			tls = " via proxy to virtual server";
+		}
+	}
+
 	snprintf(buf, buflen, "from client %.128s port %u%s%.128s%s",
 			request->client->shortname, port,
 		 (do_cli ? " cli " : ""), (do_cli ? (char *)cli->vp_strvalue : ""),
-		 (request->packet->dst_port == 0) ? " via TLS tunnel" : "");
+		 tls);
 
 	return buf;
 }
@@ -60,11 +69,14 @@ char *auth_name(char *buf, size_t buflen, REQUEST *request, int do_cli)
  * Make sure user/pass are clean
  * and then log them
  */
-static int rad_authlog(const char *msg, REQUEST *request, int goodpass) {
-
+static int rad_authlog(const char *msg, REQUEST *request, int goodpass)
+{
+	int logit;
+	const char *extra_msg = NULL;
 	char clean_password[1024];
 	char clean_username[1024];
 	char buf[1024];
+	char extra[1024];
 	VALUE_PAIR *username = NULL;
 
 	if (!request->root->log_auth) {
@@ -100,10 +112,11 @@ static int rad_authlog(const char *msg, REQUEST *request, int goodpass) {
 
 			auth_type = pairfind(request->config_items,
 					     PW_AUTH_TYPE);
-			if (auth_type && (auth_type->vp_strvalue[0] != '\0')) {
+			if (auth_type) {
 				snprintf(clean_password, sizeof(clean_password),
 					 "<via Auth-Type = %s>",
-					 auth_type->vp_strvalue);
+					 dict_valnamebyattr(PW_AUTH_TYPE,
+							    auth_type->vp_integer));
 			} else {
 				strcpy(clean_password, "<no User-Password attribute>");
 			}
@@ -117,21 +130,29 @@ static int rad_authlog(const char *msg, REQUEST *request, int goodpass) {
 	}
 
 	if (goodpass) {
-		radlog_request(L_AUTH, 0, request, "%s: [%s%s%s] (%s)",
-				msg,
-				clean_username,
-				request->root->log_auth_goodpass ? "/" : "",
-				request->root->log_auth_goodpass ? clean_password : "",
-				auth_name(buf, sizeof(buf), request, 1));
+		logit = request->root->log_auth_goodpass;
+		extra_msg = request->root->auth_goodpass_msg;
 	} else {
-		radlog_request(L_AUTH, 0, request, "%s: [%s%s%s] (%s)",
-				msg,
-				clean_username,
-				request->root->log_auth_badpass ? "/" : "",
-				request->root->log_auth_badpass ? clean_password : "",
-				auth_name(buf, sizeof(buf), request, 1));
+		logit = request->root->log_auth_badpass;
+		extra_msg = request->root->auth_badpass_msg;
 	}
 
+	if (extra_msg) {
+		extra[0] = ' ';
+		radius_xlat(extra + 1, sizeof(extra) - 1, extra_msg, request,
+			    NULL);
+	} else {
+		*extra = '\0';
+	}
+
+	radlog_request(L_AUTH, 0, request, "%s: [%s%s%s] (%s)%s",
+		       msg,
+		       clean_username,
+		       logit ? "/" : "",
+		       logit ? clean_password : "",
+		       auth_name(buf, sizeof(buf), request, 1),
+		       extra);
+
 	return 0;
 }
 
@@ -164,13 +185,13 @@ static int rad_check_password(REQUEST *request)
 	 */
 	cur_config_item = request->config_items;
 	while(((auth_type_pair = pairfind(cur_config_item, PW_AUTH_TYPE))) != NULL) {
+
 		auth_type = auth_type_pair->vp_integer;
 		auth_type_count++;
-		DICT_VALUE *dv = dict_valbyattr(auth_type_pair->attribute,
-						auth_type_pair->vp_integer);
 
 		RDEBUG2("Found Auth-Type = %s",
-			(dv != NULL) ? dv->name : "?");
+			dict_valnamebyattr(PW_AUTH_TYPE,
+					   auth_type_pair->vp_integer));
 		cur_config_item = auth_type_pair->next;
 
 		if (auth_type == PW_AUTHTYPE_REJECT) {
@@ -251,7 +272,7 @@ static int rad_check_password(REQUEST *request)
 		 	*
 		 	*	This is fail-safe.
 		 	*/
-			RDEBUG2("No authenticate method (Auth-Type) configuration found for the request: Rejecting the user");
+			RDEBUG2("ERROR: No authenticate method (Auth-Type) found for the request: Rejecting the user");
 			return -2;
 		}
 	}
@@ -284,6 +305,7 @@ static int rad_check_password(REQUEST *request)
 			case -1:
 			  rad_authlog("Login incorrect "
 						  "(system failed to supply an encrypted password for comparison)", request, 0);
+			  /* FALL-THROUGH */
 			case 1:
 			  return -1;
 			}
@@ -404,7 +426,8 @@ int rad_postauth(REQUEST *request)
 	 */
 	vp = pairfind(request->config_items, PW_POST_AUTH_TYPE);
 	if (vp) {
-		RDEBUG2("Using Post-Auth-Type %s", vp->vp_strvalue);
+		RDEBUG2("Using Post-Auth-Type %s",
+			dict_valnamebyattr(PW_POST_AUTH_TYPE, vp->vp_integer));
 		postauth_type = vp->vp_integer;
 	}
 	result = module_post_auth(postauth_type, request);
@@ -477,6 +500,9 @@ int rad_authenticate(REQUEST *request)
 						&request->config_items,
 						PW_AUTH_TYPE, PW_TYPE_INTEGER);
 			if (tmp) tmp->vp_integer = PW_AUTHTYPE_ACCEPT;
+#ifdef WITH_POST_PROXY_AUTHORIZE
+			if (mainconfig.post_proxy_authorize) break;
+#endif
 			goto authenticate;
 
 		/*
@@ -494,11 +520,15 @@ int rad_authenticate(REQUEST *request)
 		 *	done by the server, by rejecting them here.
 		 */
 		case PW_AUTHENTICATION_REJECT:
-		default:
 			rad_authlog("Login incorrect (Home Server says so)",
 				    request, 0);
 			request->reply->code = PW_AUTHENTICATION_REJECT;
 			return RLM_MODULE_REJECT;
+
+		default:
+			rad_authlog("Login incorrect (Home Server failed to respond)",
+				    request, 0);
+			return RLM_MODULE_REJECT;
 		}
 	}
 #endif
@@ -576,8 +606,9 @@ autz_redo:
 	if (!autz_retry) {
 		tmp = pairfind(request->config_items, PW_AUTZ_TYPE);
 		if (tmp) {
-			RDEBUG2("Using Autz-Type %s", tmp->vp_strvalue);
 			autz_type = tmp->vp_integer;
+			RDEBUG2("Using Autz-Type %s",
+				dict_valnamebyattr(PW_AUTZ_TYPE, autz_type));
 			autz_retry = 1;
 			goto autz_redo;
 		}
@@ -664,15 +695,18 @@ autz_redo:
 		/* double check: maybe the secret is wrong? */
 		if ((debug_flag > 1) && (auth_item != NULL) &&
 				(auth_item->attribute == PW_USER_PASSWORD)) {
-			char *p;
+			uint8_t *p;
 
-			p = auth_item->vp_strvalue;
-			while (*p != '\0') {
-				if (!isprint((int) *p)) {
-					log_debug("  WARNING: Unprintable characters in the password.\n\t  Double-check the shared secret on the server and the NAS!");
+			p = (uint8_t *) auth_item->vp_strvalue;
+			while (*p) {
+				int size;
+
+				size = fr_utf8_char(p);
+				if (!size) {
+					log_debug("  WARNING: Unprintable characters in the password.  Double-check the shared secret on the server and the NAS!");
 					break;
 				}
-				p++;
+				p += size;
 			}
 		}
 	}
@@ -687,8 +721,9 @@ autz_redo:
 
 		tmp = pairfind(request->config_items, PW_SESSION_TYPE);
 		if (tmp) {
-			RDEBUG2("Using Session-Type %s", tmp->vp_strvalue);
 			session_type = tmp->vp_integer;
+			RDEBUG2("Using Session-Type %s",
+				dict_valnamebyattr(PW_SESSION_TYPE, session_type));
 		}
 
 		/*
@@ -777,10 +812,11 @@ autz_redo:
 	 *	Set the reply to Access-Accept, if it hasn't already
 	 *	been set to something.  (i.e. Access-Challenge)
 	 */
-	if (request->reply->code == 0)
-	  request->reply->code = PW_AUTHENTICATION_ACK;
-
-	if ((module_msg = pairfind(request->packet->vps,PW_MODULE_SUCCESS_MESSAGE)) != NULL){
+	if (request->reply->code == 0) {
+		request->reply->code = PW_AUTHENTICATION_ACK;
+	}
+	
+	if ((module_msg = pairfind(request->packet->vps, PW_MODULE_SUCCESS_MESSAGE)) != NULL){
 		char msg[MAX_STRING_LEN+12];
 
 		snprintf(msg, sizeof(msg), "Login OK (%s)",
@@ -790,10 +826,38 @@ autz_redo:
 		rad_authlog("Login OK", request, 1);
 	}
 
+	return result;
+}
+
+/*
+ *	Run a virtual server auth and postauth
+ *
+ */
+int rad_virtual_server(REQUEST *request)
+{
+	VALUE_PAIR *vp;
+	int result;
+
 	/*
-	 *	Run the modules in the 'post-auth' section.
+	 *	We currently only handle AUTH packets here.
+	 *	This could be expanded to handle other packets as well if required.
 	 */
-	result = rad_postauth(request);
+	rad_assert(request->packet->code == PW_AUTHENTICATION_REQUEST);
+
+	result = rad_authenticate(request);
+
+        if (request->reply->code == PW_AUTHENTICATION_REJECT) {
+                pairdelete(&request->config_items, PW_POST_AUTH_TYPE);
+                vp = radius_pairmake(request, &request->config_items,
+                                     "Post-Auth-Type", "Reject",
+                                     T_OP_SET);
+                if (vp) rad_postauth(request);
+        }
+
+        if (request->reply->code == PW_AUTHENTICATION_ACK) {
+                rad_postauth(request);
+        }
 
 	return result;
 }
+
