@@ -1,7 +1,7 @@
 /*
  * mainconf.c	Handle the server's configuration.
  *
- * Version:	$Id$
+ * Version:	$Id: da0c16c01ff21557362a028214a9776c0f6c7b24 $
  *
  *   This program is free software; you can redistribute it and/or modify
  *   it under the terms of the GNU General Public License as published by
@@ -22,7 +22,7 @@
  */
 
 #include <freeradius-devel/ident.h>
-RCSID("$Id$")
+RCSID("$Id: da0c16c01ff21557362a028214a9776c0f6c7b24 $")
 
 #include <freeradius-devel/radiusd.h>
 #include <freeradius-devel/modules.h>
@@ -62,6 +62,10 @@ struct main_config_t mainconfig;
 char *request_log_file = NULL;
 char *debug_condition = NULL;
 
+#ifdef HAVE_GMTIME_R
+extern int log_dates_utc;
+#endif
+
 typedef struct cached_config_t {
 	struct cached_config_t *next;
 	time_t		created;
@@ -92,8 +96,15 @@ static const char *radlog_dest = NULL;
  */
 static const char *localstatedir = NULL;
 static const char *prefix = NULL;
+static const char *my_name = NULL;
+static const char *sbindir = NULL;
+static const char *run_dir = NULL;
 static char *syslog_facility = NULL;
-static const FR_NAME_NUMBER str2fac[] = {
+
+/*
+ *	Syslog facility table.
+ */
+const FR_NAME_NUMBER syslog_str2fac[] = {
 #ifdef LOG_KERN
 	{ "kern", LOG_KERN },
 #endif
@@ -182,6 +193,9 @@ static const CONF_PARSER serverdest_config[] = {
 	{ "log", PW_TYPE_SUBSECTION, 0, NULL, (const void *) logdest_config },
 	{ "log_file", PW_TYPE_STRING_PTR, 0, &mainconfig.log_file, NULL },
 	{ "log_destination", PW_TYPE_STRING_PTR, 0, &radlog_dest, NULL },
+#ifdef HAVE_GMTIME_R
+	{ "use_utc", PW_TYPE_BOOLEAN, 0, &log_dates_utc, NULL },
+#endif
 	{ NULL, -1, 0, NULL, NULL }
 };
 
@@ -195,6 +209,10 @@ static const CONF_PARSER log_config_nodest[] = {
 	{ "msg_badpass", PW_TYPE_STRING_PTR, 0, &mainconfig.auth_badpass_msg, NULL},
 	{ "msg_goodpass", PW_TYPE_STRING_PTR, 0, &mainconfig.auth_goodpass_msg, NULL},
 
+#ifdef HAVE_GMTIME_R
+	{ "use_utc", PW_TYPE_BOOLEAN, 0, &log_dates_utc, NULL },
+#endif
+
 	{ NULL, -1, 0, NULL, NULL }
 };
 
@@ -210,9 +228,12 @@ static const CONF_PARSER server_config[] = {
 	 *	hard-coded defines for the locations of the various
 	 *	files.
 	 */
+	{ "name",               PW_TYPE_STRING_PTR, 0, &my_name,          "radiusd"},
 	{ "prefix",             PW_TYPE_STRING_PTR, 0, &prefix,            "/usr/local"},
 	{ "localstatedir",      PW_TYPE_STRING_PTR, 0, &localstatedir,     "${prefix}/var"},
+	{ "sbindir",            PW_TYPE_STRING_PTR, 0, &sbindir,            "${prefix}/sbin"},
 	{ "logdir",             PW_TYPE_STRING_PTR, 0, &radlog_dir,        "${localstatedir}/log"},
+	{ "run_dir",            PW_TYPE_STRING_PTR, 0, &run_dir,           "${localstatedir}/run/${name}"},
 	{ "libdir",             PW_TYPE_STRING_PTR, 0, &radlib_dir,        "${prefix}/lib"},
 	{ "radacctdir",         PW_TYPE_STRING_PTR, 0, &radacct_dir,       "${logdir}/radacct" },
 	{ "hostname_lookups",   PW_TYPE_BOOLEAN,    0, &fr_dns_lookups,      "no" },
@@ -251,8 +272,10 @@ static const CONF_PARSER server_config[] = {
 };
 
 static const CONF_PARSER bootstrap_config[] = {
+#ifdef HAVE_SETUID
 	{ "user",  PW_TYPE_STRING_PTR, 0, &uid_name, NULL },
 	{ "group",  PW_TYPE_STRING_PTR, 0, &gid_name, NULL },
+#endif
 	{ "chroot",  PW_TYPE_STRING_PTR, 0, &chroot_dir, NULL },
 	{ "allow_core_dumps", PW_TYPE_BOOLEAN, 0, &allow_core_dumps, "no" },
 
@@ -354,10 +377,13 @@ static size_t xlat_config(void *instance, REQUEST *request,
 	 *  If 'outlen' is too small, then the output is chopped to fit.
 	 */
 	value = cf_pair_value(cp);
-	if (value) {
-		if (outlen > strlen(value)) {
-			outlen = strlen(value) + 1;
-		}
+	if (!value) {
+		out[0] = '\0';
+		return 0;
+	}
+
+	if (outlen > strlen(value)) {
+		outlen = strlen(value) + 1;
 	}
 
 	return func(out, outlen, value);
@@ -786,7 +812,7 @@ int read_mainconfig(int reload)
 	snprintf(buffer, sizeof(buffer), "%.200s/%.50s.conf",
 		 radius_dir, mainconfig.name);
 	if ((cs = cf_file_read(buffer)) == NULL) {
-		radlog(L_ERR, "Errors reading %s", buffer);
+		radlog(L_ERR, "Errors reading or parsing %s", buffer);
 		return -1;
 	}
 
@@ -826,7 +852,7 @@ int read_mainconfig(int reload)
 				cf_section_free(&cs);
 				return -1;
 			}
-			mainconfig.syslog_facility = fr_str2int(str2fac, syslog_facility, -1);
+			mainconfig.syslog_facility = fr_str2int(syslog_str2fac, syslog_facility, -1);
 			if (mainconfig.syslog_facility < 0) {
 				fprintf(stderr, "radiusd: Error: Unknown syslog_facility %s\n",
 					syslog_facility);
@@ -834,11 +860,13 @@ int read_mainconfig(int reload)
 				return -1;
 			}
 
+#ifdef HAVE_SYSLOG_H
 			/*
 			 *	Call openlog only once, when the
 			 *	program starts.
 			 */
 			openlog(progname, LOG_PID, mainconfig.syslog_facility);
+#endif
 
 		} else if (mainconfig.radlog_dest == RADLOG_FILES) {
 			if (!mainconfig.log_file) {
@@ -901,11 +929,13 @@ int read_mainconfig(int reload)
 
 	DEBUG2("%s: #### Loading Realms and Home Servers ####", mainconfig.name);
 	if (!realms_init(cs)) {
+		DEBUG2("Failed to load realms and home servers\n");
 		return -1;
 	}
 
 	DEBUG2("%s: #### Loading Clients ####", mainconfig.name);
 	if (!clients_parse_section(cs)) {
+		DEBUG2("Failed to load clients\n");
 		return -1;
 	}
 
@@ -991,6 +1021,29 @@ int free_mainconfig(void)
 	return 0;
 }
 
+void hup_logfile(void)
+{
+		int fd, old_fd;
+		
+		if (mainconfig.radlog_dest != RADLOG_FILES) return;
+
+ 		fd = open(mainconfig.log_file,
+			  O_WRONLY | O_APPEND | O_CREAT, 0640);
+		if (fd >= 0) {
+			/*
+			 *	Atomic swap. We'd like to keep the old
+			 *	FD around so that callers don't
+			 *	suddenly find the FD closed, and the
+			 *	writes go nowhere.  But that's hard to
+			 *	do.  So... we have the case where a
+			 *	log message *might* be lost on HUP.
+			 */
+			old_fd = mainconfig.radlog_fd;
+			mainconfig.radlog_fd = fd;
+			close(old_fd);
+		}
+}
+
 void hup_mainconfig(void)
 {
 	cached_config_t *cc;
@@ -1003,7 +1056,7 @@ void hup_mainconfig(void)
 	snprintf(buffer, sizeof(buffer), "%.200s/%.50s.conf",
 		 radius_dir, mainconfig.name);
 	if ((cs = cf_file_read(buffer)) == NULL) {
-		radlog(L_ERR, "Failed to re-read %s", buffer);
+		radlog(L_ERR, "Failed to re-read or parse %s", buffer);
 		return;
 	}
 
@@ -1031,25 +1084,7 @@ void hup_mainconfig(void)
 	 *	The "open log file" code is here rather than in log.c,
 	 *	because it makes that function MUCH simpler.
 	 */
-	if (mainconfig.radlog_dest == RADLOG_FILES) {
-		int fd, old_fd;
-		
-		fd = open(mainconfig.log_file,
-			  O_WRONLY | O_APPEND | O_CREAT, 0640);
-		if (fd >= 0) {
-			/*
-			 *	Atomic swap. We'd like to keep the old
-			 *	FD around so that callers don't
-			 *	suddenly find the FD closed, and the
-			 *	writes go nowhere.  But that's hard to
-			 *	do.  So... we have the case where a
-			 *	log message *might* be lost on HUP.
-			 */
-			old_fd = mainconfig.radlog_fd;
-			mainconfig.radlog_fd = fd;
-			close(old_fd);
-		}
-	}
+	hup_logfile();
 
 	radlog(L_INFO, "HUP - loading modules");
 
