@@ -1,7 +1,7 @@
 /*
  * radiusd.c	Main loop of the radius server.
  *
- * Version:	$Id$
+ * Version:	$Id: bb84e829df001f7b97f191f46587d6910857ac80 $
  *
  *   This program is free software; you can redistribute it and/or modify
  *   it under the terms of the GNU General Public License as published by
@@ -26,7 +26,7 @@
  */
 
 #include <freeradius-devel/ident.h>
-RCSID("$Id$")
+RCSID("$Id: bb84e829df001f7b97f191f46587d6910857ac80 $")
 
 #include <freeradius-devel/radiusd.h>
 #include <freeradius-devel/modules.h>
@@ -65,12 +65,14 @@ int log_stripped_names;
 int debug_flag = 0;
 int check_config = FALSE;
 
-const char *radiusd_version = "FreeRADIUS Version " RADIUSD_VERSION ", for host " HOSTINFO ", built on " __DATE__ " at " __TIME__;
+const char *radiusd_version = "FreeRADIUS Version " RADIUSD_VERSION_STRING
+#ifdef RADIUSD_VERSION_COMMIT
+" (git #" RADIUSD_VERSION_COMMIT ")"
+#endif
+", for host " HOSTINFO ", built on " __DATE__ " at " __TIME__;
 
 pid_t radius_pid;
 
-static int debug_memory = 0;
-
 /*
  *  Configuration items.
  */
@@ -169,8 +171,18 @@ int main(int argc, char *argv[])
 				break;
 
 			case 'l':
+				if (strcmp(optarg, "stdout") == 0) {
+					goto do_stdout;
+				}
 				mainconfig.log_file = strdup(optarg);
 				mainconfig.radlog_dest = RADLOG_FILES;
+				mainconfig.radlog_fd = open(mainconfig.log_file,
+							    O_WRONLY | O_APPEND | O_CREAT, 0640);
+				if (mainconfig.radlog_fd < 0) {
+					fprintf(stderr, "radiusd: Failed to open log file %s: %s\n", mainconfig.log_file, strerror(errno));
+					exit(1);
+				}
+				fr_log_fp = fdopen(mainconfig.radlog_fd, "a");
 				break;		  
 
 			case 'i':
@@ -186,7 +198,7 @@ int main(int argc, char *argv[])
 				break;
 
 			case 'm':
-				debug_memory = 1;
+				mainconfig.debug_memory = 1;
 				break;
 
 			case 'p':
@@ -209,9 +221,14 @@ int main(int argc, char *argv[])
 				break;
 
 			case 'v':
+				/* Don't print timestamps */
+				debug_flag += 2;
+				fr_log_fp = stdout;
+				mainconfig.radlog_dest = RADLOG_STDOUT;
+				mainconfig.radlog_fd = STDOUT_FILENO;
+				
 				version();
-				break;
-
+				exit(0);
 			case 'X':
 				spawn_flag = FALSE;
 				dont_fork = TRUE;
@@ -219,9 +236,10 @@ int main(int argc, char *argv[])
 				mainconfig.log_auth = TRUE;
 				mainconfig.log_auth_badpass = TRUE;
 				mainconfig.log_auth_goodpass = TRUE;
+		do_stdout:
+				fr_log_fp = stdout;
 				mainconfig.radlog_dest = RADLOG_STDOUT;
 				mainconfig.radlog_fd = STDOUT_FILENO;
-				fr_log_fp = stdout;
 				break;
 
 			case 'x':
@@ -234,20 +252,22 @@ int main(int argc, char *argv[])
 		}
 	}
 
+	/*
+	 *	Mismatch between build time OpenSSL and linked SSL,
+	 *	better to die here than segfault later.
+	 */
+	if (ssl_check_version() < 0) {
+		exit(1);
+	}
+
 	if (flag && (flag != 0x03)) {
 		fprintf(stderr, "radiusd: The options -i and -p cannot be used individually.\n");
 		exit(1);
 	}
 
-	if (debug_flag) {
-		radlog(L_INFO, "%s", radiusd_version);
-		radlog(L_INFO, "Copyright (C) 1999-2008 The FreeRADIUS server project and contributors.\n");
-		radlog(L_INFO, "There is NO warranty; not even for MERCHANTABILITY or FITNESS FOR A\n");
-		radlog(L_INFO, "PARTICULAR PURPOSE.\n");
-		radlog(L_INFO, "You may redistribute copies of FreeRADIUS under the terms of the\n");
-		radlog(L_INFO, "GNU General Public License v2.\n");
-		fflush(NULL);
-	}
+	if (debug_flag)
+		version();
+		
 
 	/*  Read the configuration files, BEFORE doing anything else.  */
 	if (read_mainconfig(0) < 0) {
@@ -284,30 +304,6 @@ int main(int argc, char *argv[])
 	 */
 	radius_pid = getpid();
 
-	/*
-	 *  Only write the PID file if we're running as a daemon.
-	 *
-	 *  And write it AFTER we've forked, so that we write the
-	 *  correct PID.
-	 */
-	if (dont_fork == FALSE) {
-		FILE *fp;
-
-		fp = fopen(mainconfig.pid_file, "w");
-		if (fp != NULL) {
-			/*
-			 *	FIXME: What about following symlinks,
-			 *	and having it over-write a normal file?
-			 */
-			fprintf(fp, "%d\n", (int) radius_pid);
-			fclose(fp);
-		} else {
-			radlog(L_ERR|L_CONS, "Failed creating PID file %s: %s\n",
-			       mainconfig.pid_file, strerror(errno));
-			exit(1);
-		}
-	}
-
 	/*
 	 *	If we're running as a daemon, close the default file
 	 *	descriptors, AFTER forking.
@@ -323,18 +319,26 @@ int main(int argc, char *argv[])
 		}
 		dup2(devnull, STDIN_FILENO);
 		if (mainconfig.radlog_dest == RADLOG_STDOUT) {
-			mainconfig.radlog_fd = dup(STDOUT_FILENO);
+			setlinebuf(stdout);
+			mainconfig.radlog_fd = STDOUT_FILENO;
+		} else {
+			dup2(devnull, STDOUT_FILENO);
 		}
-		dup2(devnull, STDOUT_FILENO);
 		if (mainconfig.radlog_dest == RADLOG_STDERR) {
-			mainconfig.radlog_fd = dup(STDERR_FILENO);
+			setlinebuf(stderr);
+			mainconfig.radlog_fd = STDERR_FILENO;
+		} else {
+			dup2(devnull, STDERR_FILENO);
 		}
-		dup2(devnull, STDERR_FILENO);
 		close(devnull);
 
 	} else {
 		setlinebuf(stdout); /* unbuffered output */
 	}
+	
+	/*
+	 *	Now we have logging check that the OpenSSL 
+	 */
 
 	/*
 	 *	Initialize the event pool, including threads.
@@ -365,7 +369,7 @@ int main(int argc, char *argv[])
 	 *	server to die immediately.  Use SIGTERM to shut down
 	 *	the server cleanly in that case.
 	 */
-	if ((debug_memory == 1) || (debug_flag == 0)) {
+	if ((mainconfig.debug_memory == 1) || (debug_flag == 0)) {
 #ifdef HAVE_SIGACTION
 	        act.sa_handler = sig_fatal;
 		sigaction(SIGINT, &act, NULL);
@@ -388,17 +392,47 @@ int main(int argc, char *argv[])
 
 	radius_stats_init(0);
 
+	/*
+	 *  Only write the PID file if we're running as a daemon.
+	 *
+	 *  And write it AFTER we've forked, so that we write the
+	 *  correct PID.
+	 */
+	if (dont_fork == FALSE) {
+		FILE *fp;
+
+		fp = fopen(mainconfig.pid_file, "w");
+		if (fp != NULL) {
+			/*
+			 *	FIXME: What about following symlinks,
+			 *	and having it over-write a normal file?
+			 */
+			fprintf(fp, "%d\n", (int) radius_pid);
+			fclose(fp);
+		} else {
+			radlog(L_ERR|L_CONS, "Failed creating PID file %s: %s\n",
+			       mainconfig.pid_file, strerror(errno));
+			exit(1);
+		}
+	}
+
 	/*
 	 *	Process requests until HUP or exit.
 	 */
 	while ((rcode = radius_event_process()) == 0x80) {
 		radius_stats_init(1);
-		radlog(L_INFO, "Received HUP.");
-		module_hup(cf_section_sub_find(mainconfig.config, "modules"));
+		hup_mainconfig();
 	}
-	
-	radlog(L_INFO, "Exiting normally.");
-	
+
+	if (rcode < 0) {
+		radlog(L_ERR, "Exiting due to internal error: %s",
+		       fr_strerror());
+		rcode = 2;
+	} else {
+		radlog(L_INFO, "Exiting normally.");
+		rcode = 1;
+	}
+
 	/*
 	 *	Ignore the TERM signal: we're
 	 *	about to die.
@@ -426,11 +460,6 @@ int main(int argc, char *argv[])
 		
 	radius_event_free();
 	
-	/*
-	 *	Free the configuration items.
-	 */
-	free_mainconfig();
-	
 	/*
 	 *	Detach any modules.
 	 */
@@ -438,6 +467,11 @@ int main(int argc, char *argv[])
 	
 	xlat_free();		/* modules may have xlat's */
 
+	/*
+	 *	Free the configuration items.
+	 */
+	free_mainconfig();
+	
 	free(radius_dir);
 		
 #ifdef WIN32
@@ -462,10 +496,13 @@ static void NEVER_RETURNS usage(int status)
 	fprintf(output, "  -d raddb_dir    Configuration files are in \"raddbdir/*\".\n");
 	fprintf(output, "  -f              Run as a foreground process, not a daemon.\n");
 	fprintf(output, "  -h              Print this help message.\n");
-	fprintf(output, "  -i ipaddr       Listen on ipaddr ONLY\n");
+	fprintf(output, "  -i ipaddr       Listen on ipaddr ONLY.\n");
+	fprintf(output, "  -l log_file     Logging output will be written to this file.\n");
+	fprintf(output, "  -m              On SIGINT or SIGQUIT exit cleanly instead of immediately.\n");
 	fprintf(output, "  -n name         Read raddb/name.conf instead of raddb/radiusd.conf\n");
-	fprintf(output, "  -p port         Listen on port ONLY\n");
+	fprintf(output, "  -p port         Listen on port ONLY.\n");
 	fprintf(output, "  -s              Do not spawn child processes to handle requests.\n");
+	fprintf(output, "  -t              Disable threads.\n");
 	fprintf(output, "  -v              Print server version information.\n");
 	fprintf(output, "  -X              Turn on full debugging.\n");
 	fprintf(output, "  -x              Turn on additional debugging. (-xx gives more debugging).\n");
@@ -478,14 +515,9 @@ static void NEVER_RETURNS usage(int status)
  */
 static void sig_fatal(int sig)
 {
-	switch(sig) {
-		case SIGSEGV:
-			/*
-			 *	We can't really do anything
-			 *	intelligent here so just die
-			 */
-			_exit(1);
+	if (getpid() != radius_pid) _exit(sig);
 
+	switch(sig) {
 		case SIGTERM:
 			radius_signal_self(RADIUS_SIGNAL_SELF_TERM);
 			break;
@@ -494,15 +526,14 @@ static void sig_fatal(int sig)
 #ifdef SIGQUIT
 		case SIGQUIT:
 #endif
-			if (debug_memory) {
+			if (mainconfig.debug_memory) {
 				radius_signal_self(RADIUS_SIGNAL_SELF_TERM);
 				break;
 			}
 			/* FALL-THROUGH */
 
 		default:
-			radius_signal_self(RADIUS_SIGNAL_SELF_EXIT);
-			break;
+			_exit(sig);
 	}
 }
 
