@@ -1,7 +1,7 @@
 /*
  * vmps.c	Handle VMPS traffic.
  *
- * Version:	$Id: vmps.c,v 1.6 2007/12/15 18:10:53 aland Exp $
+ * Version:	$Id: b88b11a189f69f5e76af696896caf6599d14c7c5 $
  *
  *   This program is free software; you can redistribute it and/or modify
  *   it under the terms of the GNU General Public License as published by
@@ -22,11 +22,10 @@
  */
 
 #include <freeradius-devel/ident.h>
-RCSID("$Id: vmps.c,v 1.6 2007/12/15 18:10:53 aland Exp $")
+RCSID("$Id: b88b11a189f69f5e76af696896caf6599d14c7c5 $")
 
 #include <freeradius-devel/radiusd.h>
 #include <freeradius-devel/modules.h>
-#include <freeradius-devel/radius_snmp.h>
 #include <freeradius-devel/vqp.h>
 #include <freeradius-devel/vmps.h>
 #include <freeradius-devel/rad_assert.h>
@@ -43,24 +42,17 @@ int vqp_socket_recv(rad_listen_t *listener,
 {
 	RADIUS_PACKET	*packet;
 	RAD_REQUEST_FUNP fun = NULL;
-	char		buffer[128];
 	RADCLIENT	*client;
 
 	packet = vqp_recv(listener->fd);
 	if (!packet) {
-		radlog(L_ERR, "%s", librad_errstr);
+		radlog(L_ERR, "%s", fr_strerror());
 		return 0;
 	}
 
 	if ((client = client_listener_find(listener,
-					   &packet->src_ipaddr)) == NULL) {
-		RAD_SNMP_TYPE_INC(listener, total_invalid_requests);
-		
-		radlog(L_ERR, "Ignoring request from unknown client %s port %d",
-		       inet_ntop(packet->src_ipaddr.af,
-				 &packet->src_ipaddr.ipaddr,
-				 buffer, sizeof(buffer)),
-		       packet->src_port);
+					   &packet->src_ipaddr,
+					   packet->src_port)) == NULL) {
 		rad_free(&packet);
 		return 0;
 	}
@@ -90,7 +82,7 @@ int vqp_socket_send(rad_listen_t *listener, REQUEST *request)
 	rad_assert(listener->send == vqp_socket_send);
 
 	if (vqp_encode(request->reply, request->packet) < 0) {
-		DEBUG2("Failed encoding packet: %s\n", librad_errstr);
+		DEBUG2("Failed encoding packet: %s\n", fr_strerror());
 		return -1;
 	}
 
