@@ -2,7 +2,7 @@
  * xlat.c	Translate strings.  This is the first version of xlat
  * 		incorporated to RADIUS
  *
- * Version:	$Id$
+ * Version:	$Id: 9b97db0820836850b3569f64ff216e09efd594c5 $
  *
  *   This program is free software; you can redistribute it and/or modify
  *   it under the terms of the GNU General Public License as published by
@@ -23,13 +23,17 @@
  */
 
 #include <freeradius-devel/ident.h>
-RCSID("$Id$")
+RCSID("$Id: 9b97db0820836850b3569f64ff216e09efd594c5 $")
 
 #include	<freeradius-devel/radiusd.h>
 #include	<freeradius-devel/rad_assert.h>
+#include	<freeradius-devel/base64.h>
+#include	<freeradius-devel/dhcp.h>
 
 #include	<ctype.h>
 
+extern int log_dates_utc; /* log.c */
+
 typedef struct xlat_t {
 	char		module[MAX_STRING_LEN];
 	int		length;
@@ -50,6 +54,7 @@ static const char * const internal_xlat[] = {"check",
 					     "proxy-reply",
 					     "outer.request",
 					     "outer.reply",
+					     "outer.control",
 					     NULL};
 
 #if REQUEST_MAX_REGEX > 8
@@ -57,9 +62,8 @@ static const char * const internal_xlat[] = {"check",
 #endif
 static const int xlat_inst[] = { 0, 1, 2, 3, 4, 5, 6, 7, 8 };	/* up to 8 for regex */
 
-
-/*
- *	Convert the value on a VALUE_PAIR to string
+/**
+ * @brief Convert the value on a VALUE_PAIR to string
  */
 static int valuepair2str(char * out,int outlen,VALUE_PAIR * pair,
 			 int type, RADIUS_ESCAPE_STRING func)
@@ -93,6 +97,23 @@ static int valuepair2str(char * out,int outlen,VALUE_PAIR * pair,
 	return strlen(out);
 }
 
+static VALUE_PAIR *pairfind_tag(VALUE_PAIR *vps, int attr, int tag)
+{
+	VALUE_PAIR *vp = vps;
+
+redo:
+	vp = pairfind(vp, attr);
+	if (!tag) return vp;
+
+	if (!vp) return NULL;
+
+	if (!vp->flags.has_tag) return NULL;
+
+	if (vp->flags.tag == tag) return vp;
+	
+	vp = vp->next;
+	goto redo;
+}
 
 /*
  *	Dynamically translate for check:, request:, reply:, etc.
@@ -149,6 +170,12 @@ static size_t xlat_packet(void *instance, REQUEST *request,
 		}
 		break;
 			
+	case 7:
+		if (request->parent) {
+			vps = request->parent->config_items;
+		}
+		break;
+			
 	default:		/* WTF? */
 		return 0;
 	}
@@ -158,31 +185,96 @@ static size_t xlat_packet(void *instance, REQUEST *request,
 	 */
 	da = dict_attrbyname(fmt);
 	if (!da) {
-		size_t count;
-		const char *p = strchr(fmt, '[');
+		int do_number = FALSE;
+		int do_array = FALSE;
+		int do_count = FALSE;
+		int do_all = FALSE;
+		int tag = 0;
+		size_t count = 0, total;
+		char *p;
 		char buffer[256];
 
-		if (!p) return 0;
 		if (strlen(fmt) > sizeof(buffer)) return 0;
 
-		strlcpy(buffer, fmt, p - fmt + 1);
+		strlcpy(buffer, fmt, sizeof(buffer));
+
+		/*
+		 *	%{Attribute-name#} - print integer version of it.
+		 */
+		p = buffer + strlen(buffer) - 1;
+		if (*p == '#') {
+			*p = '\0';
+			do_number = TRUE;
+		}
+
+		/*
+		 *	%{Attribute-Name:tag} - get the name with the specified
+		 *	value of the tag.
+		 */
+		p = strchr(buffer, ':');
+		if (p && (p[1] != '-')) {
+			tag = atoi(p + 1);
+			*p = '\0';
+			p++;
+
+		} else {
+			/*
+			 *	Allow %{Attribute-Name:tag[...]}
+			 */
+			p = buffer;
+		}
+
+		/*
+		 *	%{Attribute-Name[...] does more stuff
+		 */
+		p = strchr(p, '[');
+		if (p) {
+			*p = '\0';
+			do_array = TRUE;
+			if (p[1] == '#') {
+				do_count = TRUE;
+			} else if (p[1] == '*') {
+				do_all = TRUE;
+			} else {
+				count = atoi(p + 1);
+				p += 1 + strspn(p + 1, "0123456789");
+				if (*p != ']') {
+					RDEBUG2("xlat: Invalid array reference in string at %s %s",
+						fmt, p);
+					return 0;
+				}
+			}
+		}
+
+		/*
+		 *	We COULD argue about %{Attribute-Name[#]#} etc.
+		 *	But that looks like more work than it's worth.
+		 */
 
 		da = dict_attrbyname(buffer);
 		if (!da) return 0;
 
 		/*
-		 *	%{Attribute-Name[#]} returns the count of
-		 *	attributes of that name in the list.
+		 *	No array, print the tagged attribute.
 		 */
-		if ((p[1] == '#') && (p[2] == ']')) {
-			count = 0;
+		if (!do_array) {
+			vp = pairfind_tag(vps, da->attr, tag);
+			goto just_print;
+		}
 
-			for (vp = pairfind(vps, da->attr);
+		total = 0;
+
+		/*
+		 *	Array[#] - return the total
+		 */
+		if (do_count) {
+			for (vp = pairfind_tag(vps, da->attr, tag);
 			     vp != NULL;
-			     vp = pairfind(vp->next, da->attr)) {
-				count++;
+			     vp = pairfind_tag(vp->next, da->attr, tag)) {
+				total++;
 			}
-			snprintf(out, outlen, "%d", count);
+
+			snprintf(out, outlen, "%d", (int) total);
 			return strlen(out);
 		}
 
@@ -190,12 +282,10 @@ static size_t xlat_packet(void *instance, REQUEST *request,
 		 *	%{Attribute-Name[*]} returns ALL of the
 		 *	the attributes, separated by a newline.
 		 */
-		if ((p[1] == '*') && (p[2] == ']')) {
-			int total = 0;
-
-			for (vp = pairfind(vps, da->attr);
+		if (do_all) {
+			for (vp = pairfind_tag(vps, da->attr, tag);
 			     vp != NULL;
-			     vp = pairfind(vp->next, da->attr)) {
+			     vp = pairfind_tag(vp->next, da->attr, tag)) {
 				count = valuepair2str(out, outlen - 1, vp, da->type, func);
 				rad_assert(count <= outlen);
 				total += count + 1;
@@ -204,39 +294,46 @@ static size_t xlat_packet(void *instance, REQUEST *request,
 
 				*(out++) = '\n';
 
-				if (outlen == 0) break;
+				if (outlen <= 1) break;
 			}
 
+			*out = '\0';
 			return total;
 		}
 
-		count = atoi(p + 1);
-
-		/*
-		 *	Skip the numbers.
-		 */
-		p += 1 + strspn(p + 1, "0123456789");
-		if (*p != ']') {
-			RDEBUG2("xlat: Invalid array reference in string at %s %s",
-			       fmt, p);
-			return 0;
-		}
-
 		/*
 		 *	Find the N'th value.
 		 */
-		for (vp = pairfind(vps, da->attr);
+		for (vp = pairfind_tag(vps, da->attr, tag);
 		     vp != NULL;
-		     vp = pairfind(vp->next, da->attr)) {
-			if (count == 0) break;
-			count--;
+		     vp = pairfind_tag(vp->next, da->attr, tag)) {
+			if (total == count) break;
+			total++;
+			if (total > count) {
+				vp = NULL;
+				break;
+			}
 		}
 
 		/*
 		 *	Non-existent array reference.
 		 */
+	just_print:
 		if (!vp) return 0;
 
+		if (do_number) {
+			if ((vp->type != PW_TYPE_IPADDR) &&
+			    (vp->type != PW_TYPE_INTEGER) &&
+			    (vp->type != PW_TYPE_SHORT) &&
+			    (vp->type != PW_TYPE_BYTE) &&
+			    (vp->type != PW_TYPE_DATE)) {
+				*out = '\0';
+				return 0;
+			}
+			
+			return snprintf(out, outlen, "%u", vp->vp_integer);
+		}
+
 		return valuepair2str(out, outlen, vp, da->type, func);
 	}
 
@@ -382,6 +479,187 @@ static size_t xlat_packet(void *instance, REQUEST *request,
 	return valuepair2str(out, outlen, vp, da->type, func);
 }
 
+/*
+ *	Print data as integer, not as VALUE.
+ */
+static size_t xlat_integer(UNUSED void *instance, REQUEST *request,
+			   char *fmt, char *out, size_t outlen,
+			   UNUSED RADIUS_ESCAPE_STRING func)
+{
+	VALUE_PAIR *vp;
+
+	while (isspace((int) *fmt)) fmt++;
+
+	if (!radius_get_vp(request, fmt, &vp) || !vp) {
+		*out = '\0';
+		return 0;
+	}
+
+	if ((vp->type != PW_TYPE_IPADDR) &&
+	    (vp->type != PW_TYPE_INTEGER) &&
+	    (vp->type != PW_TYPE_SHORT) &&
+	    (vp->type != PW_TYPE_BYTE) &&
+	    (vp->type != PW_TYPE_DATE)) {
+		*out = '\0';
+		return 0;
+	}
+
+	return snprintf(out, outlen, "%u", vp->vp_integer);
+}
+
+/*
+ *	Print data as string, if possible.
+ */
+static size_t xlat_string(UNUSED void *instance, REQUEST *request,
+			  char *fmt, char *out, size_t outlen,
+			  UNUSED RADIUS_ESCAPE_STRING func)
+{
+	int len;
+	VALUE_PAIR *vp;
+
+	while (isspace((int) *fmt)) fmt++;
+
+	if (outlen < 3) {
+	nothing:
+		*out = '\0';
+		return 0;
+	}
+
+	if (!radius_get_vp(request, fmt, &vp)) goto nothing;
+
+	if (!vp) goto nothing;
+
+	if (vp->type != PW_TYPE_OCTETS) goto nothing;
+
+	len = fr_print_string(vp->vp_strvalue, vp->length, out, outlen);
+	out[len] = '\0';
+
+	return len;
+}
+
+/**
+ * @brief Print data as hex, not as VALUE.
+ */
+static size_t xlat_hex(UNUSED void *instance, REQUEST *request,
+		       char *fmt, char *out, size_t outlen,
+		       UNUSED RADIUS_ESCAPE_STRING func)
+{
+	size_t i;
+	VALUE_PAIR *vp;
+	uint8_t	buffer[MAX_STRING_LEN];
+	ssize_t	ret;
+	size_t	len;
+
+	while (isspace((int) *fmt)) fmt++;
+
+	if (!radius_get_vp(request, fmt, &vp) || !vp) {
+		*out = '\0';
+		return 0;
+	}
+	
+	ret = rad_vp2data(vp, buffer, sizeof(buffer));
+	len = (size_t) ret;
+	
+	/*
+	 *	Don't truncate the data.
+	 */
+	if ((ret < 0 ) || (outlen < (len * 2))) {
+		*out = 0;
+		return 0;
+	}
+
+	for (i = 0; i < len; i++) {
+		snprintf(out + 2*i, 3, "%02x", buffer[i]);
+	}
+
+	return len * 2;
+}
+
+/**
+ * @brief Print data as base64, not as VALUE
+ */
+static size_t xlat_base64(UNUSED void *instance, REQUEST *request,
+			  char *fmt, char *out, size_t outlen,
+			  UNUSED RADIUS_ESCAPE_STRING func)
+{
+	VALUE_PAIR *vp;
+	uint8_t buffer[MAX_STRING_LEN];
+	ssize_t	ret;
+	size_t	len;
+	size_t	enc;
+	
+	while (isspace((int) *fmt)) fmt++;
+
+	if (!radius_get_vp(request, fmt, &vp) || !vp) {
+		*out = '\0';
+		return 0;
+	}
+	
+	ret = rad_vp2data(vp, buffer, sizeof(buffer));
+	if (ret < 0) {
+		*out = 0;
+		return 0;
+	}
+	
+	len = (size_t) ret;
+	
+	enc = FR_BASE64_ENC_LENGTH(len);
+	
+	/*
+	 *	Don't truncate the data.
+	 */
+	if (outlen < (enc + 1)) {
+		*out = 0;
+		return 0;
+	}
+	
+	fr_base64_encode(buffer, len, out, outlen);
+
+	return enc;
+}
+
+#ifdef WITH_DHCP
+static size_t xlat_dhcp_options(UNUSED void *instance, REQUEST *request,
+			       char *fmt, char *out, size_t outlen,
+			       UNUSED RADIUS_ESCAPE_STRING func)
+{
+	VALUE_PAIR *vp, *head = NULL, *next;
+	int decoded = 0;
+
+	while (isspace((int) *fmt)) fmt++;
+
+	if (!radius_get_vp(request, fmt, &vp) || !vp) {
+		*out = '\0';
+		
+		return 0;
+	}
+	
+	if ((fr_dhcp_decode_options(vp->vp_octets, vp->length, &head) < 0) ||
+	    (head == NULL)) {
+		RDEBUG("WARNING: DHCP option decoding failed");
+		goto fail;
+	}
+
+	next = head;
+	
+	do {
+		next = next->next;
+		decoded++;
+	} while (next);
+	
+	pairmove(&(request->packet->vps), &head);
+	
+	/* Free any unmoved pairs */
+	pairfree(&head);
+
+	fail:
+	
+	snprintf(out, outlen, "%i", decoded);
+		  	
+	return strlen(out);
+}
+#endif
+
 #ifdef HAVE_REGEX_H
 /*
  *	Pull %{0} to %{8} out of the packet.
@@ -413,17 +691,29 @@ static size_t xlat_regex(void *instance, REQUEST *request,
 #endif				/* HAVE_REGEX_H */
 
 
-/*
- *	Change the debugging level.
+/**
+ * @brief Dynamically change the debugging level for the current request
+ *
+ * Example %{debug:3}
  */
 static size_t xlat_debug(UNUSED void *instance, REQUEST *request,
 			  char *fmt, char *out, size_t outlen,
 			  UNUSED RADIUS_ESCAPE_STRING func)
 {
 	int level = 0;
+	
+	/* 
+	 *  Expand to previous (or current) level
+	 */
+	snprintf(out, outlen, "%d", request->options & RAD_REQUEST_OPTION_DEBUG4);
 
-	if (*fmt) level = atoi(fmt);
-
+	/* 
+	 *  Assume we just want to get the current value and NOT set it to 0
+	 */
+	if (!*fmt)
+		goto done;
+		
+	level = atoi(fmt);
 	if (level == 0) {
 		request->options = RAD_REQUEST_OPTION_NONE;
 		request->radlog = NULL;
@@ -433,12 +723,11 @@ static size_t xlat_debug(UNUSED void *instance, REQUEST *request,
 		request->options = level;
 		request->radlog = radlog_request;
 	}
-
-	snprintf(out, outlen, "%d", level);
+	
+	done:
 	return strlen(out);
 }
 
-
 /*
  *	Compare two xlat_t structs, based ONLY on the module name.
  */
@@ -461,14 +750,6 @@ static xlat_t *xlat_find(const char *module)
 {
 	xlat_t my_xlat;
 
-	/*
-	 *	Look for dictionary attributes first.
-	 */
-	if ((dict_attrbyname(module) != NULL) ||
-	    (strchr(module, '[') != NULL)) {
-		module = "request";
-	}
-
 	strlcpy(my_xlat.module, module, sizeof(my_xlat.module));
 	my_xlat.length = strlen(my_xlat.module);
 
@@ -484,7 +765,7 @@ int xlat_register(const char *module, RAD_XLAT_FUNC func, void *instance)
 	xlat_t	*c;
 	xlat_t	my_xlat;
 
-	if ((module == NULL) || (strlen(module) == 0)) {
+	if (!module || !*module) {
 		DEBUG("xlat_register: Invalid module name");
 		return -1;
 	}
@@ -524,6 +805,33 @@ int xlat_register(const char *module, RAD_XLAT_FUNC func, void *instance)
 		c = xlat_find("control");
 		rad_assert(c != NULL);
 		c->internal = TRUE;
+		
+		xlat_register("hex", xlat_hex, "");
+		c = xlat_find("hex");
+		rad_assert(c != NULL);
+		c->internal = TRUE;
+
+		xlat_register("integer", xlat_integer, "");
+		c = xlat_find("integer");
+		rad_assert(c != NULL);
+		c->internal = TRUE;
+		
+		xlat_register("base64", xlat_base64, "");
+		c = xlat_find("base64");
+		rad_assert(c != NULL);
+		c->internal = TRUE;
+
+		xlat_register("string", xlat_string, "");
+		c = xlat_find("string");
+		rad_assert(c != NULL);
+		c->internal = TRUE;
+
+#ifdef WITH_DHCP
+		xlat_register("dhcp_options", xlat_dhcp_options, "");
+		c = xlat_find("dhcp_options");
+		rad_assert(c != NULL);
+		c->internal = TRUE;
+#endif
 
 #ifdef HAVE_REGEX_H
 		/*
@@ -585,9 +893,9 @@ int xlat_register(const char *module, RAD_XLAT_FUNC func, void *instance)
  *	We can only have one function to call per name, so the
  *	passing of "func" here is extraneous.
  */
-void xlat_unregister(const char *module, RAD_XLAT_FUNC func)
+void xlat_unregister(const char *module, RAD_XLAT_FUNC func, void *instance)
 {
-	rbnode_t	*node;
+	xlat_t	*c;
 	xlat_t		my_xlat;
 
 	func = func;		/* -Wunused */
@@ -597,10 +905,12 @@ void xlat_unregister(const char *module, RAD_XLAT_FUNC func)
 	strlcpy(my_xlat.module, module, sizeof(my_xlat.module));
 	my_xlat.length = strlen(my_xlat.module);
 
-	node = rbtree_find(xlat_root, &my_xlat);
-	if (!node) return;
+	c = rbtree_finddata(xlat_root, &my_xlat);
+	if (!c) return;
 
-	rbtree_delete(xlat_root, node);
+	if (c->instance != instance) return;
+
+	rbtree_deletebydata(xlat_root, c);
 }
 
 /*
@@ -616,34 +926,39 @@ void xlat_free(void)
 /*
  *	Decode an attribute name into a string.
  */
-static void decode_attribute(const char **from, char **to, int freespace,
-			     int *open_p, REQUEST *request,
+static int decode_attribute(const char **from, char **to, int freespace,
+			     REQUEST *request,
 			     RADIUS_ESCAPE_STRING func)
 {
 	int	do_length = 0;
-	char	xlat_name[128];
-	char	*xlat_string = NULL; /* can be large */
-	int	free_xlat_string = FALSE;
-	const char *p;
-	char *q, *pa;
-	int found=0, retlen=0;
-	int openbraces = *open_p;
+	char	*module_name, *xlat_str;
+	char *p, *q, *l, *next = NULL;
+	int retlen=0;
 	const xlat_t *c;
-	int spaces = FALSE;
+	int varlen;
+	char buffer[8192];
 
-	p = *from;
 	q = *to;
-	pa = &xlat_name[0];
 
 	*q = '\0';
 
 	/*
-	 * Skip the '{' at the front of 'p'
-	 * Increment open braces
+	 *	Copy the input string to an intermediate buffer where
+	 *	we can mangle it.
 	 */
-	p++;
-	openbraces++;
+	varlen = rad_copy_variable(buffer, *from);
+	if (varlen < 0) {
+		RDEBUG2("ERROR: Badly formatted variable: %s", *from);
+		return -1;
+	}
+	*from += varlen;
 
+	/*
+	 *	Kill the %{} around the data we are looking for.
+	 */
+	p = buffer;
+	p[varlen - 1] = '\0';	/*  */
+	p += 2;
 	if (*p == '#') {
 		p++;
 		do_length = 1;
@@ -655,284 +970,194 @@ static void decode_attribute(const char **from, char **to, int freespace,
 	 *	Did I mention that this parser is garbage?
 	 */
 	if ((p[0] == '%') && (p[1] == '{')) {
-		/*
-		 *	This is really bad, but it works.
-		 */
 		int len1, len2;
-		size_t mylen = strlen(p);
-		char *first = rad_malloc(mylen + 1);
-		char *second = rad_malloc(mylen + 1);
 		int expand2 = FALSE;
 
-		len1 = rad_copy_variable(first, p);
+		/*
+		 *	'p' is after the start of 'buffer', so we can
+		 *	safely do this.
+		 */
+		len1 = rad_copy_variable(buffer, p);
 		if (len1 < 0) {
-			RDEBUG2("Badly formatted variable: %s", p);
-			goto free_and_done;
+			RDEBUG2("ERROR: Badly formatted variable: %s", p);
+			return -1;
 		}
 
+		/*
+		 *	They did %{%{foo}}, which is stupid, but allowed.
+		 */
+		if (!p[len1]) {
+			RDEBUG2("Improperly nested variable; %%{%s}", p);
+			return -1;
+		}
+
+		/*
+		 *	It SHOULD be %{%{foo}:-%{bar}}.  If not, it's
+		 *	an error.
+		 */
 		if ((p[len1] != ':') || (p[len1 + 1] != '-')) {
 			RDEBUG2("No trailing :- after variable at %s", p);
-			goto free_and_done;
+			return -1;
 		}
 
+		/*
+		 *	Parse the second bit.  The second bit can be
+		 *	either %{foo}, or a string "foo", or a string
+		 *	'foo', or just a bare word: foo
+		 */
 		p += len1 + 2;
+		l = buffer + len1 + 1;
 
 		if ((p[0] == '%') && (p[1] == '{')) {
-			len2 = rad_copy_variable(second, p);
+			len2 = rad_copy_variable(l, p);
 
-			expand2 = TRUE;
 			if (len2 < 0) {
-				RDEBUG2("Invalid text after :- at %s", p);
-				goto free_and_done;
+				RDEBUG2("ERROR: Invalid text after :- at %s", p);
+				return -1;
 			}
 			p += len2;
+			expand2 = TRUE;
 
 		} else if ((p[0] == '"') || p[0] == '\'') {
-			getstring(&p, second, mylen);
+			getstring(&p, l, strlen(l));
 
 		} else {
-			char *s = second;
-
-			while (*p && (*p != '}')) {
-				*(s++) = *(p++);
-			}
-			*s = '\0';
+			l = p;
 		}
 
-		if (*p != '}') {
-			RDEBUG2("Failed to find trailing '}' in string");
-			goto free_and_done;
+		/*
+		 *	Expand the first one.  If we did, exit the
+		 *	conditional.
+		 */
+		retlen = radius_xlat(q, freespace, buffer, request, func);
+		if (retlen) {
+			q += retlen;
+			goto done;
 		}
 
-		mylen = radius_xlat(q, freespace, first, request, func);
-		if (mylen) {
-			q += mylen;
-			goto free_and_done;
-		}
-
-		if (!expand2) {
-			strlcpy(q, second, freespace);
-			q += strlen(q);
-		} else {
-			mylen = radius_xlat(q, freespace, second,
+		RDEBUG2("\t... expanding second conditional");
+		/*
+		 *	Expand / copy the second string if required.
+		 */
+		if (expand2) {
+			retlen = radius_xlat(q, freespace, l,
 					    request, func);
-			if (mylen) {
-				q += mylen;
-				goto free_and_done;
+			if (retlen) {
+				q += retlen;
 			}
+		} else {
+			strlcpy(q, l, freespace);
+			q += strlen(q);
 		}
 
 		/*
 		 *	Else the output is an empty string.
 		 */
-	free_and_done:
-		free(first);
-		free(second);
 		goto done;
 	}
 
+	/*
+	 *	See if we're supposed to expand a module name.
+	 */
+	module_name = NULL;
+	for (l = p; *l != '\0'; l++) {
+		if (*l == '\\') {
+			l++;
+			continue;
+		}
+
+		if (*l == ':') {
+			if (l[1] == '-') {
+				RDEBUG2("WARNING: Deprecated conditional expansion \":-\".  See \"man unlang\" for details");
+				module_name = internal_xlat[1];
+				xlat_str = p;
+				*l = '\0';
+				next = l + 2;
+				goto do_xlat;
+			}
+
+			module_name = p; /* start of name */
+			*l = '\0';
+
+			/*
+			 *	%{Tunnel-Password:0}
+			 *
+			 *		OR
+			 *
+			 *	%{expr:0 + 1}
+			 */
+			if (isdigit(l[1]) &&
+			    (dict_attrbyname(module_name) != NULL)) {
+				module_name = NULL;
+				*l = ':';
+				break;
+			}
+
+			p = l + 1;
+			break;
+		}
+
+		/*
+		 *	Module names can't have spaces.
+		 */
+		if ((*l == ' ') || (*l == '\t')) break;
+	}
 
 	/*
-	 *	First, copy the xlat key name to one buffer
+	 *	%{name} is a simple attribute reference,
+	 *	or regex reference.
 	 */
-	while (*p && (*p != '}') && (*p != ':')) {
-		*pa++ = *p++;
-
-		if (pa >= (xlat_name + sizeof(xlat_name) - 1)) {
-			/*
-			 *	Skip to the end of the input
-			 */
-			p += strlen(p);
-			RDEBUG("xlat: Module name is too long in string %%%s",
-			      *from);
-			goto done;
-		}
-	}
-	*pa = '\0';
-
-	if (!*p) {
-		RDEBUG("xlat: Invalid syntax in %s", *from);
-
-		/*
-		 *	%{name} is a simple attribute reference,
-		 *	or regex reference.
-		 */
-	} else if (*p == '}') {
-		openbraces--;
-		rad_assert(openbraces == *open_p);
-
-		p++;
-		xlat_string = xlat_name;
-		goto do_xlat;
-
-	} else if ((p[0] == ':') && (p[1] == '-')) { /* handle ':- */
-		RDEBUG2("WARNING: Deprecated conditional expansion \":-\".  See \"man unlang\" for details");
-		p += 2;
-		xlat_string = xlat_name;
-		goto do_xlat;
-
-	} else {      /* module name, followed by per-module string */
-		int stop = 0;
-		int delimitbrace = *open_p;
-
-		rad_assert(*p == ':');
-		p++;			/* skip the ':' */
-
-		/*
-		 *  If there's a brace immediately following the colon,
-		 *  then we've chosen to delimite the per-module string,
-		 *  so keep track of that.
-		 */
-		if (*p == '{') {
-			delimitbrace = openbraces;
-			openbraces++;
-			p++;
-		}
-
-		xlat_string = rad_malloc(strlen(p) + 1); /* always returns */
-		free_xlat_string = TRUE;
-		pa = xlat_string;
-
-		/*
-		 *  Copy over the rest of the string, which is per-module
-		 *  data.
-		 */
-		while (*p && !stop) {
-			switch(*p) {
-
-				/*
-				 *	What the heck is this supposed
-				 *	to be doing?
-				 */
-			case '\\':
-				p++; /* skip it */
-				*pa++ = *p++;
-				break;
-
-			case ':':
-				if (!spaces && p[1] == '-') {
-					p += 2;
-					stop = 1;
-					break;
-				}
-				*pa++ = *p++;
-				break;
-
-				/*
-				 *	This is pretty hokey...  we
-				 *	should use the functions in
-				 *	util.c
-				 */
-			case '{':
-				openbraces++;
-				*pa++ = *p++;
-				break;
-
-			case '}':
-				openbraces--;
-				if (openbraces == delimitbrace) {
-					p++;
-					stop=1;
-				} else {
-					*pa++ = *p++;
-				}
-				break;
-
-			case ' ':
-			case '\t':
-				spaces = TRUE;
-				/* FALL-THROUGH */
-
-			default:
-				*pa++ = *p++;
-				break;
-			}
-		}
-
-		*pa = '\0';
-
-		/*
-		 *	Now check to see if we're at the end of the string
-		 *	we were sent.  If we're not, check for :-
-		 */
-		if (openbraces == delimitbrace) {
-			if (p[0] == ':' && p[1] == '-') {
-				p += 2;
-			}
-		}
-
-		/*
-		 *	Look up almost everything in the new tree of xlat
-		 *	functions.  This makes it a little quicker...
-		 */
-	do_xlat:
-		if ((c = xlat_find(xlat_name)) != NULL) {
-			if (!c->internal) RDEBUG3("radius_xlat: Running registered xlat function of module %s for string \'%s\'",
-						c->module, xlat_string);
-			retlen = c->do_xlat(c->instance, request, xlat_string,
-					    q, freespace, func);
-			/* If retlen is 0, treat it as not found */
-			if (retlen > 0) found = 1;
-
-#ifndef NRDEBUG
+	if (!module_name) {
+		if (isdigit(*p) && !p[1]) { /* regex 0..8 */
+			module_name = xlat_str = p;
 		} else {
-			/*
-			 *	No attribute by that name, return an error.
-			 */
-			RDEBUG2("WARNING: Unknown module \"%s\" in string expansion \"%%%s\"", xlat_name, *from);
-#endif
+			module_name = internal_xlat[1];
+			xlat_str = p;
 		}
+		goto do_xlat;
 	}
 
 	/*
-	 * Skip to last '}' if attr is found
-	 * The rest of the stuff within the braces is
-	 * useless if we found what we need
+	 *	FIXME: For backwards "WTF" compatibility, check for
+	 *	{...}, (after the :), and copy that, too.
 	 */
-	if (found) {
+
+	/* module name, followed by (possibly) per-module string */
+	xlat_str = p;
+	
+do_xlat:
+	c = xlat_find(module_name);
+	if (!c) {
+		if (module_name == internal_xlat[1]) {
+			RDEBUG2("WARNING: Unknown Attribute \"%s\" in string expansion \"%%%s\"", module_name, *from);
+		} else {
+			RDEBUG2("WARNING: Unknown module \"%s\" in string expansion \"%%%s\"", module_name, *from);
+		}
+		return -1;
+	}
+
+	if (!c->internal) RDEBUG3("radius_xlat: Running registered xlat function of module %s for string \'%s\'",
+				  c->module, xlat_str);
+	retlen = c->do_xlat(c->instance, request, xlat_str,
+			    q, freespace, func);
+	if (retlen > 0) {
 		if (do_length) {
 			snprintf(q, freespace, "%d", retlen);
 			retlen = strlen(q);
 		}
-
-		q += retlen;
-
-		while((*p != '\0') && (openbraces > *open_p)) {
-			/*
-			 *	Handle escapes outside of the loop.
-			 */
-			if (*p == '\\') {
-				p++;
-				if (!*p) break;
-				p++; /* get & ignore next character */
-				continue;
-			}
-
-			switch (*p) {
-			default:
-				break;
-
-				/*
-				 *  Bare brace
-				 */
-			case '{':
-				openbraces++;
-				break;
-
-			case '}':
-				openbraces--;
-				break;
-			}
-			p++;	/* skip the character */
-		}
+		
+	} else if (next) {
+		/*
+		 *	Expand the second bit.
+		 */
+		RDEBUG2("\t... expanding second conditional");
+		retlen = radius_xlat(q, freespace, next, request, func);
 	}
+	q += retlen;
 
-	done:
-	if (free_xlat_string) free(xlat_string);
-
-	*open_p = openbraces;
-	*from = p;
+done:
 	*to = q;
+	return 0;
 }
 
 /*
@@ -944,7 +1169,7 @@ static size_t xlat_copy(char *out, size_t outlen, const char *in)
 {
 	int freespace = outlen;
 
-	rad_assert(outlen > 0);
+	if (outlen < 1) return 0;
 
 	while ((*in) && (freespace > 1)) {
 		/*
@@ -1044,7 +1269,8 @@ int radius_xlat(char *out, int outlen, const char *fmt,
 
 		} else if (c == '%') switch(*p) {
 			case '{':
-				decode_attribute(&p, &q, freespace, &openbraces, request, func);
+				p--;
+				if (decode_attribute(&p, &q, freespace, request, func) < 0) return 0;
 				break;
 
 			case '%':
@@ -1077,7 +1303,7 @@ int radius_xlat(char *out, int outlen, const char *fmt,
 				break;
 			case 'l': /* request timestamp */
 				snprintf(tmpdt, sizeof(tmpdt), "%lu",
-					 (unsigned long) request->received.tv_sec);
+					 (unsigned long) request->timestamp);
 				strlcpy(q,tmpdt,freespace);
 				q += strlen(q);
 				p++;
@@ -1104,7 +1330,14 @@ int radius_xlat(char *out, int outlen, const char *fmt,
 				p++;
 				break;
 			case 't': /* request timestamp */
-				CTIME_R(&request->timestamp, tmpdt, sizeof(tmpdt));
+#ifdef HAVE_GMTIME_R
+				if (log_dates_utc) {
+					struct tm utc;
+					gmtime_r(&request->timestamp, &utc);
+					asctime_r(&utc, tmpdt);
+				} else
+#endif
+					CTIME_R(&request->timestamp, tmpdt, sizeof(tmpdt));
 				nl = strchr(tmpdt, '\n');
 				if (nl) *nl = '\0';
 				strlcpy(q, tmpdt, freespace);
@@ -1143,11 +1376,26 @@ int radius_xlat(char *out, int outlen, const char *fmt,
 				}
 				p++;
 				break;
+			case 'I': /* Request ID */
+				snprintf(tmpdt, sizeof(tmpdt), "%i", request->packet->id);
+				strlcpy(q, tmpdt, freespace);
+				q += strlen(q);
+				p++;
+				break;
 			case 'L': /* radlog_dir */
 				strlcpy(q,radlog_dir,freespace);
 				q += strlen(q);
 				p++;
 				break;
+			case 'G': /* request minute */
+				TM = localtime_r(&request->timestamp, &s_TM);
+				len = strftime(tmpdt, sizeof(tmpdt), "%M", TM);
+				if (len > 0) {
+					strlcpy(q, tmpdt, freespace);
+					q += strlen(q);
+				}
+				p++;
+				break;
 			case 'M': /* MTU */
 				q += valuepair2str(q,freespace,pairfind(request->reply->vps,PW_FRAMED_MTU),PW_TYPE_INTEGER, func);
 				p++;
