@@ -1,7 +1,7 @@
 /*
  * eap_tls.c
  *
- * Version:     $Id$
+ * Version:     $Id: 3e2efbf244262334dcfb71f8fcff353a35b593c0 $
  *
  *   This program is free software; you can redistribute it and/or modify
  *   it under the terms of the GNU General Public License as published by
@@ -39,7 +39,7 @@
  */
 
 #include <freeradius-devel/ident.h>
-RCSID("$Id$")
+RCSID("$Id: 3e2efbf244262334dcfb71f8fcff353a35b593c0 $")
 
 #include <freeradius-devel/autoconf.h>
 #include <assert.h>
@@ -109,6 +109,7 @@ int eaptls_success(EAP_HANDLER *handler, int peap_flag)
 	REQUEST *request = handler->request;
 	tls_session_t *tls_session = handler->opaque;
 
+	handler->finished = TRUE;
 	reply.code = EAPTLS_SUCCESS;
 	reply.length = TLS_HEADER_LEN;
 	reply.flags = peap_flag;
@@ -148,11 +149,18 @@ int eaptls_success(EAP_HANDLER *handler, int peap_flag)
 		RDEBUG2("Saving response in the cache");
 		
 		vp = paircopy2(request->reply->vps, PW_USER_NAME);
-		pairadd(&vps, vp);
+		if (vp) pairadd(&vps, vp);
 		
 		vp = paircopy2(request->packet->vps, PW_STRIPPED_USER_NAME);
-		pairadd(&vps, vp);
-		
+		if (vp) pairadd(&vps, vp);
+
+		vp = paircopy2(request->reply->vps, PW_CACHED_SESSION_POLICY);
+		if (vp) pairadd(&vps, vp);
+
+		if (handler->certs) {
+			pairadd(&vps, paircopy(handler->certs));
+		}
+
 		if (vps) {
 			SSL_SESSION_set_ex_data(tls_session->ssl->session,
 						eaptls_session_idx, vps);
@@ -168,15 +176,29 @@ int eaptls_success(EAP_HANDLER *handler, int peap_flag)
 		 */
 	} else {
 	       
-		vp = SSL_SESSION_get_ex_data(tls_session->ssl->session,
+		vps = SSL_SESSION_get_ex_data(tls_session->ssl->session,
 					     eaptls_session_idx);
-		if (!vp) {
+		if (!vps) {
 			RDEBUG("WARNING: No information in cached session!");
 			return eaptls_fail(handler, peap_flag);
 		} else {
-			RDEBUG("Adding cached attributes to the reply:");
-			debug_pair_list(vp);
-			pairadd(&request->reply->vps, paircopy(vp));
+			RDEBUG("Adding cached attributes:");
+			debug_pair_list(vps);
+
+			for (vp = vps; vp != NULL; vp = vp->next) {
+				/*
+				 *	TLS-* attrs get added back to
+				 *	the request list.
+				 */
+				if ((vp->attribute >= 1910) &&
+				    (vp->attribute < 1929)) {
+					pairadd(&request->packet->vps,
+						paircopyvp(vp));
+				} else {
+					pairadd(&request->reply->vps,
+						paircopyvp(vp));
+				}
+			}
 
 			/*
 			 *	Mark the request as resumed.
@@ -201,6 +223,8 @@ int eaptls_success(EAP_HANDLER *handler, int peap_flag)
 		RDEBUG("WARNING: Not adding MPPE keys because there is no PRF label");
 	}
 
+	eaptls_gen_eap_key(tls_session->ssl,
+			   handler->eap_type, &handler->request->reply->vps);
 	return 1;
 }
 
@@ -209,6 +233,7 @@ int eaptls_fail(EAP_HANDLER *handler, int peap_flag)
 	EAPTLS_PACKET	reply;
 	tls_session_t *tls_session = handler->opaque;
 
+	handler->finished = TRUE;
 	reply.code = EAPTLS_FAIL;
 	reply.length = TLS_HEADER_LEN;
 	reply.flags = peap_flag;
@@ -760,7 +785,7 @@ static eaptls_status_t eaptls_operation(eaptls_status_t status,
 	 *	If more info
 	 *	is required then send another request.
 	 */
-	if (!tls_handshake_recv(tls_session)) {
+	if (!tls_handshake_recv(handler->request, tls_session)) {
 		DEBUG2("TLS receive handshake failed during operation");
 		eaptls_fail(handler, tls_session->peap_flag);
 		return EAPTLS_FAIL;
@@ -835,6 +860,8 @@ eaptls_status_t eaptls_process(EAP_HANDLER *handler)
 	REQUEST *request = handler->request;
 
 	RDEBUG2("processing EAP-TLS");
+	if (handler->certs) pairadd(&request->packet->vps,
+				    paircopy(handler->certs));
 
 	/* This case is when SSL generates Alert then we
 	 * send that alert to the client and then send the EAP-Failure
