@@ -1,7 +1,7 @@
 /*
  * rlm_eap.c  contains handles that are called from modules.
  *
- * Version:     $Id: rlm_eap.c,v 1.55 2008/02/10 09:18:10 aland Exp $
+ * Version:     $Id: a4a01f5a7166e09156a8bef3fe56a2f2a223bf26 $
  *
  *   This program is free software; you can redistribute it and/or modify
  *   it under the terms of the GNU General Public License as published by
@@ -23,7 +23,7 @@
  */
 
 #include <freeradius-devel/ident.h>
-RCSID("$Id: rlm_eap.c,v 1.55 2008/02/10 09:18:10 aland Exp $")
+RCSID("$Id: a4a01f5a7166e09156a8bef3fe56a2f2a223bf26 $")
 
 #include <freeradius-devel/radiusd.h>
 #include <freeradius-devel/modules.h>
@@ -39,6 +39,8 @@ static const CONF_PARSER module_config[] = {
 	  offsetof(rlm_eap_t, ignore_unknown_eap_types), NULL, "no" },
 	{ "cisco_accounting_username_bug", PW_TYPE_BOOLEAN,
 	  offsetof(rlm_eap_t, cisco_accounting_username_bug), NULL, "no" },
+	{ "max_sessions", PW_TYPE_INTEGER,
+	  offsetof(rlm_eap_t, max_sessions), NULL, "2048"},
 
  	{ NULL, -1, 0, NULL, NULL }           /* end the list */
 };
@@ -53,7 +55,13 @@ static int eap_detach(void *instance)
 
 	inst = (rlm_eap_t *)instance;
 
+#ifdef HAVE_PTHREAD_H
+	pthread_mutex_destroy(&(inst->session_mutex));
+	if (inst->handler_tree) pthread_mutex_destroy(&(inst->handler_mutex));
+#endif
+
 	rbtree_free(inst->session_tree);
+	if (inst->handler_tree) rbtree_free(inst->handler_tree);
 	inst->session_tree = NULL;
 	eaplist_free(inst);
 
@@ -62,8 +70,6 @@ static int eap_detach(void *instance)
 		inst->types[i] = NULL;
 	}
 
-	pthread_mutex_destroy(&(inst->session_mutex));
-
 	free(inst);
 
 	return 0;
@@ -82,10 +88,31 @@ static int eap_handler_cmp(const void *a, const void *b)
 	if (one->eap_id < two->eap_id) return -1;
 	if (one->eap_id > two->eap_id) return +1;
 
-	rcode = fr_ipaddr_cmp(&one->src_ipaddr, &two->src_ipaddr);
+	rcode = memcmp(one->state, two->state, sizeof(one->state));
 	if (rcode != 0) return rcode;
 
-	return memcmp(one->state, two->state, sizeof(one->state));
+	/*
+	 *	As of 2.1.8, we don't key off of source IP.  This
+	 *	a NAS to send packets load-balanced (or fail-over)
+	 *	across multiple intermediate proxies, and still have
+	 *	EAP work.
+	 */
+	if (fr_ipaddr_cmp(&one->src_ipaddr, &two->src_ipaddr) != 0) {
+		DEBUG("WARNING: EAP packets are arriving from two different upstream servers.  Has there been a proxy fail-over?");
+	}
+
+	return 0;
+}
+
+
+/*
+ *	Compare two handler pointers
+ */
+static int eap_handler_ptr_cmp(const void *a, const void *b)
+{
+	if (a < b) return -1;
+	if (a > b) return +1;
+	return 0;
 }
 
 
@@ -118,6 +145,9 @@ static int eap_instantiate(CONF_SECTION *cs, void **instance)
 	fr_randinit(&inst->rand_pool, 1);
 	inst->rand_pool.randcnt = 0;
 
+	inst->xlat_name = cf_section_name2(cs);
+	if (!inst->xlat_name) inst->xlat_name = "EAP";
+
 	/* Load all the configured EAP-Types */
 	num_types = 0;
 	for(scs=cf_subsection_find_next(cs, NULL, NULL);
@@ -132,7 +162,7 @@ static int eap_instantiate(CONF_SECTION *cs, void **instance)
 
 		eap_type = eaptype_name2type(auth_type);
 		if (eap_type < 0) {
-			radlog(L_ERR|L_CONS, "rlm_eap: Unknown EAP type %s",
+			radlog(L_ERR, "rlm_eap: Unknown EAP type %s",
 			       auth_type);
 			eap_detach(inst);
 			return -1;
@@ -154,7 +184,7 @@ static int eap_instantiate(CONF_SECTION *cs, void **instance)
 		if ((eap_type == PW_EAP_TLS) ||
 		    (eap_type == PW_EAP_TTLS) ||
 		    (eap_type == PW_EAP_PEAP)) {
-			DEBUG2("rlm_eap: Ignoring EAP-Type/%s because we do not have OpenSSL support.", auth_type);
+			DEBUG2("Ignoring EAP-Type/%s because we do not have OpenSSL support.", auth_type);
 			continue;
 		}
 #endif
@@ -223,7 +253,30 @@ static int eap_instantiate(CONF_SECTION *cs, void **instance)
 		return -1;
 	}
 
-	pthread_mutex_init(&(inst->session_mutex), NULL);
+	if (fr_debug_flag) {
+		inst->handler_tree = rbtree_create(eap_handler_ptr_cmp, NULL, 0);
+		if (!inst->handler_tree) {
+			radlog(L_ERR|L_CONS, "rlm_eap: Cannot initialize tree");
+			eap_detach(inst);
+			return -1;
+		}
+
+#ifdef HAVE_PTHREAD_H
+		if (pthread_mutex_init(&(inst->handler_mutex), NULL) < 0) {
+			radlog(L_ERR|L_CONS, "rlm_eap: Failed initializing mutex: %s", strerror(errno));
+			eap_detach(inst);
+			return -1;
+		}
+#endif
+	}
+
+#ifdef HAVE_PTHREAD_H
+	if (pthread_mutex_init(&(inst->session_mutex), NULL) < 0) {
+		radlog(L_ERR|L_CONS, "rlm_eap: Failed initializing mutex: %s", strerror(errno));
+		eap_detach(inst);
+		return -1;
+	}
+#endif
 
 	*instance = inst;
 	return 0;
@@ -242,12 +295,17 @@ static int eap_authenticate(void *instance, REQUEST *request)
 
 	inst = (rlm_eap_t *) instance;
 
+	if (!pairfind(request->packet->vps, PW_EAP_MESSAGE)) {
+		RDEBUG("ERROR: You set 'Auth-Type = EAP' for a request that does not contain an EAP-Message attribute!");
+		return RLM_MODULE_INVALID;
+	}
+
 	/*
 	 *	Get the eap packet  to start with
 	 */
 	eap_packet = eap_vp2packet(request->packet->vps);
 	if (eap_packet == NULL) {
-		radlog(L_ERR, "rlm_eap: Malformed EAP Message");
+		radlog_request(L_ERR, 0, request, "Malformed EAP Message");
 		return RLM_MODULE_FAIL;
 	}
 
@@ -258,7 +316,7 @@ static int eap_authenticate(void *instance, REQUEST *request)
 	 */
 	handler = eap_handler(inst, &eap_packet, request);
 	if (handler == NULL) {
-		DEBUG2("  rlm_eap: Failed in handler");
+		RDEBUG2("Failed in handler");
 		return RLM_MODULE_INVALID;
 	}
 
@@ -273,31 +331,34 @@ static int eap_authenticate(void *instance, REQUEST *request)
 	 */
 	if (rcode == EAP_INVALID) {
 		eap_fail(handler);
-		eap_handler_free(handler);
-		DEBUG2("  rlm_eap: Failed in EAP select");
+		eap_handler_free(inst, handler);
+		RDEBUG2("Failed in EAP select");
 		return RLM_MODULE_INVALID;
 	}
 
+#ifdef WITH_PROXY
 	/*
 	 *	If we're doing horrible tunneling work, remember it.
 	 */
 	if ((request->options & RAD_REQUEST_OPTION_PROXY_EAP) != 0) {
-		DEBUG2("  Not-EAP proxy set.  Not composing EAP");
+		RDEBUG2("  Not-EAP proxy set.  Not composing EAP");
 		/*
 		 *	Add the handle to the proxied list, so that we
 		 *	can retrieve it in the post-proxy stage, and
 		 *	send a response.
 		 */
+		handler->inst_holder = inst;
 		rcode = request_data_add(request,
 					 inst, REQUEST_DATA_EAP_HANDLER,
 					 handler,
-					 (void *) eap_handler_free);
+					 (void *) eap_opaque_free);
 		rad_assert(rcode == 0);
 
 		return RLM_MODULE_HANDLED;
 	}
+#endif
 
-
+#ifdef WITH_PROXY
 	/*
 	 *	Maybe the request was marked to be proxied.  If so,
 	 *	proxy it.
@@ -312,10 +373,11 @@ static int eap_authenticate(void *instance, REQUEST *request)
 		 *	can retrieve it in the post-proxy stage, and
 		 *	send a response.
 		 */
+		handler->inst_holder = inst;
 		rcode = request_data_add(request,
 					 inst, REQUEST_DATA_EAP_HANDLER,
 					 handler,
-					 (void *) eap_handler_free);
+					 (void *) eap_opaque_free);
 		rad_assert(rcode == 0);
 
 		/*
@@ -340,9 +402,10 @@ static int eap_authenticate(void *instance, REQUEST *request)
 		 */
 		pairdelete(&request->proxy->vps, PW_FREERADIUS_PROXIED_TO);
 
-		DEBUG2("  Tunneled session will be proxied.  Not doing EAP.");
+		RDEBUG2("  Tunneled session will be proxied.  Not doing EAP.");
 		return RLM_MODULE_HANDLED;
 	}
+#endif
 
 	/*
 	 *	We are done, wrap the EAP-request in RADIUS to send
@@ -371,16 +434,26 @@ static int eap_authenticate(void *instance, REQUEST *request)
 	     (handler->eap_ds->request->code == PW_EAP_SUCCESS) &&
 	     (handler->eap_ds->request->type.type == 0))) {
 
+		/*
+		 *	Return FAIL if we can't remember the handler.
+		 *	This is actually disallowed by the
+		 *	specification, as unexpected FAILs could have
+		 *	been forged.  However, we want to signal to
+		 *	everyone else involved that we are
+		 *	intentionally failing the session, as opposed
+		 *	to accidentally failing it.
+		 */
 		if (!eaplist_add(inst, handler)) {
+			RDEBUG("Failed adding handler to the list");
 			eap_fail(handler);
-			eap_handler_free(handler);
+			eap_handler_free(inst, handler);
 			return RLM_MODULE_FAIL;
 		}
 
 	} else {
-		DEBUG2("  rlm_eap: Freeing handler");
+		RDEBUG2("Freeing handler");
 		/* handler is not required any more, free it now */
-		eap_handler_free(handler);
+		eap_handler_free(inst, handler);
 	}
 
 	/*
@@ -397,8 +470,11 @@ static int eap_authenticate(void *instance, REQUEST *request)
 		 */
 		vp = pairfind(request->reply->vps, PW_USER_NAME);
 		if (!vp) {
-			vp = pairmake("User-Name", request->username->vp_strvalue,
+			vp = pairmake("User-Name", "",
 				      T_OP_EQ);
+			strlcpy(vp->vp_strvalue, request->username->vp_strvalue,
+				sizeof(vp->vp_strvalue));
+			vp->length = request->username->length;
 			rad_assert(vp != NULL);
 			pairadd(&(request->reply->vps), vp);
 		}
@@ -430,12 +506,14 @@ static int eap_authorize(void *instance, REQUEST *request)
 
 	inst = (rlm_eap_t *)instance;
 
+#ifdef WITH_PROXY
 	/*
 	 *	We don't do authorization again, once we've seen the
 	 *	proxy reply (or the proxied packet)
 	 */
 	if (request->proxy != NULL)
                 return RLM_MODULE_NOOP;
+#endif
 
 	/*
 	 *	For EAP_START, send Access-Challenge with EAP Identity
@@ -473,11 +551,15 @@ static int eap_authorize(void *instance, REQUEST *request)
 	vp = pairfind(request->config_items, PW_AUTH_TYPE);
 	if ((!vp) ||
 	    (vp->vp_integer != PW_AUTHTYPE_REJECT)) {
-		vp = pairmake("Auth-Type", "EAP", T_OP_EQ);
+		vp = pairmake("Auth-Type", inst->xlat_name, T_OP_EQ);
 		if (!vp) {
+			RDEBUG2("Failed to create Auth-Type %s: %s\n",
+				inst->xlat_name, fr_strerror());
 			return RLM_MODULE_FAIL;
 		}
 		pairadd(&request->config_items, vp);
+	} else {
+		RDEBUG2("WARNING: Auth-Type already set.  Not setting to EAP");
 	}
 
 	if (status == EAP_OK) return RLM_MODULE_OK;
@@ -485,6 +567,8 @@ static int eap_authorize(void *instance, REQUEST *request)
 	return RLM_MODULE_UPDATED;
 }
 
+
+#ifdef WITH_PROXY
 /*
  *	If we're proxying EAP, then there may be magic we need
  *	to do.
@@ -512,19 +596,21 @@ static int eap_post_proxy(void *inst, REQUEST *request)
 							      request->proxy,
 							      REQUEST_DATA_EAP_TUNNEL_CALLBACK);
 		if (!data) {
-			radlog(L_ERR, "rlm_eap: Failed to retrieve callback for tunneled session!");
-			eap_handler_free(handler);
+			radlog_request(L_ERR, 0, request, "Failed to retrieve callback for tunneled session!");
+			eap_handler_free(inst, handler);
 			return RLM_MODULE_FAIL;
 		}
 
 		/*
 		 *	Do the callback...
 		 */
+		RDEBUG2("Doing post-proxy callback");
 		rcode = data->callback(handler, data->tls_session);
 		free(data);
 		if (rcode == 0) {
+			RDEBUG2("Failed in post-proxy callback");
 			eap_fail(handler);
-			eap_handler_free(handler);
+			eap_handler_free(inst, handler);
 			return RLM_MODULE_REJECT;
 		}
 
@@ -532,7 +618,7 @@ static int eap_post_proxy(void *inst, REQUEST *request)
 		 *	We are done, wrap the EAP-request in RADIUS to send
 		 *	with all other required radius attributes
 		 */
-		rcode = eap_compose(handler);
+		eap_compose(handler);
 
 		/*
 		 *	Add to the list only if it is EAP-Request, OR if
@@ -540,12 +626,16 @@ static int eap_post_proxy(void *inst, REQUEST *request)
 		 */
 		if ((handler->eap_ds->request->code == PW_EAP_REQUEST) &&
 		    (handler->eap_ds->request->type.type >= PW_EAP_MD5)) {
-			eaplist_add(inst, handler);
-
+			if (!eaplist_add(inst, handler)) {
+				eap_fail(handler);
+				eap_handler_free(inst, handler);
+				return RLM_MODULE_FAIL;
+			}
+			
 		} else {	/* couldn't have been LEAP, there's no tunnel */
-			DEBUG2("  rlm_eap: Freeing handler");
+			RDEBUG2("Freeing handler");
 			/* handler is not required any more, free it now */
-			eap_handler_free(handler);
+			eap_handler_free(inst, handler);
 		}
 
 		/*
@@ -568,9 +658,10 @@ static int eap_post_proxy(void *inst, REQUEST *request)
 		}
 
 		return RLM_MODULE_OK;
+	} else {
+		RDEBUG2("No pre-existing handler found");
 	}
 
-
 	/*
 	 *	There may be more than one Cisco-AVPair.
 	 *	Ensure we find the one with the LEAP attribute.
@@ -608,7 +699,7 @@ static int eap_post_proxy(void *inst, REQUEST *request)
 	 *	The format is very specific.
 	 */
 	if (vp->length != 17 + 34) {
-		DEBUG2("  rlm_eap: Cisco-AVPair with leap:session-key has incorrect length %d: Expected %d",
+		RDEBUG2("Cisco-AVPair with leap:session-key has incorrect length %d: Expected %d",
 		       vp->length, 17 + 34);
 		return RLM_MODULE_NOOP;
 	}
@@ -634,7 +725,7 @@ static int eap_post_proxy(void *inst, REQUEST *request)
 
 	return RLM_MODULE_UPDATED;
 }
-
+#endif
 
 /*
  *	The module name should be the only globally exported symbol.
@@ -653,7 +744,11 @@ module_t rlm_eap = {
 		NULL,			/* accounting */
 		NULL,			/* checksimul */
 		NULL,			/* pre-proxy */
+#ifdef WITH_PROXY
 		eap_post_proxy,		/* post-proxy */
+#else
+		NULL,
+#endif
 		NULL			/* post-auth */
 	},
 };
