@@ -1,7 +1,7 @@
 /*
  * rlm_eap_mschapv2.c    Handles that are called from eap
  *
- * Version:     $Id$
+ * Version:     $Id: d1b5ed575ebcbf57adc0de6778f553cff36aa9a4 $
  *
  *   This program is free software; you can redistribute it and/or modify
  *   it under the terms of the GNU General Public License as published by
@@ -21,7 +21,7 @@
  */
 
 #include <freeradius-devel/ident.h>
-RCSID("$Id$")
+RCSID("$Id: d1b5ed575ebcbf57adc0de6778f553cff36aa9a4 $")
 
 #include <freeradius-devel/autoconf.h>
 
@@ -34,16 +34,38 @@ RCSID("$Id$")
 
 typedef struct rlm_eap_mschapv2_t {
         int with_ntdomain_hack;
+	int send_error;
 } rlm_eap_mschapv2_t;
 
 static CONF_PARSER module_config[] = {
 	{ "with_ntdomain_hack",     PW_TYPE_BOOLEAN,
 	  offsetof(rlm_eap_mschapv2_t,with_ntdomain_hack), NULL, "no" },
 
+	{ "send_error",     PW_TYPE_BOOLEAN,
+	  offsetof(rlm_eap_mschapv2_t,send_error), NULL, "no" },
+
 	{ NULL, -1, 0, NULL, NULL }		/* end the list */
 };
 
 
+static void fix_mppe_keys(EAP_HANDLER *handler, mschapv2_opaque_t *data)
+{
+	pairmove2(&data->mppe_keys, &handler->request->reply->vps, ((311 << 16) | 7));
+	pairmove2(&data->mppe_keys, &handler->request->reply->vps, ((311 << 16) | 8));
+	pairmove2(&data->mppe_keys, &handler->request->reply->vps, ((311 << 16) | 16));
+	pairmove2(&data->mppe_keys, &handler->request->reply->vps, ((311 << 16) | 17));
+  
+}
+
+static void free_data(void *ptr)
+{
+	mschapv2_opaque_t *data = ptr;
+
+	pairfree(&data->mppe_keys);
+	pairfree(&data->reply);
+	free(data);
+}
+
 /*
  *	Detach the module.
  */
@@ -178,7 +200,7 @@ static int eapmschapv2_compose(EAP_HANDLER *handler, VALUE_PAIR *reply)
 
 	case PW_MSCHAP_ERROR:
 		DEBUG2("MSCHAP Failure\n");
-		length = 4 + MSCHAPV2_FAILURE_MESSAGE_LEN;
+		length = 4 + reply->length - 1;
 		eap_ds->request->type.data = malloc(length);
 
 		/*
@@ -195,7 +217,11 @@ static int eapmschapv2_compose(EAP_HANDLER *handler, VALUE_PAIR *reply)
 		eap_ds->request->type.data[1] = eap_ds->response->id;
 		length = htons(length);
 		memcpy((eap_ds->request->type.data + 2), &length, sizeof(uint16_t));
-		memcpy((eap_ds->request->type.data + 4), MSCHAPV2_FAILURE_MESSAGE, MSCHAPV2_FAILURE_MESSAGE_LEN);
+		/*
+		 *	Copy the entire failure message.
+		 */
+		memcpy((eap_ds->request->type.data + 4),
+		       reply->vp_strvalue + 1, reply->length - 1);
 		break;
 
 	default:
@@ -245,9 +271,11 @@ static int mschapv2_initiate(void *type_data, EAP_HANDLER *handler)
 	 */
 	data->code = PW_EAP_MSCHAPV2_CHALLENGE;
 	memcpy(data->challenge, challenge->vp_strvalue, MSCHAPV2_CHALLENGE_LEN);
+	data->mppe_keys = NULL;
+	data->reply = NULL;
 
 	handler->opaque = data;
-	handler->free_opaque = free;
+	handler->free_opaque = free_data;
 
 	/*
 	 *	Compose the EAP-MSCHAPV2 packet out of the data structure,
@@ -256,11 +284,13 @@ static int mschapv2_initiate(void *type_data, EAP_HANDLER *handler)
 	eapmschapv2_compose(handler, challenge);
 	pairfree(&challenge);
 
+#ifdef WITH_PROXY
 	/*
 	 *	The EAP session doesn't have enough information to
 	 *	proxy the "inside EAP" protocol.  Disable EAP proxying.
 	 */
 	handler->request->options &= ~RAD_REQUEST_OPTION_PROXY_EAP;
+#endif
 
 	/*
 	 *	We don't need to authorize the user at this point.
@@ -274,7 +304,7 @@ static int mschapv2_initiate(void *type_data, EAP_HANDLER *handler)
 	return 1;
 }
 
-
+#ifdef WITH_PROXY
 /*
  *	Do post-proxy processing,
  *	0 = fail
@@ -336,10 +366,14 @@ static int mschap_postproxy(EAP_HANDLER *handler, void *tunnel_data)
 	 *
 	 *	FIXME: Use intelligent names...
 	 */
-	pairdelete(&handler->request->reply->vps, ((311 << 16) | 7));
-	pairdelete(&handler->request->reply->vps, ((311 << 16) | 8));
-	pairdelete(&handler->request->reply->vps, ((311 << 16) | 16));
-	pairdelete(&handler->request->reply->vps, ((311 << 16) | 17));
+	fix_mppe_keys(handler, data);
+
+	/*
+	 * save any other attributes for re-use in the final
+	 * access-accept e.g. vlan, etc. This lets the PEAP
+	 * use_tunneled_reply code work
+	 */
+	data->reply = paircopy(handler->request->reply->vps);
 
 	/*
 	 *	And we need to challenge the user, not ack/reject them,
@@ -350,7 +384,7 @@ static int mschap_postproxy(EAP_HANDLER *handler, void *tunnel_data)
 
 	return 1;
 }
-
+#endif
 
 /*
  *	Authenticate a previously sent challenge.
@@ -360,7 +394,8 @@ static int mschapv2_authenticate(void *arg, EAP_HANDLER *handler)
 	int rcode;
 	mschapv2_opaque_t *data;
 	EAP_DS *eap_ds = handler->eap_ds;
-	VALUE_PAIR *challenge, *response;
+	VALUE_PAIR *challenge, *response, *name;
+	rlm_eap_mschapv2_t *inst = (rlm_eap_mschapv2_t *) arg;
 
 	rad_assert(handler->request != NULL);
 	rad_assert(handler->stage == AUTHENTICATE);
@@ -370,7 +405,7 @@ static int mschapv2_authenticate(void *arg, EAP_HANDLER *handler)
 	/*
 	 *	Sanity check the response.
 	 */
-	if (eap_ds->response->length <= 4) {
+	if (eap_ds->response->length <= 5) {
 		radlog(L_ERR, "rlm_eap_mschapv2: corrupted data");
 		return 0;
 	}
@@ -407,6 +442,8 @@ static int mschapv2_authenticate(void *arg, EAP_HANDLER *handler)
 		 *	a challenge.
 		 */
 	case PW_EAP_MSCHAPV2_RESPONSE:
+		if (data->code == PW_EAP_MSCHAPV2_FAILURE) goto failure;
+
 		if (data->code != PW_EAP_MSCHAPV2_CHALLENGE) {
 			radlog(L_ERR, "rlm_eap_mschapv2: Unexpected response received");
 			return 0;
@@ -454,15 +491,36 @@ static int mschapv2_authenticate(void *arg, EAP_HANDLER *handler)
 			return 0;
 		}
 
+#ifdef WITH_PROXY
 		/*
 		 *	It's a success.  Don't proxy it.
 		 */
 		handler->request->options &= ~RAD_REQUEST_OPTION_PROXY_EAP;
+#endif
+		pairadd(&handler->request->reply->vps, data->reply);
+		data->reply = NULL;
 
 		eap_ds->request->code = PW_EAP_SUCCESS;
+
+		pairadd(&handler->request->reply->vps, data->mppe_keys);
+		data->mppe_keys = NULL;
 		return 1;
 		break;
 
+		/*
+		 *	Ack of a failure message
+		 */
+        case PW_EAP_MSCHAPV2_FAILURE:
+		if (data->code != PW_EAP_MSCHAPV2_FAILURE) {
+			radlog(L_ERR, "rlm_eap_mschapv2: Unexpected FAILURE received");
+			return 0;
+		}
+
+	failure:
+                handler->request->options &= ~RAD_REQUEST_OPTION_PROXY_EAP;
+                eap_ds->request->code = PW_EAP_FAILURE;
+                return 1;
+
 		/*
 		 *	Something else, we don't know what it is.
 		 */
@@ -502,13 +560,38 @@ static int mschapv2_authenticate(void *arg, EAP_HANDLER *handler)
 	response->vp_strvalue[0] = eap_ds->response->type.data[1];
 	response->vp_strvalue[1] = eap_ds->response->type.data[5 + MSCHAPV2_RESPONSE_LEN];
 
+	name = pairmake("MS-CHAP-User-Name", "", T_OP_EQ);
+	if (!name) {
+		pairfree(&challenge);
+		pairfree(&response);
+		radlog(L_ERR, "rlm_eap_mschapv2: Failed creating MS-CHAP-User-Name: %s", fr_strerror());
+		return 0;
+	}
+	
+	/*
+	 *	MS-Length - MS-Value - 5.
+	 */
+	name->length = (((eap_ds->response->type.data[2] << 8) |
+                         eap_ds->response->type.data[3]) -
+			eap_ds->response->type.data[4] - 5);
+	if (name->length >= sizeof(name->vp_strvalue)) {
+		name->length = sizeof(name->vp_strvalue) - 1;
+	}
+
+	memcpy(name->vp_strvalue,
+	       &eap_ds->response->type.data[4 + MSCHAPV2_RESPONSE_LEN],
+	       name->length);
+	name->vp_strvalue[name->length] = '\0';
+
 	/*
 	 *	Add the pairs to the request, and call the 'mschap'
 	 *	module.
 	 */
 	pairadd(&handler->request->packet->vps, challenge);
 	pairadd(&handler->request->packet->vps, response);
+	pairadd(&handler->request->packet->vps, name);
 
+#ifdef WITH_PROXY
 	/*
 	 *	If this options is set, then we do NOT authenticate the
 	 *	user here.  Instead, now that we've added the MS-CHAP
@@ -522,7 +605,6 @@ static int mschapv2_authenticate(void *arg, EAP_HANDLER *handler)
 	if (handler->request->options & RAD_REQUEST_OPTION_PROXY_EAP) {
 		char *username = NULL;
 		eap_tunnel_data_t *tunnel;
-		rlm_eap_mschapv2_t *inst = (rlm_eap_mschapv2_t *) arg;
 
 		/*
 		 *	Set up the callbacks for the tunnel
@@ -583,6 +665,7 @@ static int mschapv2_authenticate(void *arg, EAP_HANDLER *handler)
 		 */
 		return 1;
 	}
+#endif
 
 	/*
 	 *	This is a wild & crazy hack.
@@ -593,10 +676,7 @@ static int mschapv2_authenticate(void *arg, EAP_HANDLER *handler)
 	 *	Delete MPPE keys & encryption policy.  We don't
 	 *	want these here.
 	 */
-	pairdelete(&handler->request->reply->vps, ((311 << 16) | 7));
-	pairdelete(&handler->request->reply->vps, ((311 << 16) | 8));
-	pairdelete(&handler->request->reply->vps, ((311 << 16) | 16));
-	pairdelete(&handler->request->reply->vps, ((311 << 16) | 17));
+	fix_mppe_keys(handler, data);
 
 	/*
 	 *	Take the response from the mschap module, and
@@ -607,17 +687,14 @@ static int mschapv2_authenticate(void *arg, EAP_HANDLER *handler)
 		pairmove2(&response, &handler->request->reply->vps,
 			 PW_MSCHAP2_SUCCESS);
 		data->code = PW_EAP_MSCHAPV2_SUCCESS;
+
+	} else if (inst->send_error) {
+	  pairmove2(&response, &handler->request->reply->vps,
+		    PW_MSCHAP_ERROR);
+		data->code = PW_EAP_MSCHAPV2_FAILURE;
 	} else {
-		/*
-		 *	Don't return anything in the error message.
-		 */
 		eap_ds->request->code = PW_EAP_FAILURE;
 		return 1;
-#if 0
-		pairmove2(&handler->request->reply->vps, &response
-			  PW_MSCHAP_ERROR);
-		data->code = PW_EAP_MSCHAPV2_FAILURE;
-#endif
 	}
 
 	/*
