@@ -4,7 +4,7 @@
  * The development of the EAP/SIM support was funded by Internet Foundation
  * Austria (http://www.nic.at/ipa).
  *
- * Version:     $Id: rlm_eap_sim.c,v 1.20 2007/11/25 14:02:09 aland Exp $
+ * Version:     $Id: 87cba647e0cd1e9ba49900c65320bb5c76ceed06 $
  *
  *   This program is free software; you can redistribute it and/or modify
  *   it under the terms of the GNU General Public License as published by
@@ -26,7 +26,7 @@
  */
 
 #include <freeradius-devel/ident.h>
-RCSID("$Id: rlm_eap_sim.c,v 1.20 2007/11/25 14:02:09 aland Exp $")
+RCSID("$Id: 87cba647e0cd1e9ba49900c65320bb5c76ceed06 $")
 
 #include <freeradius-devel/autoconf.h>
 
@@ -56,7 +56,7 @@ static void add_reply(VALUE_PAIR** vp,
 	if (!reply_attr) {
 		DEBUG("rlm_eap_sim: "
 		      "add_reply failed to create attribute %s: %s\n",
-		      name, librad_errstr);
+		      name, fr_strerror());
 		return;
 	}
 
@@ -226,8 +226,10 @@ static int eap_sim_sendchallenge(EAP_HANDLER *handler)
 	/* outvps is the data to the client. */
 	outvps= &handler->request->reply->vps;
 
-	printf("+++> EAP-sim decoded packet:\n");
-	vp_printlist(stdout, *invps);
+	if ((debug_flag > 0) && fr_log_fp) {
+		fprintf(fr_log_fp, "+++> EAP-sim decoded packet:\n");
+		debug_pair_list(*invps);
+	}
 
 	/* okay, we got the challenges! Put them into an attribute */
 	newvp = paircreate(ATTRIBUTE_EAP_SIM_BASE+PW_EAP_SIM_RAND,
@@ -245,8 +247,14 @@ static int eap_sim_sendchallenge(EAP_HANDLER *handler)
 	pairreplace(outvps, newvp);
 
 	/* make a copy of the identity */
-	ess->keys.identitylen = strlen(handler->identity);
-	memcpy(ess->keys.identity, handler->identity, ess->keys.identitylen);
+	newvp = pairfind(*invps, ATTRIBUTE_EAP_SIM_BASE + PW_EAP_SIM_IDENTITY);
+	if (newvp) {
+		ess->keys.identitylen = newvp->length;
+		memcpy(ess->keys.identity, newvp->vp_octets, newvp->length);
+	} else {
+		ess->keys.identitylen = strlen(handler->identity);
+		memcpy(ess->keys.identity, handler->identity, ess->keys.identitylen);
+	}
 
 	/* all set, calculate keys! */
 	eapsim_calculate_keys(&ess->keys);
@@ -549,7 +557,8 @@ static int eap_sim_authenticate(void *arg, EAP_HANDLER *handler)
 					 handler->eap_ds->response->type.length);
 
 	if(!success) {
-	  return 0;
+		DEBUG("Failed to decode EAP-SIM");
+		return 0;
 	}
 
 	/* see what kind of message we have gotten */
@@ -564,6 +573,7 @@ static int eap_sim_authenticate(void *arg, EAP_HANDLER *handler)
 	 *	Client error supersedes anything else.
 	 */
 	if (subtype == eapsim_client_error) {
+		DEBUG2("Client says error.  Stopping!");
 		return 0;
 	}
 
