@@ -1,7 +1,7 @@
 /*
  * rlm_eap_tls.c  contains the interfaces that are called from eap
  *
- * Version:     $Id$
+ * Version:     $Id: 27db07d7272e9e0d3dfd7c356f867daa049226a2 $
  *
  *   This program is free software; you can redistribute it and/or modify
  *   it under the terms of the GNU General Public License as published by
@@ -24,7 +24,7 @@
  */
 
 #include <freeradius-devel/ident.h>
-RCSID("$Id$")
+RCSID("$Id: 27db07d7272e9e0d3dfd7c356f867daa049226a2 $")
 
 #include <freeradius-devel/autoconf.h>
 
@@ -36,6 +36,8 @@ RCSID("$Id$")
 #include <openssl/evp.h>
 #endif
 
+#include <openssl/x509.h>
+
 #include "rlm_eap_tls.h"
 #include "config.h"
 
@@ -43,6 +45,10 @@ RCSID("$Id$")
 #include <sys/stat.h>
 #endif
 
+#ifdef HAVE_OPENSSL_OCSP_H
+#include <openssl/ocsp.h>
+#endif
+
 static CONF_PARSER cache_config[] = {
 	{ "enable", PW_TYPE_BOOLEAN,
 	  offsetof(EAP_TLS_CONF, session_cache_enable), NULL, "no" },
@@ -55,6 +61,32 @@ static CONF_PARSER cache_config[] = {
  	{ NULL, -1, 0, NULL, NULL }           /* end the list */
 };
 
+static CONF_PARSER verify_config[] = {
+	{ "tmpdir", PW_TYPE_STRING_PTR,
+	  offsetof(EAP_TLS_CONF, verify_tmp_dir), NULL, NULL},
+	{ "client", PW_TYPE_STRING_PTR,
+	  offsetof(EAP_TLS_CONF, verify_client_cert_cmd), NULL, NULL},
+ 	{ NULL, -1, 0, NULL, NULL }           /* end the list */
+};
+
+#ifdef HAVE_OPENSSL_OCSP_H
+static CONF_PARSER ocsp_config[] = {
+	{ "enable", PW_TYPE_BOOLEAN,
+	  offsetof(EAP_TLS_CONF, ocsp_enable), NULL, "no"},
+	{ "override_cert_url", PW_TYPE_BOOLEAN,
+	  offsetof(EAP_TLS_CONF, ocsp_override_url), NULL, "no"},
+	{ "url", PW_TYPE_STRING_PTR,
+	  offsetof(EAP_TLS_CONF, ocsp_url), NULL, NULL },
+	{ "use_nonce", PW_TYPE_BOOLEAN,
+	  offsetof(EAP_TLS_CONF, ocsp_use_nonce), NULL, "yes"},
+	{ "timeout", PW_TYPE_INTEGER,
+	  offsetof(EAP_TLS_CONF, ocsp_timeout), NULL, "0" },
+	{ "softfail", PW_TYPE_BOOLEAN,
+	  offsetof(EAP_TLS_CONF, ocsp_softfail), NULL, "no"},
+ 	{ NULL, -1, 0, NULL, NULL }           /* end the list */
+};
+#endif
+
 static CONF_PARSER module_config[] = {
 	{ "rsa_key_exchange", PW_TYPE_BOOLEAN,
 	  offsetof(EAP_TLS_CONF, rsa_key), NULL, "no" },
@@ -88,6 +120,8 @@ static CONF_PARSER module_config[] = {
 	  offsetof(EAP_TLS_CONF, include_length), NULL, "yes" },
 	{ "check_crl", PW_TYPE_BOOLEAN,
 	  offsetof(EAP_TLS_CONF, check_crl), NULL, "no"},
+	{ "allow_expired_crl", PW_TYPE_BOOLEAN,
+	  offsetof(EAP_TLS_CONF, allow_expired_crl), NULL, NULL},
 	{ "check_cert_cn", PW_TYPE_STRING_PTR,
 	  offsetof(EAP_TLS_CONF, check_cert_cn), NULL, NULL},
 	{ "cipher_list", PW_TYPE_STRING_PTR,
@@ -96,9 +130,24 @@ static CONF_PARSER module_config[] = {
 	  offsetof(EAP_TLS_CONF, check_cert_issuer), NULL, NULL},
 	{ "make_cert_command", PW_TYPE_STRING_PTR,
 	  offsetof(EAP_TLS_CONF, make_cert_command), NULL, NULL},
+	{ "virtual_server", PW_TYPE_STRING_PTR,
+	  offsetof(EAP_TLS_CONF, virtual_server), NULL, NULL },
+
+#if OPENSSL_VERSION_NUMBER >= 0x0090800fL
+#ifndef OPENSSL_NO_ECDH
+	{ "ecdh_curve", PW_TYPE_STRING_PTR,
+	  offsetof(EAP_TLS_CONF, ecdh_curve), NULL, "prime256v1"},
+#endif
+#endif
 
 	{ "cache", PW_TYPE_SUBSECTION, 0, NULL, (const void *) cache_config },
 
+	{ "verify", PW_TYPE_SUBSECTION, 0, NULL, (const void *) verify_config },
+
+#ifdef HAVE_OPENSSL_OCSP_H
+	{ "ocsp", PW_TYPE_SUBSECTION, 0, NULL, (const void *) ocsp_config },
+#endif
+
  	{ NULL, -1, 0, NULL, NULL }           /* end the list */
 };
 
@@ -155,9 +204,6 @@ static int generate_eph_rsa_key(SSL_CTX *ctx)
 
 
 /*
- *	These functions don't do anything other than print debugging
- *	messages.
- *
  *	FIXME: Write sessions to some long-term storage, so that
  *	       session resumption can still occur after the server
  *	       restarts.
@@ -213,6 +259,264 @@ static SSL_SESSION *cbtls_get_session(UNUSED SSL *s,
 	return NULL;
 }
 
+#ifdef HAVE_OPENSSL_OCSP_H
+/*
+ * This function extracts the OCSP Responder URL
+ * from an existing x509 certificate.
+ */
+static int ocsp_parse_cert_url(X509 *cert, char **phost, char **pport,
+			       char **ppath, int *pssl)
+{
+	int i;
+
+	AUTHORITY_INFO_ACCESS *aia;
+	ACCESS_DESCRIPTION *ad;
+
+	aia = X509_get_ext_d2i(cert, NID_info_access, NULL, NULL);
+
+	for (i = 0; i < sk_ACCESS_DESCRIPTION_num(aia); i++) {
+		ad = sk_ACCESS_DESCRIPTION_value(aia, 0);
+		if (OBJ_obj2nid(ad->method) == NID_ad_OCSP) {
+			if (ad->location->type == GEN_URI) {
+				if(OCSP_parse_url(ad->location->d.ia5->data,
+					phost, pport, ppath, pssl))
+					return 1;
+			}
+		}
+	}
+	return 0;
+}
+
+/*
+ * This function sends a OCSP request to a defined OCSP responder
+ * and checks the OCSP response for correctness.
+ */
+
+/* Maximum leeway in validity period: default 5 minutes */
+#define MAX_VALIDITY_PERIOD     (5 * 60)
+
+static int ocsp_check(X509_STORE *store, X509 *issuer_cert, X509 *client_cert,
+		      EAP_TLS_CONF *conf)
+{
+	OCSP_CERTID *certid;
+	OCSP_REQUEST *req;
+	OCSP_RESPONSE *resp = NULL;
+	OCSP_BASICRESP *bresp = NULL;
+	char *host = NULL;
+	char *port = NULL;
+	char *path = NULL;
+	int use_ssl = -1;
+	long nsec = MAX_VALIDITY_PERIOD, maxage = -1;
+	BIO *cbio, *bio_out;
+	int ocsp_ok = 0;
+	int status ;
+	ASN1_GENERALIZEDTIME *rev, *thisupd, *nextupd;
+	int reason;
+#if OPENSSL_VERSION_NUMBER >= 0x1000003f
+	OCSP_REQ_CTX *ctx;
+	int rc;
+	struct timeval now;
+	struct timeval when;
+#endif
+
+	/*
+	 * Create OCSP Request
+	 */
+	certid = OCSP_cert_to_id(NULL, client_cert, issuer_cert);
+	req = OCSP_REQUEST_new();
+	OCSP_request_add0_id(req, certid);
+	if(conf->ocsp_use_nonce){
+		OCSP_request_add1_nonce(req, NULL, 8);
+	}
+
+	/*
+	 * Send OCSP Request and get OCSP Response
+	 */
+
+	/* Get OCSP responder URL */
+	if(conf->ocsp_override_url) {
+		OCSP_parse_url(conf->ocsp_url, &host, &port, &path, &use_ssl);
+	}
+	else {
+		ocsp_parse_cert_url(client_cert, &host, &port, &path, &use_ssl);
+	}
+
+	if (!host || !port || !path) {
+		DEBUG2("[ocsp] - Host / port / path missing.  Not doing OCSP.");
+		ocsp_ok = 2;
+		goto ocsp_skip;
+	}
+
+	DEBUG2("[ocsp] --> Responder URL = http://%s:%s%s", host, port, path);
+
+	/* Setup BIO socket to OCSP responder */
+	cbio = BIO_new_connect(host);
+
+	bio_out = BIO_new_fp(stdout, BIO_NOCLOSE);
+
+	BIO_set_conn_port(cbio, port);
+#if OPENSSL_VERSION_NUMBER < 0x1000003f
+	BIO_do_connect(cbio);
+
+	/* Send OCSP request and wait for response */
+	resp = OCSP_sendreq_bio(cbio, path, req);
+	if (!resp) {
+		radlog(L_ERR, "Error: Couldn't get OCSP response");
+		ocsp_ok = 2;
+		goto ocsp_end;
+	}
+#else
+	if (conf->ocsp_timeout)
+		BIO_set_nbio(cbio, 1);
+
+	rc = BIO_do_connect(cbio);
+	if ((rc <= 0) && ((!conf->ocsp_timeout) || !BIO_should_retry(cbio))) {
+		radlog(L_ERR, "Error: Couldn't connect to OCSP responder");
+		ocsp_ok = 2;
+		goto ocsp_end;
+	}
+
+	ctx = OCSP_sendreq_new(cbio, path, req, -1);
+	if (!ctx) {
+		radlog(L_ERR, "Error: Couldn't send OCSP request");
+		ocsp_ok = 2;
+		goto ocsp_end;
+	}
+
+	gettimeofday(&when, NULL);
+	when.tv_sec += conf->ocsp_timeout;
+
+	do {
+		rc = OCSP_sendreq_nbio(&resp, ctx);
+		if (conf->ocsp_timeout) {
+			gettimeofday(&now, NULL);
+			if (!timercmp(&now, &when, <))
+				break;
+		}
+	} while ((rc == -1) && BIO_should_retry(cbio));
+
+	if (conf->ocsp_timeout && (rc == -1) && BIO_should_retry(cbio)) {
+		radlog(L_ERR, "Error: OCSP response timed out");
+		ocsp_ok = 2;
+		goto ocsp_end;
+	}
+
+	OCSP_REQ_CTX_free(ctx);
+
+	if (rc == 0) {
+		radlog(L_ERR, "Error: Couldn't get OCSP response");
+		ocsp_ok = 2;
+		goto ocsp_end;
+	}
+#endif
+
+	/* Verify OCSP response status */
+	status = OCSP_response_status(resp);
+	DEBUG2("[ocsp] --> Response status: %s",OCSP_response_status_str(status));
+	if(status != OCSP_RESPONSE_STATUS_SUCCESSFUL) {
+		radlog(L_ERR, "Error: OCSP response status: %s", OCSP_response_status_str(status));
+		goto ocsp_end;
+	}
+	bresp = OCSP_response_get1_basic(resp);
+	if(conf->ocsp_use_nonce && OCSP_check_nonce(req, bresp)!=1) {
+		radlog(L_ERR, "Error: OCSP response has wrong nonce value");
+		goto ocsp_end;
+	}
+	if(OCSP_basic_verify(bresp, NULL, store, 0)!=1){
+		radlog(L_ERR, "Error: Couldn't verify OCSP basic response");
+		goto ocsp_end;
+	}
+	/*	Verify OCSP cert status */
+	if(!OCSP_resp_find_status(bresp, certid, &status, &reason,
+				                      &rev, &thisupd, &nextupd)) {
+		radlog(L_ERR, "ERROR: No Status found.\n");
+		goto ocsp_end;
+	}
+
+	if (!OCSP_check_validity(thisupd, nextupd, nsec, maxage)) {
+		BIO_puts(bio_out, "WARNING: Status times invalid.\n");
+		ERR_print_errors(bio_out);
+		goto ocsp_end;
+	}
+	BIO_puts(bio_out, "\tThis Update: ");
+        ASN1_GENERALIZEDTIME_print(bio_out, thisupd);
+        BIO_puts(bio_out, "\n");
+	if (nextupd) {
+		BIO_puts(bio_out, "\tNext Update: ");
+		ASN1_GENERALIZEDTIME_print(bio_out, nextupd);
+		BIO_puts(bio_out, "\n");
+	}
+
+	switch (status) {
+	case V_OCSP_CERTSTATUS_GOOD:
+		DEBUG2("[oscp] --> Cert status: good");
+		ocsp_ok = 1;
+		break;
+
+	default:
+		/* REVOKED / UNKNOWN */
+		DEBUG2("[ocsp] --> Cert status: %s",OCSP_cert_status_str(status));
+                if (reason != -1)
+			DEBUG2("[ocsp] --> Reason: %s", OCSP_crl_reason_str(reason));
+                BIO_puts(bio_out, "\tRevocation Time: ");
+                ASN1_GENERALIZEDTIME_print(bio_out, rev);
+                BIO_puts(bio_out, "\n");
+		break;
+	}
+
+ocsp_end:
+	/* Free OCSP Stuff */
+	OCSP_REQUEST_free(req);
+	OCSP_RESPONSE_free(resp);
+	free(host);
+	free(port);
+	free(path);
+	BIO_free_all(cbio);
+	OCSP_BASICRESP_free(bresp);
+
+ocsp_skip:
+	switch (ocsp_ok) {
+	case 1:
+		DEBUG2("[ocsp] --> Certificate is valid!");
+		break;
+	case 2:
+		if (conf->ocsp_softfail) {
+			DEBUG2("[ocsp] --> Unable to check certificate; assuming valid.");
+			DEBUG2("[ocsp] --> Warning! This may be insecure.");
+			ocsp_ok = 1;
+		} else {
+			DEBUG2("[ocsp] --> Unable to check certificate; failing!");
+			ocsp_ok = 0;
+		}
+		break;
+	default:
+		DEBUG2("[ocsp] --> Certificate has been expired/revoked!");
+		break;
+	}
+
+	return ocsp_ok;
+}
+#endif	/* HAVE_OPENSSL_OCSP_H */
+
+/*
+ *	For creating certificate attributes.
+ */
+static const char *cert_attr_names[6][2] = {
+  { "TLS-Client-Cert-Serial",		"TLS-Cert-Serial" },
+  { "TLS-Client-Cert-Expiration",	"TLS-Cert-Expiration" },
+  { "TLS-Client-Cert-Subject",		"TLS-Cert-Subject" },
+  { "TLS-Client-Cert-Issuer",		"TLS-Cert-Issuer" },
+  { "TLS-Client-Cert-Common-Name",	"TLS-Cert-Common-Name" },
+  { "TLS-Client-Cert-Subject-Alt-Name-Email",	"TLS-Cert-Subject-Alt-Name-Email" }
+};
+
+#define EAPTLS_SERIAL		(0)
+#define EAPTLS_EXPIRATION	(1)
+#define EAPTLS_SUBJECT		(2)
+#define EAPTLS_ISSUER		(3)
+#define EAPTLS_CN		(4)
+#define EAPTLS_SAN_EMAIL	(5)
+
 /*
  *	Before trusting a certificate, you must make sure that the
  *	certificate is 'valid'. There are several steps that your
@@ -242,25 +546,38 @@ static int cbtls_verify(int ok, X509_STORE_CTX *ctx)
 {
 	char subject[1024]; /* Used for the subject name */
 	char issuer[1024]; /* Used for the issuer name */
+	char attribute[1024];
+	char value[1024];
 	char common_name[1024];
 	char cn_str[1024];
+	char buf[64];
 	EAP_HANDLER *handler = NULL;
 	X509 *client_cert;
+	X509_CINF *client_inf;
+	STACK_OF(X509_EXTENSION) *ext_list;
+	X509 *issuer_cert;
 	SSL *ssl;
-	int err, depth;
+	int err, depth, lookup, loc;
 	EAP_TLS_CONF *conf;
 	int my_ok = ok;
 	REQUEST *request;
+	ASN1_INTEGER *sn = NULL;
+	ASN1_TIME *asn_time = NULL;
+#ifdef HAVE_OPENSSL_OCSP_H
+	X509_STORE *ocsp_store = NULL;
+#endif
 
 	client_cert = X509_STORE_CTX_get_current_cert(ctx);
 	err = X509_STORE_CTX_get_error(ctx);
 	depth = X509_STORE_CTX_get_error_depth(ctx);
 
-	if (!my_ok) {
-		radlog(L_ERR,"--> verify error:num=%d:%s\n",err,
-			X509_verify_cert_error_string(err));
-		return my_ok;
-	}
+	lookup = depth;
+
+	/*
+	 *	Log client/issuing cert.  If there's an error, log
+	 *	issuing cert.
+	 */
+	if ((lookup > 1) && !my_ok) lookup = 1;
 
 	/*
 	 * Retrieve the pointer to the SSL of the connection currently treated
@@ -270,6 +587,45 @@ static int cbtls_verify(int ok, X509_STORE_CTX *ctx)
 	handler = (EAP_HANDLER *)SSL_get_ex_data(ssl, 0);
 	request = handler->request;
 	conf = (EAP_TLS_CONF *)SSL_get_ex_data(ssl, 1);
+#ifdef HAVE_OPENSSL_OCSP_H
+	ocsp_store = (X509_STORE *)SSL_get_ex_data(ssl, 2);
+#endif
+
+
+	/*
+	 *	Get the Serial Number
+	 */
+	buf[0] = '\0';
+	sn = X509_get_serialNumber(client_cert);
+
+	/*
+	 *	For this next bit, we create the attributes *only* if
+	 *	we're at the client or issuing certificate.
+	 */
+	if ((lookup <= 1) && sn && (sn->length < (sizeof(buf) / 2))) {
+		char *p = buf;
+		int i;
+
+		for (i = 0; i < sn->length; i++) {
+			sprintf(p, "%02x", (unsigned int)sn->data[i]);
+			p += 2;
+		}
+		pairadd(&handler->certs,
+			pairmake(cert_attr_names[EAPTLS_SERIAL][lookup], buf, T_OP_SET));
+	}
+
+
+	/*
+	 *	Get the Expiration Date
+	 */
+	buf[0] = '\0';
+	asn_time = X509_get_notAfter(client_cert);
+	if ((lookup <= 1) && asn_time && (asn_time->length < sizeof(buf))) {
+		memcpy(buf, (char*) asn_time->data, asn_time->length);
+		buf[asn_time->length] = '\0';
+		pairadd(&handler->certs,
+			pairmake(cert_attr_names[EAPTLS_EXPIRATION][lookup], buf, T_OP_SET));
+	}
 
 	/*
 	 *	Get the Subject & Issuer
@@ -277,18 +633,140 @@ static int cbtls_verify(int ok, X509_STORE_CTX *ctx)
 	subject[0] = issuer[0] = '\0';
 	X509_NAME_oneline(X509_get_subject_name(client_cert), subject,
 			  sizeof(subject));
+	subject[sizeof(subject) - 1] = '\0';
+	if ((lookup <= 1) && subject[0] && (strlen(subject) < MAX_STRING_LEN)) {
+		pairadd(&handler->certs,
+			pairmake(cert_attr_names[EAPTLS_SUBJECT][lookup], subject, T_OP_SET));
+	}
+
 	X509_NAME_oneline(X509_get_issuer_name(ctx->current_cert), issuer,
 			  sizeof(issuer));
-
-	subject[sizeof(subject) - 1] = '\0';
 	issuer[sizeof(issuer) - 1] = '\0';
+	if ((lookup <= 1) && issuer[0] && (strlen(issuer) < MAX_STRING_LEN)) {
+		pairadd(&handler->certs,
+			pairmake(cert_attr_names[EAPTLS_ISSUER][lookup], issuer, T_OP_SET));
+	}
 
 	/*
-	 *	Get the Common Name
+	 *	Get the Common Name, if there is a subject.
 	 */
 	X509_NAME_get_text_by_NID(X509_get_subject_name(client_cert),
 				  NID_commonName, common_name, sizeof(common_name));
 	common_name[sizeof(common_name) - 1] = '\0';
+	if ((lookup <= 1) && common_name[0] && subject[0] && (strlen(common_name) < MAX_STRING_LEN)) {
+		pairadd(&handler->certs,
+			pairmake(cert_attr_names[EAPTLS_CN][lookup], common_name, T_OP_SET));
+	}
+
+#ifdef GEN_EMAIL
+	/*
+	 *	Get the RFC822 Subject Alternative Name
+	 */
+	loc = X509_get_ext_by_NID(client_cert, NID_subject_alt_name, 0);
+	if (lookup <= 1 && loc >= 0) {
+		X509_EXTENSION *ext = NULL;
+		GENERAL_NAMES *names = NULL;
+		int i;
+
+		if ((ext = X509_get_ext(client_cert, loc)) &&
+		    (names = X509V3_EXT_d2i(ext))) {
+			for (i = 0; i < sk_GENERAL_NAME_num(names); i++) {
+				GENERAL_NAME *name = sk_GENERAL_NAME_value(names, i);
+
+				switch (name->type) {
+				case GEN_EMAIL:
+					if (ASN1_STRING_length(name->d.rfc822Name) >= MAX_STRING_LEN)
+						break;
+
+					pairadd(&handler->certs,
+						pairmake(cert_attr_names[EAPTLS_SAN_EMAIL][lookup],
+							 ASN1_STRING_data(name->d.rfc822Name), T_OP_SET));
+					break;
+				default:
+					/* XXX TODO handle other SAN types */
+					break;
+				}
+			}
+		}
+		if (names != NULL)
+			sk_GENERAL_NAME_free(names);
+	}
+#endif	/* GEN_EMAIL */
+
+	/*
+	 *	If the CRL has expired, that might still be OK.
+	 */
+	if (!my_ok &&
+	    (conf->allow_expired_crl) &&
+	    (err == X509_V_ERR_CRL_HAS_EXPIRED)) {
+		my_ok = 1;
+		X509_STORE_CTX_set_error( ctx, 0 );
+	}
+
+	if (!my_ok) {
+		const char *p = X509_verify_cert_error_string(err);
+		radlog(L_ERR,"--> verify error:num=%d:%s\n",err, p);
+		radius_pairmake(request, &request->packet->vps,
+				"Module-Failure-Message", p, T_OP_SET);
+		return my_ok;
+	}
+
+	if (lookup == 0) {
+		client_inf = client_cert->cert_info;
+		ext_list = client_inf->extensions;
+	} else {
+		ext_list = NULL;
+	}
+
+	/*
+	 *	Grab the X509 extensions, and create attributes out of them.
+	 *	For laziness, we re-use the OpenSSL names
+	 */
+	if (sk_X509_EXTENSION_num(ext_list) > 0) {
+		int i, len;
+		char *p;
+		BIO *out;
+
+		out = BIO_new(BIO_s_mem());
+		strlcpy(attribute, "TLS-Client-Cert-", sizeof(attribute));
+
+		for (i = 0; i < sk_X509_EXTENSION_num(ext_list); i++) {
+			ASN1_OBJECT *obj;
+			X509_EXTENSION *ext;
+			VALUE_PAIR *vp;
+
+			ext = sk_X509_EXTENSION_value(ext_list, i);
+
+			obj = X509_EXTENSION_get_object(ext);
+			i2a_ASN1_OBJECT(out, obj);
+			len = BIO_read(out, attribute + 16 , sizeof(attribute) - 16 - 1);
+			if (len <= 0) continue;
+
+			attribute[16 + len] = '\0';
+
+			X509V3_EXT_print(out, ext, 0, 0);
+			len = BIO_read(out, value, sizeof(value) - 1);
+			if (len <= 0) continue;
+
+			value[len] = '\0';
+
+			/*
+			 *	Mash the OpenSSL name to our name, and
+			 *	create the attribute.
+			 */
+			for (p = attribute + 16; *p != '\0'; p++) {
+				if (*p == ' ') *p = '-';
+			}
+
+			vp = pairmake(attribute, value, T_OP_ADD);
+			if (vp) {
+				pairadd(&handler->certs, vp);
+				debug_pair_list(vp);
+			}
+		}
+
+		BIO_free_all(out);
+	}
 
 	switch (ctx->error) {
 
@@ -346,6 +824,72 @@ static int cbtls_verify(int ok, X509_STORE_CTX *ctx)
 				}
 			}
 		} /* check_cert_cn */
+
+#ifdef HAVE_OPENSSL_OCSP_H
+		if (my_ok && conf->ocsp_enable){
+			RDEBUG2("--> Starting OCSP Request");
+			if(X509_STORE_CTX_get1_issuer(&issuer_cert, ctx, client_cert)!=1) {
+				radlog(L_ERR, "Error: Couldn't get issuer_cert for %s", common_name);
+			}
+			my_ok = ocsp_check(ocsp_store, issuer_cert, client_cert, conf);
+		}
+#endif
+
+		while (conf->verify_client_cert_cmd) {
+			char filename[256];
+			int fd;
+			FILE *fp;
+
+			snprintf(filename, sizeof(filename), "%s/%s.client.XXXXXXXX",
+				 conf->verify_tmp_dir, progname);
+			fd = mkstemp(filename);
+			if (fd < 0) {
+				RDEBUG("Failed creating file in %s: %s",
+				       conf->verify_tmp_dir, strerror(errno));
+				break;
+			}
+
+			fp = fdopen(fd, "w");
+			if (!fp) {
+				RDEBUG("Failed opening file %s: %s",
+				       filename, strerror(errno));
+				break;
+			}
+
+			if (!PEM_write_X509(fp, client_cert)) {
+				fclose(fp);
+				RDEBUG("Failed writing certificate to file");
+				goto do_unlink;
+			}
+			fclose(fp);
+
+			if (!radius_pairmake(request, &request->packet->vps,
+					     "TLS-Client-Cert-Filename",
+					     filename, T_OP_SET)) {
+				RDEBUG("Failed creating TLS-Client-Cert-Filename");
+
+				goto do_unlink;
+			}
+
+			RDEBUG("Verifying client certificate: %s",
+			       conf->verify_client_cert_cmd);
+			if (radius_exec_program(conf->verify_client_cert_cmd,
+						request, 1, NULL, 0,
+						EXEC_TIMEOUT,
+						request->packet->vps,
+						NULL, 1) != 0) {
+				radlog(L_AUTH, "rlm_eap_tls: Certificate CN (%s) fails external verification!", common_name);
+				my_ok = 0;
+			} else {
+				RDEBUG("Client certificate CN %s passed external validation", common_name);
+			}
+
+		do_unlink:
+			unlink(filename);
+			break;
+		}
+
+
 	} /* depth == 0 */
 
 	if (debug_flag > 0) {
@@ -375,6 +919,67 @@ static void eaptls_session_free(UNUSED void *parent, void *data_ptr,
 	pairfree(&vp);
 }
 
+#ifdef HAVE_OPENSSL_OCSP_H
+/*
+ * 	Create Global X509 revocation store and use it to verify
+ * 	OCSP responses
+ *
+ * 	- Load the trusted CAs
+ * 	- Load the trusted issuer certificates
+ */
+static X509_STORE *init_revocation_store(EAP_TLS_CONF *conf)
+{
+	X509_STORE *store = NULL;
+
+	store = X509_STORE_new();
+
+	/* Load the CAs we trust */
+        if (conf->ca_file || conf->ca_path)
+		if(!X509_STORE_load_locations(store, conf->ca_file, conf->ca_path)) {
+			radlog(L_ERR, "rlm_eap: X509_STORE error %s", ERR_error_string(ERR_get_error(), NULL));
+			radlog(L_ERR, "rlm_eap_tls: Error reading Trusted root CA list %s",conf->ca_file );
+			return NULL;
+		}
+
+#ifdef X509_V_FLAG_CRL_CHECK
+	if (conf->check_crl)
+		X509_STORE_set_flags(store, X509_V_FLAG_CRL_CHECK);
+#endif
+	return store;
+}
+#endif	/* HAVE_OPENSSL_OCSP_H */
+
+#if OPENSSL_VERSION_NUMBER >= 0x0090800fL
+#ifndef OPENSSL_NO_ECDH
+static int set_ecdh_curve(SSL_CTX *ctx, const char *ecdh_curve)
+{
+	int      nid;
+	EC_KEY  *ecdh;
+
+	if (!ecdh_curve || !*ecdh_curve) return 0;
+
+	nid = OBJ_sn2nid(ecdh_curve);
+	if (!nid) {
+		radlog(L_ERR, "Unknown ecdh_curve \"%s\"", ecdh_curve);
+		return -1;
+	}
+
+	ecdh = EC_KEY_new_by_curve_name(nid);
+	if (!ecdh) {
+		radlog(L_ERR, "Unable to create new curve \"%s\"", ecdh_curve);
+		return -1;
+	}
+
+	SSL_CTX_set_tmp_ecdh(ctx, ecdh);
+
+	SSL_CTX_set_options(ctx, SSL_OP_SINGLE_ECDH_USE);
+
+	EC_KEY_free(ecdh);
+
+	return 0;
+}
+#endif
+#endif
 
 /*
  *	Create Global context SSL and use it in every new session
@@ -393,11 +998,10 @@ static SSL_CTX *init_tls_ctx(EAP_TLS_CONF *conf)
 	int type;
 
 	/*
-	 *	Add all the default ciphers and message digests
-	 *	Create our context.
+	 *	Bug fix
+	 *	http://old.nabble.com/Backward-compatibility-of-private-key-files--td27937046.html
 	 */
-	SSL_library_init();
-	SSL_load_error_strings();
+	OpenSSL_add_all_algorithms();
 
 	/*
 	 *	SHA256 is in all versions of OpenSSL, but isn't
@@ -424,6 +1028,49 @@ static SSL_CTX *init_tls_ctx(EAP_TLS_CONF *conf)
 	 * Set the password to load private key
 	 */
 	if (conf->private_key_password) {
+#ifdef __APPLE__
+		/*
+		 * We don't want to put the private key password in eap.conf, so  check
+		 * for our special string which indicates we should get the password
+		 * programmatically.
+		 */
+		const char* special_string = "Apple:UseCertAdmin";
+		if (strncmp(conf->private_key_password,
+					special_string,
+					strlen(special_string)) == 0)
+		{
+			char cmd[256];
+			const long max_password_len = 128;
+			snprintf(cmd, sizeof(cmd) - 1,
+					 "/usr/sbin/certadmin --get-private-key-passphrase \"%s\"",
+					 conf->private_key_file);
+
+			DEBUG2("rlm_eap: Getting private key passphrase using command \"%s\"", cmd);
+
+			FILE* cmd_pipe = popen(cmd, "r");
+			if (!cmd_pipe) {
+				radlog(L_ERR, "rlm_eap: %s command failed.	Unable to get private_key_password", cmd);
+				radlog(L_ERR, "rlm_eap: Error reading private_key_file %s", conf->private_key_file);
+				return NULL;
+			}
+
+			free(conf->private_key_password);
+			conf->private_key_password = malloc(max_password_len * sizeof(char));
+			if (!conf->private_key_password) {
+				radlog(L_ERR, "rlm_eap: Can't malloc space for private_key_password");
+				radlog(L_ERR, "rlm_eap: Error reading private_key_file %s", conf->private_key_file);
+				pclose(cmd_pipe);
+				return NULL;
+			}
+
+			fgets(conf->private_key_password, max_password_len, cmd_pipe);
+			pclose(cmd_pipe);
+
+			/* Get rid of newline at end of password. */
+			conf->private_key_password[strlen(conf->private_key_password) - 1] = '\0';
+			DEBUG2("rlm_eap:  Password from command = \"%s\"", conf->private_key_password);
+		}
+#endif
 		SSL_CTX_set_default_passwd_cb_userdata(ctx, conf->private_key_password);
 		SSL_CTX_set_default_passwd_cb(ctx, cbtls_password);
 	}
@@ -478,6 +1125,9 @@ static SSL_CTX *init_tls_ctx(EAP_TLS_CONF *conf)
 	 */
 	ctx_options |= SSL_OP_NO_SSLv2;
    	ctx_options |= SSL_OP_NO_SSLv3;
+#ifdef SSL_OP_NO_TICKET
+	ctx_options |= SSL_OP_NO_TICKET ;
+#endif
 
 	/*
 	 *	SSL_OP_SINGLE_DH_USE must be used in order to prevent
@@ -506,6 +1156,17 @@ static SSL_CTX *init_tls_ctx(EAP_TLS_CONF *conf)
 	 *	SSL_CTX_set_tmp_dh_callback(ctx, cbtls_dh);
 	 */
 
+	/*
+	 *	Set eliptical curve crypto configuration.
+	 */
+#if OPENSSL_VERSION_NUMBER >= 0x0090800fL
+#ifndef OPENSSL_NO_ECDH
+	if (set_ecdh_curve(ctx, conf->ecdh_curve) < 0) {
+		return NULL;
+	}
+#endif
+#endif
+
 	/*
 	 *	set the message callback to identify the type of
 	 *	message.  For every new session, there can be a
@@ -519,7 +1180,7 @@ static SSL_CTX *init_tls_ctx(EAP_TLS_CONF *conf)
 
 	/*
 	 *	Callbacks, etc. for session resumption.
-	 */						      
+	 */
 	if (conf->session_cache_enable) {
 		SSL_CTX_sess_set_new_cb(ctx, cbtls_new_session);
 		SSL_CTX_sess_set_get_cb(ctx, cbtls_get_session);
@@ -557,10 +1218,12 @@ static SSL_CTX *init_tls_ctx(EAP_TLS_CONF *conf)
 	}
 
 	/* Load randomness */
-	if (!(RAND_load_file(conf->random_file, 1024*1024))) {
-		radlog(L_ERR, "rlm_eap: SSL error %s", ERR_error_string(ERR_get_error(), NULL));
-		radlog(L_ERR, "rlm_eap_tls: Error loading randomness");
-		return NULL;
+	if (conf->random_file) {
+		if (!(RAND_load_file(conf->random_file, 1024*1024))) {
+			radlog(L_ERR, "rlm_eap: SSL error %s", ERR_error_string(ERR_get_error(), NULL));
+			radlog(L_ERR, "rlm_eap_tls: Error loading randomness");
+			return NULL;
+		}
 	}
 
 	/*
@@ -583,19 +1246,19 @@ static SSL_CTX *init_tls_ctx(EAP_TLS_CONF *conf)
 		if (conf->session_id_name) {
 			snprintf(conf->session_context_id,
 				 sizeof(conf->session_context_id),
-				 "FreeRADIUS EAP-TLS %s",
+				 "FR eap %s",
 				 conf->session_id_name);
 		} else {
 			snprintf(conf->session_context_id,
 				 sizeof(conf->session_context_id),
-				 "FreeRADIUS EAP-TLS %p", conf);
+				 "FR eap %p", conf);
 		}
 
 		/*
 		 *	Cache it, and DON'T auto-clear it.
 		 */
 		SSL_CTX_set_session_cache_mode(ctx, SSL_SESS_CACHE_SERVER | SSL_SESS_CACHE_NO_AUTO_CLEAR);
-					       
+
 		SSL_CTX_set_session_id_context(ctx,
 					       (unsigned char *) conf->session_context_id,
 					       (unsigned int) strlen(conf->session_context_id));
@@ -604,7 +1267,7 @@ static SSL_CTX *init_tls_ctx(EAP_TLS_CONF *conf)
 		 *	Our timeout is in hours, this is in seconds.
 		 */
 		SSL_CTX_set_timeout(ctx, conf->session_timeout * 3600);
-		
+
 		/*
 		 *	Set the maximum number of entries in the
 		 *	session cache.
@@ -625,14 +1288,19 @@ static SSL_CTX *init_tls_ctx(EAP_TLS_CONF *conf)
 		eaptls_handle_idx = SSL_get_ex_new_index(0, "eaptls_handle_idx",
 							  NULL, NULL, NULL);
 	}
-	
+
 	if (eaptls_conf_idx < 0) {
 		eaptls_conf_idx = SSL_get_ex_new_index(0, "eaptls_conf_idx",
 							  NULL, NULL, NULL);
 	}
 
+	if (eaptls_store_idx < 0) {
+		eaptls_store_idx = SSL_get_ex_new_index(0, "eaptls_store_idx",
+							  NULL, NULL, NULL);
+	}
+
 	if (eaptls_session_idx < 0) {
-		eaptls_session_idx = SSL_get_ex_new_index(0, "eaptls_session_idx",
+		eaptls_session_idx = SSL_SESSION_get_ex_new_index(0, "eaptls_session_idx",
 							  NULL, NULL,
 							  eaptls_session_free);
 	}
@@ -650,17 +1318,16 @@ static int eaptls_detach(void *arg)
 	eap_tls_t 	 *inst;
 
 	inst = (eap_tls_t *) arg;
-	conf = inst->conf;
-
-	if (conf) {
-		memset(conf, 0, sizeof(*conf));
-		free(inst->conf);
-		inst->conf = NULL;
-	}
+	conf = &(inst->conf);
 
 	if (inst->ctx) SSL_CTX_free(inst->ctx);
 	inst->ctx = NULL;
 
+#ifdef HAVE_OPENSSL_OCSP_H
+	if (inst->store) X509_STORE_free(inst->store);
+	inst->store = NULL;
+#endif
+
 	free(inst);
 
 	return 0;
@@ -682,20 +1349,14 @@ static int eaptls_attach(CONF_SECTION *cs, void **instance)
 		return -1;
 	}
 	memset(inst, 0, sizeof(*inst));
+	conf = &(inst->conf);
 
 	/*
-	 *	Parse the config file & get all the configured values
+	 *	Hack: conf is the first structure inside of inst.  The
+	 *	CONF_PARSER stuff above uses offsetof() and
+	 *	EAP_TLS_CONF, which is technically wrong.
 	 */
-	conf = (EAP_TLS_CONF *)malloc(sizeof(*conf));
-	if (conf == NULL) {
-		free(inst);
-		radlog(L_ERR, "rlm_eap_tls: out of memory");
-		return -1;
-	}
-	memset(conf, 0, sizeof(*conf));
-
-	inst->conf = conf;
-	if (cf_section_parse(cs, conf, module_config) < 0) {
+	if (cf_section_parse(cs, inst, module_config) < 0) {
 		eaptls_detach(inst);
 		return -1;
 	}
@@ -743,7 +1404,8 @@ static int eaptls_attach(CONF_SECTION *cs, void **instance)
 		    (stat(conf->certificate_file, &buf) < 0) &&
 		    (errno == ENOENT) &&
 		    (radius_exec_program(conf->make_cert_command, NULL, 1,
-					 NULL, 0, NULL, NULL, 0) != 0)) {
+					 NULL, 0, EXEC_TIMEOUT,
+					 NULL, NULL, 0) != 0)) {
 			eaptls_detach(inst);
 			return -1;
 		}
@@ -759,15 +1421,43 @@ static int eaptls_attach(CONF_SECTION *cs, void **instance)
 		return -1;
 	}
 
+#ifdef HAVE_OPENSSL_OCSP_H
+	/*
+	 * 	Initialize OCSP Revocation Store
+	 */
+	if (conf->ocsp_enable) {
+		inst->store = init_revocation_store(conf);
+		if (inst->store == NULL) {
+			eaptls_detach(inst);
+		  return -1;
+		}
+	}
+#endif /*HAVE_OPENSSL_OCSP_H*/
+
 	if (load_dh_params(inst->ctx, conf->dh_file) < 0) {
 		eaptls_detach(inst);
 		return -1;
 	}
 
         if (generate_eph_rsa_key(inst->ctx) < 0) {
+		eaptls_detach(inst);
                 return -1;
         }
 
+	if (conf->verify_tmp_dir) {
+		if (chmod(conf->verify_tmp_dir, S_IRWXU) < 0) {
+			radlog(L_ERR, "rlm_eap_tls: Failed changing permissions on %s: %s", conf->verify_tmp_dir, strerror(errno));
+			eaptls_detach(inst);
+			return -1;
+		}
+	}
+
+	if (conf->verify_client_cert_cmd && !conf->verify_tmp_dir) {
+		radlog(L_ERR, "rlm_eap_tls: You MUST set the verify directory in order to use verify_client_cmd");
+		eaptls_detach(inst);
+		return -1;
+	}
+
 	*instance = inst;
 
 	return 0;
@@ -805,6 +1495,9 @@ static int eaptls_initiate(void *type_arg, EAP_HANDLER *handler)
 
 	inst = (eap_tls_t *)type_arg;
 
+	handler->tls = TRUE;
+	handler->finished = FALSE;
+
 	/*
 	 *	Manually flush the sessions every so often.  If HALF
 	 *	of the session lifetime has passed since we last
@@ -812,13 +1505,13 @@ static int eaptls_initiate(void *type_arg, EAP_HANDLER *handler)
 	 *
 	 *	FIXME: Also do it every N sessions?
 	 */
-	if (inst->conf->session_cache_enable &&
-	    ((inst->conf->session_last_flushed + (inst->conf->session_timeout * 1800)) <= request->timestamp)) {
+	if (inst->conf.session_cache_enable &&
+	    ((inst->conf.session_last_flushed + (inst->conf.session_timeout * 1800)) <= request->timestamp)) {
 		RDEBUG2("Flushing SSL sessions (of #%ld)",
 			SSL_CTX_sess_number(inst->ctx));
 
 		SSL_CTX_flush_sessions(inst->ctx, request->timestamp);
-		inst->conf->session_last_flushed = request->timestamp;
+		inst->conf.session_last_flushed = request->timestamp;
 	}
 
 	/*
@@ -869,9 +1562,12 @@ static int eaptls_initiate(void *type_arg, EAP_HANDLER *handler)
 	 *	this index should be global.
 	 */
 	SSL_set_ex_data(ssn->ssl, 0, (void *)handler);
-	SSL_set_ex_data(ssn->ssl, 1, (void *)inst->conf);
+	SSL_set_ex_data(ssn->ssl, 1, (void *)&(inst->conf));
+#ifdef HAVE_OPENSSL_OCSP_H
+	SSL_set_ex_data(ssn->ssl, 2, (void *)inst->store);
+#endif
 
-	ssn->length_flag = inst->conf->include_length;
+	ssn->length_flag = inst->conf.include_length;
 
 	/*
 	 *	We use default fragment size, unless the Framed-MTU
@@ -885,7 +1581,7 @@ static int eaptls_initiate(void *type_arg, EAP_HANDLER *handler)
 	 *	of EAP-TLS in order to calculate fragment sizes is
 	 *	just too much.
 	 */
-	ssn->offset = inst->conf->fragment_size;
+	ssn->offset = inst->conf.fragment_size;
 	vp = pairfind(handler->request->packet->vps, PW_FRAMED_MTU);
 	if (vp && ((vp->vp_integer - 14) < ssn->offset)) {
 		/*
@@ -947,7 +1643,7 @@ static int eaptls_initiate(void *type_arg, EAP_HANDLER *handler)
 		break;
 	}
 
-	if (inst->conf->session_cache_enable) {
+	if (inst->conf.session_cache_enable) {
 		ssn->allow_session_resumption = 1; /* otherwise it's zero */
 	}
 
@@ -988,6 +1684,48 @@ static int eaptls_authenticate(void *arg, EAP_HANDLER *handler)
 		 *	EAP-TLS-Success packet here.
 		 */
 	case EAPTLS_SUCCESS:
+		if (inst->conf.virtual_server) {
+			VALUE_PAIR *vp;
+			REQUEST *fake;
+
+			/* create a fake request */
+			fake = request_alloc_fake(request);
+			rad_assert(fake->packet->vps == NULL);
+
+			fake->packet->vps = paircopy(request->packet->vps);
+
+			/* set the virtual server to use */
+			if ((vp = pairfind(request->config_items,
+					   PW_VIRTUAL_SERVER)) != NULL) {
+				fake->server = vp->vp_strvalue;
+			} else {
+				fake->server = inst->conf.virtual_server;
+			}
+
+			RDEBUG("Processing EAP-TLS Certificate check:");
+			debug_pair_list(fake->packet->vps);
+
+			RDEBUG("server %s {", fake->server);
+
+			rad_virtual_server(fake);
+
+			RDEBUG("} # server %s", fake->server);
+
+			/* copy the reply vps back to our reply */
+			pairadd(&request->reply->vps, fake->reply->vps);
+			fake->reply->vps = NULL;
+
+			/* reject if virtual server didn't return accept */
+			if (fake->reply->code != PW_AUTHENTICATION_ACK) {
+				RDEBUG2("Certifictes were rejected by the virtual server");
+				request_free(&fake);
+				eaptls_fail(handler, 0);
+				return 0;
+			}
+
+			request_free(&fake);
+			/* success */
+		}
 		break;
 
 		/*
@@ -1034,7 +1772,7 @@ static int eaptls_authenticate(void *arg, EAP_HANDLER *handler)
 		 *	the client can't re-use it.
 		 */
 	default:
-		if (inst->conf->session_cache_enable) {	
+		if (inst->conf.session_cache_enable) {
 			SSL_CTX_remove_session(inst->ctx,
 					       tls_session->ssl->session);
 		}
@@ -1051,7 +1789,7 @@ static int eaptls_authenticate(void *arg, EAP_HANDLER *handler)
 		 *	FIXME: Store miscellaneous data.
 		 */
 		RDEBUG2("Adding user data to cached session");
-		
+
 #if 0
 		SSL_SESSION_set_ex_data(tls_session->ssl->session,
 					ssl_session_idx_user_session, session_data);
@@ -1093,3 +1831,4 @@ EAP_TYPE rlm_eap_tls = {
 	eaptls_authenticate,		/* authentication */
 	eaptls_detach			/* detach */
 };
+
