@@ -1,7 +1,7 @@
 /*
  * rlm_linelog.c
  *
- * Version:	$Id: rlm_linelog.c,v 1.17 2008/05/29 08:26:55 aland Exp $
+ * Version:	$Id: 2c322c1dbca57213d35efa18e11e35e060a947c2 $
  *
  *   This program is free software; you can redistribute it and/or modify
  *   it under the terms of the GNU General Public License as published by
@@ -22,7 +22,7 @@
  */
 
 #include <freeradius-devel/ident.h>
-RCSID("$Id: rlm_linelog.c,v 1.17 2008/05/29 08:26:55 aland Exp $")
+RCSID("$Id: 2c322c1dbca57213d35efa18e11e35e060a947c2 $")
 
 #include <freeradius-devel/radiusd.h>
 #include <freeradius-devel/modules.h>
@@ -31,32 +31,39 @@ RCSID("$Id: rlm_linelog.c,v 1.17 2008/05/29 08:26:55 aland Exp $")
 #include <fcntl.h>
 #endif
 
+#ifdef HAVE_UNISTD_H
+#include <unistd.h>
+#endif
+
+#ifdef HAVE_GRP_H
+#include <grp.h>
+#endif
+
 #ifdef HAVE_SYSLOG_H
 #include <syslog.h>
 
-#ifndef LOG_AUTHPRIV
-#define LOG_AUTHPRIV LOG_USER
-#endif
-
-#ifndef LOG_PID
-#define LOG_PID (0)
-#endif
-
 #ifndef LOG_INFO
 #define LOG_INFO (0)
 #endif
 #endif
 
+/*
+ *	Syslog facilities from main/mainconfig.c
+ */
+extern const FR_NAME_NUMBER syslog_str2fac[];
+
 /*
  *	Define a structure for our module configuration.
- *
- *	These variables do not need to be in a structure, but it's
- *	a lot cleaner to do so, and a pointer to the structure can
- *	be used as the instance handle.
  */
 typedef struct rlm_linelog_t {
+	CONF_SECTION	*cs;
 	char		*filename;
+	char		*syslog_facility;
+	int		facility;
+	int		permissions;
+	char		*group;
 	char		*line;
+	char		*reference;
 } rlm_linelog_t;
 
 /*
@@ -71,8 +78,16 @@ typedef struct rlm_linelog_t {
 static const CONF_PARSER module_config[] = {
 	{ "filename",  PW_TYPE_STRING_PTR,
 	  offsetof(rlm_linelog_t,filename), NULL,  NULL},
+	{ "syslog_facility",  PW_TYPE_STRING_PTR,
+	  offsetof(rlm_linelog_t,syslog_facility), NULL,  NULL},
+	{ "permissions",  PW_TYPE_INTEGER,
+	  offsetof(rlm_linelog_t,permissions), NULL,  "0600"},
+	{ "group",  PW_TYPE_STRING_PTR,
+	  offsetof(rlm_linelog_t,group), NULL,  NULL},
 	{ "format",  PW_TYPE_STRING_PTR,
 	  offsetof(rlm_linelog_t,line), NULL,  NULL},
+	{ "reference",  PW_TYPE_STRING_PTR,
+	  offsetof(rlm_linelog_t,reference), NULL,  NULL},
 	{ NULL, -1, 0, NULL, NULL }		/* end the list */
 };
 
@@ -81,7 +96,6 @@ static int linelog_detach(void *instance)
 {
 	rlm_linelog_t *inst = instance;
 
-
 	free(inst);
 	return 0;
 }
@@ -120,6 +134,19 @@ static int linelog_instantiate(CONF_SECTION *conf, void **instance)
 		linelog_detach(inst);
 		return -1;
 	}
+#else
+	inst->facility = 0;
+
+	if (inst->syslog_facility) {
+		inst->facility = fr_str2int(syslog_str2fac, inst->syslog_facility, -1);
+		if (inst->facility < 0) {
+			radlog(L_ERR, "rlm_linelog: Bad syslog facility '%s'", inst->syslog_facility);
+			linelog_detach(inst);
+			return -1;
+		}
+	}
+
+	inst->facility |= LOG_INFO;
 #endif
 
 	if (!inst->line) {
@@ -128,6 +155,7 @@ static int linelog_instantiate(CONF_SECTION *conf, void **instance)
 		return -1;
 	}
 
+	inst->cs = conf;
 	*instance = inst;
 
 	return 0;
@@ -199,11 +227,55 @@ static int do_linelog(void *instance, REQUEST *request)
 {
 	int fd = -1;
 	char buffer[4096];
+	char *p;
 	char line[1024];
-	rlm_linelog_t *inst;
+	rlm_linelog_t *inst = (rlm_linelog_t*) instance;
+	const char *value = inst->line;
 
-	inst = (rlm_linelog_t*) instance;
+#ifdef HAVE_GRP_H
+	gid_t gid;
+	struct group *grp;
+	char *endptr;
+#endif
 
+	if (inst->reference) {
+		CONF_ITEM *ci;
+		CONF_PAIR *cp;
+
+		radius_xlat(line + 1, sizeof(line) - 2, inst->reference,
+			    request, linelog_escape_func);
+		line[0] = '.';	/* force to be in current section */
+
+		/*
+		 *	Don't allow it to go back up
+		 */
+		if (line[1] == '.') goto do_log;
+
+		ci = cf_reference_item(NULL, inst->cs, line);
+		if (!ci) {
+			RDEBUG2("No such entry \"%s\"", line);
+			return RLM_MODULE_NOOP;
+		}
+
+		if (!cf_item_is_pair(ci)) {
+			RDEBUG2("Entry \"%s\" is not a variable assignment ", line);
+			goto do_log;
+		}
+
+		cp = cf_itemtopair(ci);
+		value = cf_pair_value(cp);
+		if (!value) {
+			RDEBUG2("Entry \"%s\" has no value", line);
+			goto do_log;
+		}
+
+		/*
+		 *	Value exists, but is empty.  Don't log anything.
+		 */
+		if (!*value) return RLM_MODULE_OK;
+	}
+
+ do_log:
 	/*
 	 *	FIXME: Check length.
 	 */
@@ -211,18 +283,49 @@ static int do_linelog(void *instance, REQUEST *request)
 		radius_xlat(buffer, sizeof(buffer), inst->filename, request,
 			    NULL);
 		
-		fd = open(buffer, O_WRONLY | O_APPEND | O_CREAT, 0600);
+		/* check path and eventually create subdirs */
+		p = strrchr(buffer,'/');
+		if (p) {
+			*p = '\0';
+			if (rad_mkdir(buffer, 0700) < 0) {
+				radlog_request(L_ERR, 0, request, "rlm_linelog: Failed to create directory %s: %s", buffer, strerror(errno));
+				return RLM_MODULE_FAIL;
+			}
+			*p = '/';
+		}
+
+		fd = open(buffer, O_WRONLY | O_APPEND | O_CREAT, inst->permissions);
 		if (fd == -1) {
 			radlog(L_ERR, "rlm_linelog: Failed to open %s: %s",
 			       buffer, strerror(errno));
 			return RLM_MODULE_FAIL;
 		}
+
+#ifdef HAVE_GRP_H
+		if (inst->group != NULL) {
+			gid = strtol(inst->group, &endptr, 10);
+			if (*endptr != '\0') {
+				grp = getgrnam(inst->group);
+				if (grp == NULL) {
+					RDEBUG2("Unable to find system group \"%s\"", inst->group);
+					goto skip_group;
+				}
+				gid = grp->gr_gid;
+			}
+
+			if (chown(buffer, -1, gid) == -1) {
+				RDEBUG2("Unable to change system group of \"%s\"", buffer);
+			}
+		}
+#endif
 	}
 
+ skip_group:
+
 	/*
 	 *	FIXME: Check length.
 	 */
-	radius_xlat(line, sizeof(line) - 1, inst->line, request,
+	radius_xlat(line, sizeof(line) - 1, value, request,
 		    linelog_escape_func);
 
 	if (fd >= 0) {
@@ -233,7 +336,7 @@ static int do_linelog(void *instance, REQUEST *request)
 
 #ifdef HAVE_SYSLOG_H
 	} else {
-		syslog(LOG_AUTHPRIV | LOG_PID | LOG_INFO, "%s", line);
+		syslog(inst->facility, "%s", line);
 #endif
 	}
 
@@ -259,5 +362,9 @@ module_t rlm_linelog = {
 		do_linelog, 	/* pre-proxy */
 		do_linelog,	/* post-proxy */
 		do_linelog	/* post-auth */
+#ifdef WITH_COA
+		, do_linelog,	/* recv-coa */
+		do_linelog	/* send-coa */
+#endif
 	},
 };
