@@ -1,7 +1,7 @@
 /*
  * rlm_mschap.c
  *
- * Version:	$Id$
+ * Version:	$Id: c26fff61f1987413400b72a756516789e5275aa4 $
  *
  *   This program is free software; you can redistribute it and/or modify
  *   it under the terms of the GNU General Public License as published by
@@ -20,30 +20,10 @@
  * Copyright 2000,2001,2006  The FreeRADIUS server project
  */
 
-
-/*
- *  mschap.c    MS-CHAP module
- *
- *  This implements MS-CHAP, as described in RFC 2548
- *
- *  http://www.freeradius.org/rfc/rfc2548.txt
- *
- */
-
-/*
- *  If you have any questions on NTLM (Samba) passwords
- *  support, LM authentication and MS-CHAP v2 support
- *  please contact
- *
- *  Vladimir Dubrovin	vlad@sandy.ru
- *  aka
- *  ZARAZA		3APA3A@security.nnov.ru
- */
-
 /*  MPPE support from Takahiro Wagatsuma <waga@sic.shibaura-it.ac.jp> */
 
 #include	<freeradius-devel/ident.h>
-RCSID("$Id$")
+RCSID("$Id: c26fff61f1987413400b72a756516789e5275aa4 $")
 
 #include	<freeradius-devel/radiusd.h>
 #include	<freeradius-devel/modules.h>
@@ -53,6 +33,7 @@ RCSID("$Id$")
 
 #include 	<ctype.h>
 
+#include	"mschap.h"
 #include	"smbdes.h"
 
 #ifdef __APPLE__
@@ -147,115 +128,6 @@ static int pdb_decode_acct_ctrl(const char *p)
 }
 
 
-/*
- *	ntpwdhash converts Unicode password to 16-byte NT hash
- *	with MD4
- */
-static void ntpwdhash (uint8_t *szHash, const char *szPassword)
-{
-	char szUnicodePass[513];
-	int nPasswordLen;
-	int i;
-
-	/*
-	 *	NT passwords are unicode.  Convert plain text password
-	 *	to unicode by inserting a zero every other byte
-	 */
-	nPasswordLen = strlen(szPassword);
-	for (i = 0; i < nPasswordLen; i++) {
-		szUnicodePass[i << 1] = szPassword[i];
-		szUnicodePass[(i << 1) + 1] = 0;
-	}
-
-	/* Encrypt Unicode password to a 16-byte MD4 hash */
-	fr_md4_calc(szHash, (uint8_t *) szUnicodePass, (nPasswordLen<<1) );
-}
-
-
-/*
- *	challenge_hash() is used by mschap2() and auth_response()
- *	implements RFC2759 ChallengeHash()
- *	generates 64 bit challenge
- */
-static void challenge_hash( const uint8_t *peer_challenge,
-			    const uint8_t *auth_challenge,
-			    const char *user_name, uint8_t *challenge )
-{
-	fr_SHA1_CTX Context;
-	uint8_t hash[20];
-
-	fr_SHA1Init(&Context);
-	fr_SHA1Update(&Context, peer_challenge, 16);
-	fr_SHA1Update(&Context, auth_challenge, 16);
-	fr_SHA1Update(&Context, (const uint8_t *) user_name,
-		      strlen(user_name));
-	fr_SHA1Final(hash, &Context);
-	memcpy(challenge, hash, 8);
-}
-
-/*
- *	auth_response() generates MS-CHAP v2 SUCCESS response
- *	according to RFC 2759 GenerateAuthenticatorResponse()
- *	returns 42-octet response string
- */
-static void auth_response(const char *username,
-			  const uint8_t *nt_hash_hash,
-			  uint8_t *ntresponse,
-			  uint8_t *peer_challenge, uint8_t *auth_challenge,
-			  char *response)
-{
-	fr_SHA1_CTX Context;
-	static const uint8_t magic1[39] =
-	{0x4D, 0x61, 0x67, 0x69, 0x63, 0x20, 0x73, 0x65, 0x72, 0x76,
-	 0x65, 0x72, 0x20, 0x74, 0x6F, 0x20, 0x63, 0x6C, 0x69, 0x65,
-	 0x6E, 0x74, 0x20, 0x73, 0x69, 0x67, 0x6E, 0x69, 0x6E, 0x67,
-	 0x20, 0x63, 0x6F, 0x6E, 0x73, 0x74, 0x61, 0x6E, 0x74};
-
-	static const uint8_t magic2[41] =
-	{0x50, 0x61, 0x64, 0x20, 0x74, 0x6F, 0x20, 0x6D, 0x61, 0x6B,
-	 0x65, 0x20, 0x69, 0x74, 0x20, 0x64, 0x6F, 0x20, 0x6D, 0x6F,
-	 0x72, 0x65, 0x20, 0x74, 0x68, 0x61, 0x6E, 0x20, 0x6F, 0x6E,
-	 0x65, 0x20, 0x69, 0x74, 0x65, 0x72, 0x61, 0x74, 0x69, 0x6F,
-	 0x6E};
-
-	static const char hex[16] = "0123456789ABCDEF";
-
-	size_t i;
-        uint8_t challenge[8];
-	uint8_t digest[20];
-
-	fr_SHA1Init(&Context);
-	fr_SHA1Update(&Context, nt_hash_hash, 16);
-	fr_SHA1Update(&Context, ntresponse, 24);
-	fr_SHA1Update(&Context, magic1, 39);
-	fr_SHA1Final(digest, &Context);
-	challenge_hash(peer_challenge, auth_challenge, username, challenge);
-	fr_SHA1Init(&Context);
-	fr_SHA1Update(&Context, digest, 20);
-	fr_SHA1Update(&Context, challenge, 8);
-	fr_SHA1Update(&Context, magic2, 41);
-	fr_SHA1Final(digest, &Context);
-
-	/*
-	 *	Encode the value of 'Digest' as "S=" followed by
-	 *	40 ASCII hexadecimal digits and return it in
-	 *	AuthenticatorResponse.
-	 *	For example,
-	 *	"S=0123456789ABCDEF0123456789ABCDEF01234567"
-	 */
- 	response[0] = 'S';
-	response[1] = '=';
-
-	/*
-	 *	The hexadecimal digits [A-F] MUST be uppercase.
-	 */
-	for (i = 0; i < sizeof(digest); i++) {
-		response[2 + (i * 2)] = hex[(digest[i] >> 4) & 0x0f];
-		response[3 + (i * 2)] = hex[digest[i] & 0x0f];
-	}
-}
-
-
 typedef struct rlm_mschap_t {
 	int use_mppe;
 	int require_encryption;
@@ -264,7 +136,10 @@ typedef struct rlm_mschap_t {
 	char *passwd_file;
 	const char *xlat_name;
 	char *ntlm_auth;
+	int ntlm_auth_timeout;
 	const char *auth_type;
+	int allow_retry;
+	char *retry_msg;
 #ifdef __APPLE__
 	int  open_directory;
 #endif  
@@ -319,9 +194,9 @@ static size_t mschap_xlat(void *instance, REQUEST *request,
 			 *	for MS-CHAPv2.
 			 */
 		} else if (chap_challenge->length == 16) {
+			VALUE_PAIR *name_attr, *response_name;
 			char *username_string;
 
-			RDEBUG2(" mschap2: %02x", chap_challenge->vp_octets[0]);
 			response = pairfind(request->packet->vps,
 					    PW_MSCHAP2_RESPONSE);
 			if (!response) {
@@ -329,6 +204,12 @@ static size_t mschap_xlat(void *instance, REQUEST *request,
 				return 0;
 			}
 
+			/*
+			 *	FIXME: Much of this is copied from
+			 *	below.  We should put it into a
+			 *	separate function.
+			 */
+
 			/*
 			 *	Responses are 50 octets.
 			 */
@@ -340,22 +221,44 @@ static size_t mschap_xlat(void *instance, REQUEST *request,
 			user_name = pairfind(request->packet->vps,
 					     PW_USER_NAME);
 			if (!user_name) {
-				RDEBUG2("User-Name is required to calculateMS-CHAPv1 Challenge.");
+				RDEBUG2("User-Name is required to calculate MS-CHAPv1 Challenge.");
 				return 0;
 			}
 
+ 			/*
+			 *      Check for MS-CHAP-User-Name and if found, use it
+			 *      to construct the MSCHAPv1 challenge.  This is
+			 *      set by rlm_eap_mschap to the MS-CHAP Response
+			 *      packet Name field.
+			 *
+			 *	We prefer this to the User-Name in the
+			 *	packet.
+			 */
+			response_name = pairfind(request->packet->vps, PW_MS_CHAP_USER_NAME);
+			if (response_name) {
+				name_attr = response_name;
+			} else {
+				name_attr = user_name;
+			}
+
 			/*
 			 *	with_ntdomain_hack moved here, too.
 			 */
-			if ((username_string = strchr(user_name->vp_strvalue, '\\')) != NULL) {
+			if ((username_string = strchr(name_attr->vp_strvalue, '\\')) != NULL) {
 				if (inst->with_ntdomain_hack) {
 					username_string++;
 				} else {
 					RDEBUG2("NT Domain delimeter found, should we have enabled with_ntdomain_hack?");
-					username_string = user_name->vp_strvalue;
+					username_string = name_attr->vp_strvalue;
 				}
 			} else {
-				username_string = user_name->vp_strvalue;
+				username_string = name_attr->vp_strvalue;
+			}
+
+			if (response_name &&
+			    ((user_name->length != response_name->length) ||
+			     (strncasecmp(user_name->vp_strvalue, response_name->vp_strvalue, user_name->length) != 0))) {
+				RDEBUG("WARNING: User-Name (%s) is not the same as MS-CHAP Name (%s) from EAP-MSCHAPv2", user_name->vp_strvalue, response_name->vp_strvalue);
 			}
 
 			/*
@@ -363,7 +266,9 @@ static size_t mschap_xlat(void *instance, REQUEST *request,
 			 *	from the MS-CHAPv2 peer challenge,
 			 *	our challenge, and the user name.
 			 */
-			challenge_hash(response->vp_octets + 2,
+			RDEBUG2("Creating challenge hash with username: %s",
+				username_string);
+			mschap_challenge_hash(response->vp_octets + 2,
 				       chap_challenge->vp_octets,
 				       username_string, buffer);
 			data = buffer;
@@ -549,7 +454,7 @@ static size_t mschap_xlat(void *instance, REQUEST *request,
 			return 0;
 		}
 
-		ntpwdhash(buffer,buf2);
+		mschap_ntpwdhash(buffer,buf2);
 
 		fr_bin2hex(buffer, out, 16);
 		out[32] = '\0';
@@ -631,6 +536,12 @@ static const CONF_PARSER module_config[] = {
 	  offsetof(rlm_mschap_t, passwd_file), NULL,  NULL },
 	{ "ntlm_auth",   PW_TYPE_STRING_PTR,
 	  offsetof(rlm_mschap_t, ntlm_auth), NULL,  NULL },
+	{ "ntlm_auth_timeout",   PW_TYPE_INTEGER,
+	  offsetof(rlm_mschap_t, ntlm_auth_timeout), NULL,  NULL },
+	{ "allow_retry",   PW_TYPE_BOOLEAN,
+	  offsetof(rlm_mschap_t, allow_retry), NULL,  "yes" },
+	{ "retry_msg",   PW_TYPE_STRING_PTR,
+	  offsetof(rlm_mschap_t, retry_msg), NULL,  NULL },
 #ifdef __APPLE__
 	{ "use_open_directory",    PW_TYPE_BOOLEAN,
 	  offsetof(rlm_mschap_t,open_directory), NULL, "yes" },
@@ -646,7 +557,7 @@ static const CONF_PARSER module_config[] = {
 static int mschap_detach(void *instance){
 #define inst ((rlm_mschap_t *)instance)
 	if (inst->xlat_name) {
-		xlat_unregister(inst->xlat_name, mschap_xlat);
+		xlat_unregister(inst->xlat_name, mschap_xlat, instance);
 		free(inst->xlat_name);
 	}
 	free(instance);
@@ -702,6 +613,23 @@ static int mschap_instantiate(CONF_SECTION *conf, void **instance)
 		inst->auth_type = inst->xlat_name;
 	}
 
+	/*
+	 *	Check ntlm_auth_timeout is sane
+	 */
+	if (!inst->ntlm_auth_timeout) {
+		inst->ntlm_auth_timeout = EXEC_TIMEOUT;
+	}
+	if (inst->ntlm_auth_timeout < 1) {
+		radlog(L_ERR, "rlm_mschap: ntml_auth_timeout '%d' is too small (minimum: 1)",
+			      inst->ntlm_auth_timeout);
+		return -1;
+	}
+	if (inst->ntlm_auth_timeout > 10) {
+		radlog(L_ERR, "rlm_mschap: ntlm_auth_timeout '%d' is too large (maximum: 10)",
+			      inst->ntlm_auth_timeout);
+		return -1;
+	}
+
 	return 0;
 }
 
@@ -770,7 +698,7 @@ static int do_mschap(rlm_mschap_t *inst,
 		}
 
 		smbdes_mschap(password->vp_strvalue, challenge, calculated);
-		if (memcmp(response, calculated, 24) != 0) {
+		if (rad_digest_cmp(response, calculated, 24) != 0) {
 			return -1;
 		}
 
@@ -796,9 +724,27 @@ static int do_mschap(rlm_mschap_t *inst,
 		result = radius_exec_program(inst->ntlm_auth, request,
 					     TRUE, /* wait */
 					     buffer, sizeof(buffer),
+					     inst->ntlm_auth_timeout,
 					     NULL, NULL, 1);
 		if (result != 0) {
+			char *p;
+			VALUE_PAIR *vp = NULL;
+
 			RDEBUG2("External script failed.");
+
+			vp = pairmake("Module-Failure-Message", "", T_OP_EQ);
+			if (!vp) {
+				radlog_request(L_ERR, 0, request, "No memory to allocate Module-Failure-Message");
+				return RLM_MODULE_FAIL;
+			}
+
+			p = strchr(buffer, '\n');
+			if (p) *p = '\0';
+			snprintf(vp->vp_strvalue, sizeof(vp->vp_strvalue),
+				"%s: External script says %s",
+				 inst->xlat_name, buffer);
+			vp->length = strlen(vp->vp_strvalue);
+			pairadd(&request->packet->vps, vp);
 			return -1;
 		}
 
@@ -981,7 +927,7 @@ static int mschap_authorize(void * instance, REQUEST *request)
 	}
 
 	if (pairfind(request->config_items, PW_AUTH_TYPE)) {
-		RDEBUG2("Found existing Auth-Type.  Not changing it.");
+		RDEBUG2("WARNING: Auth-Type already set.  Not setting to MS-CHAP");
 		return RLM_MODULE_NOOP;
 	}
 
@@ -1147,7 +1093,7 @@ static int mschap_authenticate(void * instance, REQUEST *request)
 			radlog_request(L_ERR, 0, request, "No memory");
 			return RLM_MODULE_FAIL;
 		} else {
-			ntpwdhash(nt_password->vp_octets,
+			mschap_ntpwdhash(nt_password->vp_octets,
 				  password->vp_strvalue);
 			nt_password->length = 16;
 		}
@@ -1155,7 +1101,7 @@ static int mschap_authenticate(void * instance, REQUEST *request)
 
 	challenge = pairfind(request->packet->vps, PW_MSCHAP_CHALLENGE);
 	if (!challenge) {
-		RDEBUG2("No MS-CHAP-Challenge in the request");
+		RDEBUG("ERROR: You set 'Auth-Type = MS-CHAP' for a request that does not contain any MS-CHAP attributes!");
 		return RLM_MODULE_REJECT;
 	}
 
@@ -1191,11 +1137,11 @@ static int mschap_authenticate(void * instance, REQUEST *request)
 		 *	response
 		 */
 		if (response->vp_octets[1] & 0x01) {
-			RDEBUG2("Told to do MS-CHAPv1 with NT-Password");
+			RDEBUG2("Client is using MS-CHAPv1 with NT-Password");
 			password = nt_password;
 			offset = 26;
 		} else {
-			RDEBUG2("Told to do MS-CHAPv1 with LM-Password");
+			RDEBUG2("Client is using MS-CHAPv1 with LM-Password");
 			password = lm_password;
 			offset = 2;
 		}
@@ -1207,16 +1153,14 @@ static int mschap_authenticate(void * instance, REQUEST *request)
 			      response->vp_octets + offset, nthashhash,
 			      do_ntlm_auth) < 0) {
 			RDEBUG2("MS-CHAP-Response is incorrect.");
-			mschap_add_reply(request, &request->reply->vps,
-					 *response->vp_octets,
-					 "MS-CHAP-Error", "E=691 R=1", 9);
-			return RLM_MODULE_REJECT;
+			goto do_error;
 		}
 
 		chap = 1;
 
 	} else if ((response = pairfind(request->packet->vps, PW_MSCHAP2_RESPONSE)) != NULL) {
 		uint8_t	mschapv1_challenge[16];
+		VALUE_PAIR *name_attr, *response_name;
 
 		/*
 		 *	MS-CHAPv2 challenges are 16 octets.
@@ -1243,19 +1187,51 @@ static int mschap_authenticate(void * instance, REQUEST *request)
 			return RLM_MODULE_INVALID;
 		}
 
-
 		/*
-		 *	with_ntdomain_hack moved here
+		 *      Check for MS-CHAP-User-Name and if found, use it
+		 *      to construct the MSCHAPv1 challenge.  This is
+		 *      set by rlm_eap_mschap to the MS-CHAP Response
+		 *      packet Name field.
+		 *
+		 *	We prefer this to the User-Name in the
+		 *	packet.
 		 */
-		if ((username_string = strchr(username->vp_strvalue, '\\')) != NULL) {
-		        if (inst->with_ntdomain_hack) {
-			        username_string++;
+		response_name = pairfind(request->packet->vps, PW_MS_CHAP_USER_NAME);
+		if (response_name) {
+			name_attr = response_name;
+		} else {
+			name_attr = username;
+		}
+		
+		/*
+		 *	with_ntdomain_hack moved here, too.
+		 */
+		if ((username_string = strchr(name_attr->vp_strvalue, '\\')) != NULL) {
+			if (inst->with_ntdomain_hack) {
+				username_string++;
 			} else {
-				RDEBUG2("  NT Domain delimeter found, should we have enabled with_ntdomain_hack?");
-				username_string = username->vp_strvalue;
+				RDEBUG2("NT Domain delimeter found, should we have enabled with_ntdomain_hack?");
+				username_string = name_attr->vp_strvalue;
 			}
 		} else {
-		        username_string = username->vp_strvalue;
+			username_string = name_attr->vp_strvalue;
+		}
+		
+		/*
+		 *	When the names are ASCII, they should be
+		 *	identical.  When the names are non-ASCII,
+		 *	User-Name is UTF-8, and MS-CHAP-User-Name is
+		 *	some local Windows character set.  So they
+		 *	can't be identical.  And because you don't
+		 *	know what the MS-CHAP character set is,
+		 *	there's no way to do ANY kind of comparison.
+		 *	They could be "bob" and "doug", and you'd have
+		 *	no idea.
+		 */
+		if (response_name &&
+		    ((username->length != response_name->length) ||
+		     (strncasecmp(username->vp_strvalue, response_name->vp_strvalue, username->length) != 0))) {
+			RDEBUG("WARNING: User-Name (%s) is not the same as MS-CHAP Name (%s) from EAP-MSCHAPv2", username->vp_strvalue, response_name->vp_strvalue);
 		}
 
 #ifdef __APPLE__
@@ -1282,31 +1258,45 @@ static int mschap_authenticate(void * instance, REQUEST *request)
 		 *	MS-CHAPv2 takes some additional data to create an
 		 *	MS-CHAPv1 challenge, and then does MS-CHAPv1.
 		 */
-		challenge_hash(response->vp_octets + 2, /* peer challenge */
+		RDEBUG2("Creating challenge hash with username: %s",
+			username_string);
+		mschap_challenge_hash(response->vp_octets + 2, /* peer challenge */
 			       challenge->vp_octets, /* our challenge */
 			       username_string,	/* user name */
 			       mschapv1_challenge); /* resulting challenge */
 
-		RDEBUG2("Told to do MS-CHAPv2 for %s with NT-Password",
+		RDEBUG2("Client is using MS-CHAPv2 for %s, we need NT-Password",
 		       username_string);
 
 		if (do_mschap(inst, request, nt_password, mschapv1_challenge,
 			      response->vp_octets + 26, nthashhash,
 			      do_ntlm_auth) < 0) {
+			int i;
+			char buffer[128];
+
 			RDEBUG2("FAILED: MS-CHAP2-Response is incorrect");
+
+		do_error:
+			snprintf(buffer, sizeof(buffer), "E=691 R=%d",
+				 inst->allow_retry);
+
+			if (inst->retry_msg) {
+				snprintf(buffer + 9, sizeof(buffer) - 9, " C=");
+				for (i = 0; i < 16; i++) {
+					snprintf(buffer + 12 + i*2,
+						 sizeof(buffer) - 12 - i*2, "%02x",
+						 fr_rand() & 0xff);
+				}
+				snprintf(buffer + 44, sizeof(buffer) - 44,
+					 " V=3 M=%s", inst->retry_msg);
+			}
 			mschap_add_reply(request, &request->reply->vps,
-					 *response->vp_octets,
-					 "MS-CHAP-Error", "E=691 R=1", 9);
+					 *response->vp_octets, "MS-CHAP-Error",
+					 buffer, strlen(buffer));
 			return RLM_MODULE_REJECT;
 		}
 
-		/*
-		 *	Get the NT-hash-hash, if necessary
-		 */
-		if (nt_password) {
-		}
-
-		auth_response(username_string, /* without the domain */
+		mschap_auth_response(username_string, /* without the domain */
 			      nthashhash, /* nt-hash-hash */
 			      response->vp_octets + 26, /* peer response */
 			      response->vp_octets + 2, /* peer challenge */
@@ -1317,7 +1307,7 @@ static int mschap_authenticate(void * instance, REQUEST *request)
 		chap = 2;
 
 	} else {		/* Neither CHAPv1 or CHAPv2 response: die */
-		radlog_request(L_AUTH, 0, request, "No MS-CHAP response found");
+		RDEBUG("ERROR: You set 'Auth-Type = MS-CHAP' for a request that does not contain any MS-CHAP attributes!");
 		return RLM_MODULE_INVALID;
 	}
 
@@ -1334,8 +1324,8 @@ static int mschap_authenticate(void * instance, REQUEST *request)
 		 *	return 'not found'.
 		 */
 		if (((smb_ctrl->vp_integer & ACB_DISABLED) != 0) ||
-		    ((smb_ctrl->vp_integer & ACB_NORMAL) == 0)) {
-			RDEBUG2("SMB-Account-Ctrl says that the account is disabled, or is not a normal account.");
+		    ((smb_ctrl->vp_integer & (ACB_NORMAL|ACB_WSTRUST)) == 0)) {
+			RDEBUG2("SMB-Account-Ctrl says that the account is disabled, or is not a normal or workstatin trust account.");
 			mschap_add_reply(request, &request->reply->vps,
 					  *response->vp_octets,
 					  "MS-CHAP-Error", "E=691 R=1", 9);
@@ -1414,7 +1404,7 @@ static int mschap_authenticate(void * instance, REQUEST *request)
 module_t rlm_mschap = {
 	RLM_MODULE_INIT,
 	"MS-CHAP",
-	RLM_TYPE_THREAD_SAFE,		/* type */
+	RLM_TYPE_THREAD_SAFE | RLM_TYPE_HUP_SAFE,		/* type */
 	mschap_instantiate,		/* instantiation */
 	mschap_detach,		/* detach */
 	{
