@@ -1,7 +1,7 @@
 /*
  * rlm_pap.c
  *
- * Version:  $Id$
+ * Version:  $Id: 8ef2152df423e227697a76700c95a838775e6fe5 $
  *
  *   This program is free software; you can redistribute it and/or modify
  *   it under the terms of the GNU General Public License as published by
@@ -22,7 +22,7 @@
  */
 
 #include <freeradius-devel/ident.h>
-RCSID("$Id$")
+RCSID("$Id: 8ef2152df423e227697a76700c95a838775e6fe5 $")
 
 #include <freeradius-devel/radiusd.h>
 #include <freeradius-devel/modules.h>
@@ -98,11 +98,13 @@ static const FR_NAME_NUMBER header_names[] = {
 	{ "{clear}",	PW_CLEARTEXT_PASSWORD },
 	{ "{cleartext}", PW_CLEARTEXT_PASSWORD },
 	{ "{md5}",	PW_MD5_PASSWORD },
+	{ "{BASE64_MD5}",	PW_MD5_PASSWORD },
 	{ "{smd5}",	PW_SMD5_PASSWORD },
 	{ "{crypt}",	PW_CRYPT_PASSWORD },
 	{ "{sha}",	PW_SHA_PASSWORD },
 	{ "{ssha}",	PW_SSHA_PASSWORD },
 	{ "{nt}",	PW_NT_PASSWORD },
+	{ "{nthash}",	PW_NT_PASSWORD },
 	{ "{x-nthash}",	PW_NT_PASSWORD },
 	{ "{ns-mta-md5}", PW_NS_MTA_MD5_PASSWORD },
 	{ "{x- orcllmv}", PW_LM_PASSWORD },
@@ -143,7 +145,7 @@ static int pap_instantiate(CONF_SECTION *conf, void **instance)
 		pap_detach(inst);
                 return -1;
         }
-	if (inst->scheme == NULL || strlen(inst->scheme) == 0){
+	if (!inst->scheme || !*inst->scheme) {
 		radlog(L_ERR, "rlm_pap: No scheme defined");
 		pap_detach(inst);
 		return -1;
@@ -222,6 +224,8 @@ static int base64_decode (const char *src, uint8_t *dst)
 
 	num = (length + equals) / 4;
 
+	if (!num || (num > MAX_STRING_LEN)) return 0;
+
 	for (i = 0; i < num - 1; i++) {
 		if (!decode_it(src, dst)) return 0;
 		src += 4;
@@ -264,14 +268,13 @@ static void normify(REQUEST *request, VALUE_PAIR *vp, size_t min_length)
 	 *	Base 64 encoding.  It's at least 4/3 the original size,
 	 *	and we want to avoid division...
 	 */
-	if ((vp->length * 3) >= ((min_length * 4))) {
-		decoded = base64_decode(vp->vp_strvalue, buffer);
-		if (decoded >= min_length) {
-			RDEBUG2("Normalizing %s from base64 encoding", vp->name);
-			memcpy(vp->vp_octets, buffer, decoded);
-			vp->length = decoded;
-			return;
-		}
+	if (((vp->length * 3) >= ((min_length * 4))) &&
+	    ((decoded = base64_decode(vp->vp_strvalue, buffer)) > 0) &&
+	    (decoded >= min_length)) {
+		RDEBUG2("Normalizing %s from base64 encoding", vp->name);
+		memcpy(vp->vp_octets, buffer, decoded);
+		vp->length = decoded;
+		return;
 	}
 
 	/*
@@ -313,19 +316,44 @@ static int pap_authorize(void *instance, REQUEST *request)
 		{
 			int attr;
 			char *p, *q;
-			char buffer[64];
+			char buffer[128];
 			VALUE_PAIR *new_vp;
 
 			found_pw = TRUE;
+		redo:
 			q = vp->vp_strvalue;
 			p = strchr(q + 1, '}');
 			if (!p) {
+				int decoded;
+
 				/*
-				 *	FIXME: Turn it into a
-				 *	cleartext-password, unless it,
-				 *	or user-password already
-				 *	exists.
+				 *	Password already exists: use
+				 *	that instead of this one.
 				 */
+				if (pairfind(request->config_items, PW_USER_PASSWORD) ||
+				    pairfind(request->config_items, PW_CLEARTEXT_PASSWORD)) {
+					RDEBUG("Config already contains \"known good\" password.  Ignoring Password-With-Header");
+					break;
+				}
+
+				/*
+				 *	If it's binary, it may be
+				 *	base64 encoded.  Decode it,
+				 *	and re-write the attribute to
+				 *	have the decoded value.
+				 */
+				decoded = base64_decode(vp->vp_strvalue, buffer);
+				if ((decoded > 0) && (buffer[0] == '{') &&
+				    (strchr(buffer, '}') != NULL)) {
+					memcpy(vp->vp_octets, buffer, decoded);
+					vp->length = decoded;
+					goto redo;
+				}
+
+				RDEBUG("No {...} in Password-With-Header = \"%s\", re-writing to Cleartext-Password", vp->vp_strvalue);
+				radius_pairmake(request, &request->config_items,
+						"Cleartext-Password",
+						vp->vp_strvalue, T_OP_SET);
 				break;
 			}
 
@@ -343,8 +371,14 @@ static int pap_authorize(void *instance, REQUEST *request)
 			new_vp = radius_paircreate(request,
 						   &request->config_items,
 						   attr, PW_TYPE_STRING);
-			strcpy(new_vp->vp_strvalue, p + 1);/* bounds OK */
-			new_vp->length = strlen(new_vp->vp_strvalue);
+			
+			/*
+			 *	The data after the '}' may be binary,
+			 *	so we copy it via memcpy.
+			 */
+			new_vp->length = vp->length;
+			new_vp->length -= (p - q + 1);
+			memcpy(new_vp->vp_strvalue, p + 1, new_vp->length);
 
 			/*
 			 *	May be old-style User-Password with header.
@@ -440,7 +474,7 @@ static int pap_authorize(void *instance, REQUEST *request)
 	 *	Don't touch existing Auth-Types.
 	 */
 	if (auth_type) {
-		RDEBUG2("Found existing Auth-Type, not changing it.");
+		RDEBUG2("WARNING: Auth-Type already set.  Not setting to PAP");
 		return RLM_MODULE_NOOP;
 	}
 
@@ -487,16 +521,9 @@ static int pap_authenticate(void *instance, REQUEST *request)
 	char buff2[MAX_STRING_LEN + 50];
 	int scheme = PAP_ENC_INVALID;
 
-	if (!request->password){
-		radlog_request(L_AUTH, 0, request, "Attribute \"Password\" is required for authentication.");
-		return RLM_MODULE_INVALID;
-	}
-
-	/*
-	 *	Clear-text passwords are the only ones we support.
-	 */
-	if (request->password->attribute != PW_USER_PASSWORD) {
-		radlog_request(L_AUTH, 0, request, "Attribute \"User-Password\" is required for authentication. Cannot use \"%s\".", request->password->name);
+	if (!request->password ||
+	    (request->password->attribute != PW_USER_PASSWORD)) {
+		RDEBUG("ERROR: You set 'Auth-Type = PAP' for a request that does not contain a User-Password attribute!");
 		return RLM_MODULE_INVALID;
 	}
 
@@ -582,8 +609,10 @@ static int pap_authenticate(void *instance, REQUEST *request)
 	do_clear:
 		RDEBUG("Using clear text password \"%s\"",
 		      vp->vp_strvalue);
-		if (strcmp((char *) vp->vp_strvalue,
-			   (char *) request->password->vp_strvalue) != 0){
+		if ((vp->length != request->password->length) ||
+		    (rad_digest_cmp(vp->vp_strvalue,
+				request->password->vp_strvalue,
+				vp->length) != 0)) {
 			snprintf(module_fmsg,sizeof(module_fmsg),"rlm_pap: CLEAR TEXT password check failed");
 			goto make_msg;
 		}
@@ -594,9 +623,10 @@ static int pap_authenticate(void *instance, REQUEST *request)
 
 	case PAP_ENC_CRYPT:
 	do_crypt:
-		RDEBUG("Using CRYPT encryption.");
-		if (fr_crypt_check((char *) request->password->vp_strvalue,
-				     (char *) vp->vp_strvalue) != 0) {
+		RDEBUG("Using CRYPT password \"%s\"",
+		       vp->vp_strvalue);
+		if (fr_crypt_check(request->password->vp_strvalue,
+				   vp->vp_strvalue) != 0) {
 			snprintf(module_fmsg,sizeof(module_fmsg),"rlm_pap: CRYPT password check failed");
 			goto make_msg;
 		}
@@ -618,7 +648,7 @@ static int pap_authenticate(void *instance, REQUEST *request)
 		fr_MD5Update(&md5_context, request->password->vp_octets,
 			     request->password->length);
 		fr_MD5Final(digest, &md5_context);
-		if (memcmp(digest, vp->vp_octets, vp->length) != 0) {
+		if (rad_digest_cmp(digest, vp->vp_octets, vp->length) != 0) {
 			snprintf(module_fmsg,sizeof(module_fmsg),"rlm_pap: MD5 password check failed");
 			goto make_msg;
 		}
@@ -645,7 +675,7 @@ static int pap_authenticate(void *instance, REQUEST *request)
 		/*
 		 *	Compare only the MD5 hash results, not the salt.
 		 */
-		if (memcmp(digest, vp->vp_octets, 16) != 0) {
+		if (rad_digest_cmp(digest, vp->vp_octets, 16) != 0) {
 			snprintf(module_fmsg,sizeof(module_fmsg),"rlm_pap: SMD5 password check failed");
 			goto make_msg;
 		}
@@ -667,7 +697,7 @@ static int pap_authenticate(void *instance, REQUEST *request)
 		fr_SHA1Update(&sha1_context, request->password->vp_octets,
 			      request->password->length);
 		fr_SHA1Final(digest,&sha1_context);
-		if (memcmp(digest, vp->vp_octets, vp->length) != 0) {
+		if (rad_digest_cmp(digest, vp->vp_octets, vp->length) != 0) {
 			snprintf(module_fmsg,sizeof(module_fmsg),"rlm_pap: SHA1 password check failed");
 			goto make_msg;
 		}
@@ -691,7 +721,7 @@ static int pap_authenticate(void *instance, REQUEST *request)
 			   request->password->length);
 		fr_SHA1Update(&sha1_context, &vp->vp_octets[20], vp->length - 20);
 		fr_SHA1Final(digest,&sha1_context);
-		if (memcmp(digest, vp->vp_octets, 20) != 0) {
+		if (rad_digest_cmp(digest, vp->vp_octets, 20) != 0) {
 			snprintf(module_fmsg,sizeof(module_fmsg),"rlm_pap: SSHA password check failed");
 			goto make_msg;
 		}
@@ -717,7 +747,7 @@ static int pap_authenticate(void *instance, REQUEST *request)
 			goto make_msg;
 		}
 		if ((fr_hex2bin(digest, digest, 16) != vp->length) ||
-		    (memcmp(digest, vp->vp_octets, vp->length) != 0)) {
+		    (rad_digest_cmp(digest, vp->vp_octets, vp->length) != 0)) {
 			snprintf(module_fmsg,sizeof(module_fmsg),"rlm_pap: NT password check failed");
 			goto make_msg;
 		}
@@ -741,7 +771,7 @@ static int pap_authenticate(void *instance, REQUEST *request)
 			goto make_msg;
 		}
 		if ((fr_hex2bin(digest, digest, 16) != vp->length) ||
-		    (memcmp(digest, vp->vp_octets, vp->length) != 0)) {
+		    (rad_digest_cmp(digest, vp->vp_octets, vp->length) != 0)) {
 			snprintf(module_fmsg,sizeof(module_fmsg),"rlm_pap: LM password check failed");
 		make_msg:
 			RDEBUG("Passwords don't match");
@@ -803,7 +833,7 @@ static int pap_authenticate(void *instance, REQUEST *request)
 				     p - buff2);
 			fr_MD5Final(digest, &md5_context);
 		}
-		if (memcmp(digest, buff, 16) != 0) {
+		if (rad_digest_cmp(digest, buff, 16) != 0) {
 			snprintf(module_fmsg,sizeof(module_fmsg),"rlm_pap: NS-MTA-MD5 password check failed");
 			goto make_msg;
 		}
